/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.api;

import java.io.Serializable;
import java.util.BitSet;

import cf.common.Engine;
import cf.common.Param;
import cf.real.CyberEngineIdList;
import cf.util.Log;

public abstract class ApplicationKnowledge implements Serializable , Cloneable
{
    private static final long serialVersionUID = 1L;
    private String creator; // Creator of this information.
    private double creationTime; // Creation time of this information.
    private double activationTime; // Activation time of this information.
    private double expirationTime; // Expiration time of this information.

    public long engineId;

    /** Observers. */

    public String creator() { return creator; }
    public double creationTime() { return creationTime; }
    public double activationTime() { return activationTime; }
    public double expirationTime() { return expirationTime; }
    public boolean expired() { return Engine.time() >= expirationTime; }
    public long engineId() { return engineId; }

    /** Knowledge constructor. */

    public double transmissionProbability;

    public ApplicationKnowledge(CyberApplication app)
    {
        init_sets(this);
        this.creator = app.name();
        this.creationTime = app.time();
        this.activationTime = this.creationTime;
        this.expirationTime = Double.POSITIVE_INFINITY;
	this.transmissionProbability = 1.0;
    }

    public ApplicationKnowledge(CyberApplication app, double expirationDelay)
    {
        init_sets(this);
        assert(0 <= expirationDelay);

        this.creator = app.name();
        this.creationTime = app.time();
        this.activationTime = this.creationTime;
        this.expirationTime = this.creationTime + expirationDelay;
	this.transmissionProbability = 1.0;
    }

    public ApplicationKnowledge(CyberApplication app, double activationDelay, double expirationDelay)
    {
        init_sets(this);
        assert(0 <= activationDelay);
        assert(activationDelay < expirationDelay);

        this.creator = app.name();
        this.creationTime = app.time();
        this.activationTime = this.creationTime + activationDelay;
        this.expirationTime = this.creationTime + expirationDelay;
	this.transmissionProbability = 1.0;
    }

    /* constructors with no Application, not public, inside using purposes */

    /* constructor for IPKnowledge, PingRespKnowledge , and
       {NodeParam, NodeStatistics, LinkParam, LinkStatistics, BundleDeliveredStats ,BundleReceivedKnowledge} in KnowledgeManager
       this knowledge ,
       will be handled specially
       and IPKnowledge is not handled in KB, in world level
    */
    protected ApplicationKnowledge(String creator, double creationTime, double expirationTime)
    {
        this.creator = creator;
        this.creationTime = creationTime;
        this.activationTime = this.creationTime;
        this.expirationTime = expirationTime;
        init_sets(this);
    }

    public boolean related(ApplicationKnowledge k)
    {
        return this.getClass() == k.getClass() && creator.equals(k.creator());
    }

    public boolean subsumes(ApplicationKnowledge k)
    {
        return k.creationTime == creationTime;
    }

    public boolean replaces(ApplicationKnowledge k)
    {
        return k.creationTime < creationTime;
    }


    /** Print knowledge - to be extended by subclass. */

    public String logRepShort()
    {
        //if(Param.SILENT){ return ""; }      

        return Log.getClassName(this) + "(" + creator + "," + creationTime + ")";
    }

    public String logRepLong()
    {
        //if(Param.SILENT){ return ""; }
        return Log.getClassName(this) +
            (" Creator: " + creator) +
            (" TimeCreated: " + creationTime) + 
            (expirationTime < Double.POSITIVE_INFINITY ? " TimeToExpire: " + expirationTime : "");
    }

    public void handle(CyberApplication app)
    {
        app.handleKnowledge(this);
    }

    public void handleKnowledgeExpired(CyberApplication app) {
        app.handleKnowledgeExpired(this);
    }


    public BitSet aware;

    public CyberEngineIdList aware_engine;
    public CyberEngineIdList aware_engine() { return aware_engine; }

    public BitSet processedfor;
    public BitSet sentto;  

    private CyberEngineIdList sentto_engine;
    public CyberEngineIdList sentto_engine() { return sentto_engine; }

    public ApplicationKnowledge copy()
    {
        try 
            {
                ApplicationKnowledge k = (ApplicationKnowledge)clone();
                //ApplicationKnowledge k = new ApplicationKnowledge(creator, timeCreated, timeToExpire);
                init_sets(k);
      
                //k.aware_engine().addAll(this.aware_engine);
                synchronized(k.aware_engine()){
                    k.aware_engine = (CyberEngineIdList)(this.aware_engine).clone();
                }
                k.aware =  (BitSet)aware.clone();

                return k;
            } 
        catch (CloneNotSupportedException e)
            {
                e.printStackTrace();
                assert false;
                return null;
            }
    }

    /* will be called receving knowledge from UDP */
    public void clear_sets()
    {
        processedfor.clear();
        aware.clear();
        sentto.clear();
        sentto_engine.clear();
    }

    void init_sets(ApplicationKnowledge k)
    {
        k.aware = new BitSet(Param.MaxNodes);
        k.aware_engine = new CyberEngineIdList();
        k.processedfor = new BitSet(Param.MaxNodes);
        k.sentto = new BitSet(Param.MaxNodes);
        k.sentto_engine = new CyberEngineIdList();
    }

    public boolean fresh(ApplicationKnowledge k ) 
    {
        return k.replaces(this);
    }

    //public boolean equal(ApplicationKnowledge k) 
    //{
    //  return this.creationTime == k.creationTime() && this.creator().equals(k.creator()) &&
    //    this.getClass() == k.getClass();
    //}

    public void print()
    {
        Engine.rep.println(logRepLong());
    }

    public String toString() 
    {
        return Log.getClassName(this) + "(" + creator + "," + creationTime + ")";
    }
    
    // JS: It's called when it's replaced by other one.
    public void onReplaced(){}	
}

