/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/
// optimistic distributed mutex
// each node has a notion of available resources
// but node knowledge can be inconsistent
// mutual exclusion becomes a probabilistic property

package cf.app.adaptivemutex;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.sim.SimEngine;
import cf.util.Dist;

// knowledge ordering depends on creator and timestamp
// furthermore time is compared across nodes (need of synchronization)

class KnowledgeOwner extends ApplicationKnowledge
{
    String name;
    
    KnowledgeOwner(CyberApplication creator)
    {
	super(creator);
    }

    public boolean related(ApplicationKnowledge k)
    {
	return this.getClass() == k.getClass();
    }

    public boolean subsumes(ApplicationKnowledge k)
    {
	return k.creationTime() == creationTime();
    }

    public boolean replaces(ApplicationKnowledge k)
    {
	return k.creationTime() < creationTime();
    }

    public String logRepLong()
    {
	return super.logRepLong() + " Name: " + name;
    }
}

// not a stateless node
// local sate encapsulates program state (in critical  or not, encoded in events)

class MutexApp extends CyberApp
{
    double TargetFailProb = 0.05;

    double delay = Dist.dist.sampleUniform(0.0,10.0);
    double failprob = TargetFailProb;

    boolean requesting = false;
    boolean acquiring = false;
    boolean critical = false;

    String owner = null;

    int entries = 0;
    boolean violation = false;
    int violations = 0;

    double entrytime[] = new double[1000000];
    boolean isviolation[] = new boolean[1000000];

    int logcount = 0;

    MutexApp(CyberNode node)
    {	
	super(node);
    }

    public void initialize()
    {
	requesting = true;
	postEvent(Dist.dist.sampleUniform(0.0,10.0));
    }

    public void handleEvent()
    {
	log("handing event");
	if(requesting) handleRequestEvent();
	else if(acquiring) handleAcquireEvent();
	else if(critical) handleReleaseEvent();
    }

    void handleRequestEvent()
    {
	log("requesting resource");

	if(owner == null)
	    {
		log("no interference detected - acquiring");
		owner = name(); 
		postOwner();
		requesting = false;
		acquiring = true;
		// postEvent(Dist.dist.sampleUniform(0.0,2*delay));
		// postEvent(delay);
		postEvent(3.9);
	    }
	else
	    {
		if(owner != null && owner.equals(name()))
		    {
			log("!!!!!!!!!!!!!!!!!!!!!!!!");
		    }
		log("interference detected - backoff");
		initialize();
	    }
    }

    void handleAcquireEvent()
    {
	log("acquiring resource");
	if(owner != null && owner.equals(name()))
	    {
		log("no interference detected - critical");
		acquiring = false;
		critical = true;
		violation = false;
		entrytime[entries] = SimEngine.time(); 
		// postEvent(Dist.dist.sampleUniform(0.0,300.0));
		postEvent(Dist.dist.sampleUniform(0.0,300.0));
		// postEvent(150);
	    }
	else
	    {
		log("interference detected - backoff");
		acquiring = false;
		initialize();
	    }
    }

    void handleReleaseEvent()
    {
	log("resource release");
	critical = false;
	owner = null;
	postOwner();

	if(violation) 
	    {
		isviolation[entries] = true;
		violations++;

		log("violations/entries = " + violations + " / " + entries + " " + "(" + ((double)violations/(double)entries) + ")");
		log("entries/time = " + entries + " / " + time() + " " + "(" + ((double)entries/(double)time()) + ")");
		log("violations/time = " + violations + " / " + time() + " " + "(" + ((double)violations/(double)time()) + ")");
	    }

	int v = 0;
	int e = 0;
	for(int i = entries; i >= 0 && i >= entries - 1000; i--)
	    {
		e += 1; 
		if(isviolation[i]) v +=1;
	    }

	failprob = (double)v/(double)e;
	
	double error = failprob - TargetFailProb;
	delay = delay + 0.5 * error;
	if(delay < 0) delay = 0;

	logcount++;
	if(logcount%10 == 0)
	    log("violations/entries = " + violations + " / " + entries + " failprob = " + failprob + " error = " + error + " delay = " + delay);

	entries++;

	initialize();
    }

    void postOwner()
    {
	log("posting Owner = " + owner);
	KnowledgeOwner kowner = new KnowledgeOwner(this);
	kowner.name = owner;
	postKnowledge(kowner);
    }

    public void handleKnowledge(ApplicationKnowledge k)
    {
	if(k instanceof KnowledgeOwner) handleOwner((KnowledgeOwner)k);
    }

    boolean eq(String n1, String n2)
    {
	if(n1 == null && n2 == null) return true;
	if(n1 == null && n2 != null) return false;
	if(n2 == null && n1 != null) return false;
	return n1.equals(n2);
    }
    
    void handleOwner(KnowledgeOwner kowner)
    {
	log("handling Owner = " + kowner.name);
	owner = kowner.name;
	if(critical && owner != null && !owner.equals(name())) 
	    {
		log("violation detected");
		violation = true;
	    }
    }

}
