/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/
// optimistic distributed mutex
// each node has a notion of available resources
// but node knowledge can be inconsistent
// mutual exclusion becomes a probabilistic property

package cf.app.altmutex;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.util.Dist;

// knowledge ordering depends on creator and timestamp
// furthermore time is compared across nodes (need of synchronization)

class ResourceKnowledge extends ApplicationKnowledge
{
    int tokens;
    
    ResourceKnowledge(CyberApplication creator)
    {
	super(creator);
    }

    public boolean related(ApplicationKnowledge k)
    {
	return this.getClass() == k.getClass();
    }

    public String logRepLong()
    {
	return super.logRepLong() + " Tokens: " + tokens;
    }
}

// not a stateless node
// local sate encapsulates program state (in critical  or not, encoded in events)

class MutexApp extends CyberApp
{
    int tokens = -1;

    boolean requesting = false;
    boolean critical = false;

    MutexApp(CyberNode node)
    {	
	super(node);
    }

    public void initialize()
    {
	postResource(1);
	requesting = true;
	postEvent(100 + Dist.dist.sampleUniform(0.0,10.0));
    }

    public void handleEvent()
    {
	if(requesting) handleRequestEvent();
	else if(critical) handleReleaseEvent();
    }

    void handleRequestEvent()
    {
	log("resource request");
	if(tokens >= 1)
	    {
		log("resource granted");
		requesting = false;
		critical = true;
		postResource(0);
		postEvent(150);
	    }
	else
	    {
		log("resource not available");
		requesting = true;
		postEvent(100);
	    }
    }

    void handleReleaseEvent()
    {
	log("resource release");
	critical = false;
	postResource(1);
    }

    void postResource(int t)
    {
	log("posting resource - tokens = " + tokens);
	ResourceKnowledge kresource = new ResourceKnowledge(this);
	kresource.tokens = t;
	postKnowledge(kresource);
    }

    public void handleKnowledge(ApplicationKnowledge k)
    {
	if(k instanceof ResourceKnowledge) handleResource((ResourceKnowledge)k);
    }
    
    void handleResource(ResourceKnowledge kresource)
    {
	log("handling resource - tokens = " + kresource.tokens);
	tokens = kresource.tokens;
    }
}

