/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/
// Object-oriented dispatching of events and knowledge
// MutexKnowledge as an outer class and explicit MutexKnowledgeHandler interface
// Knowledge could be used by other node classes that implement MutexKnowledgeHandler interface

// optimistic distributed mutex
// each node has a notion of available resources
// but node knowledge can be inconsistent
// mutual exclusion becomes a probabilistic property

package cf.app.approximation;

import cf.api.ApplicationEvent;
import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.util.Dist;

// not a stateless node
// local sate encapsulates program state (in critical section or not, encoded in events)

interface ApproxKnowledgeHandler
{
    public void handleApproxKnowledge(ApproxKnowledge k);
}

// knowledge ordering depends on creator and timestamp
// furthermore time is compared across nodes (need of synchronization)

class ApproxKnowledge extends ApplicationKnowledge
{
    int request;

    public ApproxKnowledge(ApproxApp app)
    {
	super(app);
    }       
    
    public String logRepLong()
    {
	return super.logRepLong() + " Request: " + request;
    }
    
    public void handle(CyberApplication app) {
((ApproxKnowledgeHandler)app).handleApproxKnowledge((ApproxKnowledge)this); }
}

class ApproxApp extends CyberApp implements ApproxKnowledgeHandler
{
    static double AcquireDelay = 0;
    // static double AcquireDelay = 10; // less optimistic version

    int numnodes = 0;
    String node[] = new String[100];
    int request[] = new int[100];
    int requests = 0;

    boolean requesting = false;
    boolean acquiring = false;
    boolean critical = false;

    ApproxApp(CyberNode node)
    {	
	super(node);
    }

    public void initialize()
    {
	postApproxKnowledge(0);
	requesting = true;
	log("initial request event");
	postRequestEvent(100 + Dist.dist.sampleUniform(0.0,10.0));
    }

    public boolean known(String n)
    {
	for(int i = 0; i < numnodes; i++)
	    {
		if(node[i].equals(n)) return true;
	    }
	return false;
    }

    class RequestEvent extends ApplicationEvent
    {
	public RequestEvent(ApproxApp app, double delay) { super(app,delay); }
	public void handle() { ((ApproxApp)application()).handleRequestEvent(this); }
    }
    
    class AcquireEvent extends ApplicationEvent
    {
	public AcquireEvent(ApproxApp app, double delay) { super(app,delay); }
	public void handle() { ((ApproxApp)application()).handleAcquireEvent(this); }
    }
    
    class ReleaseEvent extends ApplicationEvent
    {
	public ReleaseEvent(ApproxApp app, double delay) { super(app,delay); }
	public void handle() { ((ApproxApp)application()).handleReleaseEvent(this); }
    }
    
    void postRequestEvent(double delay)
    {
	postEvent(new RequestEvent(this,delay));
    }

    public void handleRequestEvent(RequestEvent e)
    {
	log("attempting to request resource");
	if(requests == 0)
	    {
		log("request granted");
		postApproxKnowledge(1);
		requesting = false;
		acquiring = true;
		postAcquireEvent(AcquireDelay);
	    }
	else
	    {
		log("resource already requested");
		requesting = true;
		postRequestEvent(100);
	    }
    }

    void postAcquireEvent(double delay)
    {
	postEvent(new AcquireEvent(this,delay));
    }

    public void handleAcquireEvent(AcquireEvent e)
    {
	log("attempting to acquire resource");
	if(requests <= 1)
	    {
		log("resource granted");
		acquiring = false;
		critical = true;
		postReleaseEvent(50);
	    }
	else
	    {
		log("resource already requested");
		postApproxKnowledge(0);
		acquiring = false;
		requesting = true;
		postRequestEvent(100);
	    }
    }

    void postReleaseEvent(double delay)
    {
	postEvent(new ReleaseEvent(this,delay));
    }

    public void handleReleaseEvent(ReleaseEvent e)
    {
	log("request release");
	critical = false;
	postApproxKnowledge(0);
	// postRequestEvent(0); 
    }

    void postApproxKnowledge(int r)
    {
	log("posting request = " + r);
	ApproxKnowledge krequest = new ApproxKnowledge(this);
	krequest.request = r;
	postKnowledge(krequest);
    }

    public void handleApproxKnowledge(ApproxKnowledge knew)
    {
	log("handling request = " + knew.request);

	String n = knew.creator();

	if(!known(n))
	    {
		log("request for new node");
		node[numnodes] = knew.creator();
		request[numnodes] = knew.request;
		numnodes++;
	    }
	else
	    {
		for(int i = 0; i < numnodes; i++)
		    {
			if(knew.creator().equals(node[i]))
			    {
				log("request for known node");
				request[i] = knew.request;
			    }
		    }
	    }

	log("counting requests");

	requests = 0;	
	for(int i = 0; i < numnodes; i++)
	    {
		log("processing node " + node[i] + " request = " + request[i]);

		if(request[i] >= 0)
		    {
			requests += request[i];
		    }
	    }
	
	log("requests = " + requests);
    }

}
