/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/
// competitive computation example
// production node generates a computation task
// competition nodes try to perform computation task in a race

package cf.app.competition;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.util.Dist;

// creator and timestamp is not used
// knowledge ordering only depends on input and result
// currently results are assumed to be consistent

class TaskKnowledge extends ApplicationKnowledge
{
    double input = Double.NaN;
    double result = Double.NaN;

    public TaskKnowledge(CyberApplication creator)
    {
	super(creator);
    }       

    public boolean related(ApplicationKnowledge k)
    {
	return this.getClass() == k.getClass();
    }

    public boolean subsumes(ApplicationKnowledge k)
    {
	return ((TaskKnowledge)k).input == input && (Double.isNaN(((TaskKnowledge)k).result) || ((TaskKnowledge)k).result == result);
    }

    public String logRepLong()
    {
	return super.logRepLong() + 
	    " Input: " + input +
	    " Result: " + result;
    }
}

class ProductionApp extends CyberApp
{
    ProductionApp(CyberNode node)
    {
	super(node);
    }

    public void initialize()
    {
	log("initialization");
	postEvent(100);
    }

    public void handleEvent()
    {
	handleInitialTaskEvent();
    }

    void handleInitialTaskEvent()
    {
	double i = Dist.dist.sampleUniform(0.0,1.0);    
	log("initial task - input = " + i);
	postTask(i);
    }

    void postTask(double i)
    {
	log("posting task - input = " + i);

	TaskKnowledge ktask = new TaskKnowledge(this);
	ktask.input = i;
	postKnowledge(ktask);
    }
}

class CompetitionApp extends CyberApp
{
    double input = Double.NaN;
    double result = Double.NaN;
    boolean computationPerformed = false;

    CompetitionApp(CyberNode node)
    {
	super(node);
    }

    public void initialize()
    {
    }

    public void handleEvent()
    {
	handleComputeEvent();
    }

    void postTask(double input, double result)
    {
	log("posting task - input = " + input + " result = " + result);

	TaskKnowledge ktask = new TaskKnowledge(this);

	ktask.input = input;
	ktask.result = result;

	postKnowledge(ktask);
    }

    public void handleKnowledge(ApplicationKnowledge knew)
    {
	if(knew instanceof TaskKnowledge) handleTask((TaskKnowledge)knew);
    }
    
    void handleTask(TaskKnowledge knew)
    {
	log("handling task - input = " + knew.input + " result = " + knew.result + " from " + knew.creator());

	input = knew.input;
	result = knew.result;

	postEvent(Dist.dist.sampleUniform(0,10));
    }

    void handleComputeEvent()
    {
	log("handling computation event - input = " + input + " result = " + result);

	boolean pending = Double.isNaN(result);

	if(pending) 
	    {	    		
		double r = input+1.0;
		computationPerformed = true;
		log("performing computation - input = " + input + " result = " + r);
		postTask(input,r);
	    }
	else
	    {
		log("no computation - result already available");
	    }
    }

}
