/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/
// cooperative distributed computing example
// compute min and max of distributed values
// each node's contribution improves the solution (monotonicity)

package cf.app.computing;


import java.util.LinkedList;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.util.Dist;

// knowledge ordering does not use timestamps
// but sets of node ids

class IntervalKnowledge extends ApplicationKnowledge
{
    double min,max;
    StringList contrib;

    IntervalKnowledge(CyberApplication creator)
    {
	super(creator);
    }

    public boolean related(ApplicationKnowledge k)
    {
	return this.getClass() == k.getClass();
    }

    public boolean subsumes(ApplicationKnowledge k)
    {
	IntervalKnowledge ki = (IntervalKnowledge)k;
	return ki.contrib.sublist(contrib);
    }

    public String logRepLong()
    {
	return super.logRepLong() + 
	    " MinMax: [" + min + "," + max + "]" + 
	    " Contrib: " + contrib.toString();
    }
}

class StringList extends LinkedList<String>
{
    public StringList copy()
    {
	StringList l = new StringList();
	for(String n : this) l.add(n);
	return l;
    }

    public String toString()
    {
	String r = null;
	for(String pn : this) 
	    {
		if(r == null) r = ""; else r+= ", ";
		r += pn;
	    }
	return r;
    }
    
    public boolean addOnce(String s)
    {
	if(contains(s)) return false;
	add(s);
	return true;
    }

    public boolean addOnce(StringList l)
    {
	boolean changed = false;

	for(String s : l)
	    {
		if(addOnce(s)) changed = true;
	    }

	return changed;
    }

    public boolean sublist(StringList l)
    {
	for(String n : this) if(!l.contains(n)) return false;
	return true;
    }

    public boolean equal(StringList l)
    {
	return this.sublist(l) && l.sublist(this);
    }
}

// node is not stateless (e.g. node can have real world state or sensor reading)
// local state contains initial value (initialized once and never changed)
// other than that local state is abstraction of all knowledge generated and received

class ComputingApp extends CyberApp
{
    double value = Dist.dist.sampleUniform(0.0,1.0);
    
    StringList contrib;
    double min, max;

    ComputingApp(CyberNode node)
    {
	super(node);
    }

    public void initialize()
    {
	log("initialization");
	log("value = " + value);

	contrib = new StringList(); contrib.addOnce(name());
	min = max = value;

	postEvent(100);
    }

    public void handleEvent()
    {
	handleIntervalEvent();
    }

    void handleIntervalEvent()
    {
	requestInterval();
    }

    void requestInterval()
    {
	log("posting interval = [" + min + "," + max + "] from " + contrib.toString());

	IntervalKnowledge kinterval = new IntervalKnowledge(this);
	kinterval.contrib = contrib.copy();
	kinterval.min = min;
	kinterval.max = max;
	postKnowledge(kinterval);
    }

    public void handleKnowledge(ApplicationKnowledge knew)
    {
	if(knew instanceof IntervalKnowledge) handleInterval((IntervalKnowledge)knew);
    }
    
    void handleInterval(IntervalKnowledge knew)
    {
	log("handling interval = [" + knew.min + "," + knew.max + "] from " + knew.contrib.toString());

	boolean changed = contrib.addOnce(knew.contrib);

	if(knew.min < min) min = knew.min;
	if(knew.max > max) max = knew.max;

	if(changed) requestInterval();
    }
}
