/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/
// self-stabilizing consensus example

package cf.app.consensus;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.util.Dist;

// makes use of creator and timestamp to allow proposal updates
// but timestamp is only used locally (i.e. to compare knowledge from same creator)

class ProposalKnowledge extends ApplicationKnowledge
{
    int proposal;

    public ProposalKnowledge(CyberApplication creator)
    {
	super(creator);
    }       

    public boolean related(ApplicationKnowledge k)
    {
	return this.getClass() == k.getClass() && creator() == ((ProposalKnowledge)k).creator();
    }

    public boolean replaces(ApplicationKnowledge k)
    {
	return k.creationTime() <= creationTime();
    }

    public String logRepLong()
    {
	return super.logRepLong() + " proposal: " + proposal;
    }
}

// each consensus node makes use of node ids to keep track of proposal from all nodes (inclusing its own)

class ConsensusApp extends CyberApp
{
    static double MaxWaitingTime = 10;
    
    int myproposal = -1;

    int numnodes = 0;
    String node[] = new String[100];
    int proposal[] = new int[100];

    boolean scheduledInitialProposalEvent = false;
    boolean scheduledCheckProposalEvent = false;

    public ConsensusApp(CyberNode node)
    {
	super(node);
    }

    public boolean known(String r)
    {
	for(int i = 0; i < numnodes; i++)
	    {
		if(node[i].equals(r)) return true;
	    }
	return false;
    }

    public String logRepLong()
    {
	String res = "Representing Nodes: ";

	for(int i = 0; i < numnodes; i++)
	    {
		res += " " + node[i] + "[" + proposal[i] + "]";
	    }
	
	return res;
    }

    public void initialize()
    {
	log("initialization");

	scheduledInitialProposalEvent = true;
	postEvent(Dist.dist.sampleUniform(0,MaxWaitingTime));
    }

    public void handleEvent()
    {
	if(scheduledInitialProposalEvent) handleInitialProposalEvent();
	else if(scheduledCheckProposalEvent) handleCheckProposalEvent();
    }

    void handleInitialProposalEvent()
    {
	scheduledInitialProposalEvent = false;
	log("initial proposal event");

	int newproposal = Dist.dist.sampleUniform(0,2);
	log("initial proposal = " + newproposal);
	postProposal(newproposal);
    }

    void handleCheckProposalEvent()
    {
	scheduledCheckProposalEvent = false;
	log("check proposal event");

	int votes0 = 0;
	int votes1 = 0;
	
	for(int i = 0; i < numnodes; i++)
	    {
		log("processing node " + node[i] + " proposal = " + proposal[i]);

		if(proposal[i] >= 0)
		    {
			votes0 += 1 - proposal[i];
			votes1 += proposal[i];
		    }
	    }
	
	log("votes = " + votes0 + "/" + votes1);
		
	if(myproposal >= 0 && votes0 == votes1)
	    {
		int newproposal = 1 - myproposal;
		log("modified proposal = " + newproposal);
		postProposal(newproposal);
	    }
    }

    void postProposal(int p)
    {
	log("posting proposal = " + p);
	
	ProposalKnowledge kproposal = new ProposalKnowledge(this);
	kproposal.proposal = p;
	postKnowledge(kproposal);
    }

    public void handleKnowledge(ApplicationKnowledge knew)
    {
	if(knew instanceof ProposalKnowledge) handleProposal((ProposalKnowledge)knew);
    }

    void handleProposal(ProposalKnowledge knew)
    {
	log("handling proposal = " + knew.proposal + 
			       " for " + knew.creator() + " from " + knew.creator());

	String n = knew.creator();
	if(n.equals(name()))
	    {
		myproposal = knew.proposal;
		log("my proposal = " + myproposal);
	    }

	if(!known(n))
	    {
		log("proposal for new node");
		node[numnodes] = knew.creator();
		proposal[numnodes] = knew.proposal;
		numnodes++;
	    }
	else
	    {
		for(int i = 0; i < numnodes; i++)
		    {
			if(node[i].equals(knew.creator()))
			    {
				log("proposal for known node");
				proposal[i] = knew.proposal;
			    }
		    }
	    }

	scheduledCheckProposalEvent = true;
	postEvent(Dist.dist.sampleUniform(0,MaxWaitingTime));
    }
}
