/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/
// Fisher-style distributed mutex but with knowledge sharing instead of shared memory
// uses a single shared knowledge unit (approximating a shared register)
// but node knowledge can be inconsistent
// mutual exclusion becomes a probabilistic property

package cf.app.fishermutex;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.util.Dist;

// knowledge ordering depends on creator and timestamp
// furthermore time is compared across nodes (need of synchronization)

class KnowledgeOwner extends ApplicationKnowledge
{
    String name;
    
    KnowledgeOwner(CyberApplication creator)
    {
	super(creator);
    }

    public boolean related(ApplicationKnowledge k)
    {
	return this.getClass() == k.getClass();
    }

    public boolean subsumes(ApplicationKnowledge k)
    {
	return k.creationTime() == creationTime();
    }

    public boolean replaces(ApplicationKnowledge k)
    {
	return k.creationTime() < creationTime();
    }

    public String logRepLong()
    {
	return super.logRepLong() + " Name: " + name;
    }
}

// not a stateless node
// local sate encapsulates program state (in critical  or not, encoded in events)

class MutexApp extends CyberApp
{
    static double Delay = 1;

    boolean requesting = false;
    boolean acquiring = false;
    boolean critical = false;

    String owner = null;

    MutexApp(CyberNode node)
    {	
	super(node);
    }

    public void initialize()
    {
	requesting = true;
	postEvent(100 + Dist.dist.sampleUniform(0.0,10.0));
    }

    public void handleEvent()
    {
	log("handing event");
	if(requesting) handleRequestEvent();
	else if(acquiring) handleAcquireEvent();
	else if(critical) handleReleaseEvent();
    }

    void handleRequestEvent()
    {
	log("requesting resource");

	if(owner == null)
	    {
		log("no interference detected - acquiring");
		owner = name(); 
		postOwner();
		requesting = false;
		acquiring = true;
		postEvent(Delay);
	    }
	else
	    {
		log("interference detected - backoff");
		postEvent(100);
	    }


    }

    void handleAcquireEvent()
    {
	log("acquiring resource");
	if(owner != null && owner.equals(name()))
	    {
		log("no interference detected - critical");
		acquiring = false;
		critical = true;
		postEvent(150);
	    }
	else
	    {
		log("interference detected - backoff");
		acquiring = false;
		requesting = true;
		postEvent(100);
	    }
    }

    void handleReleaseEvent()
    {
	log("resource release");
	critical = false;
	owner = null;
	postOwner();
    }

    void postOwner()
    {
	log("posting Owner = " + owner);
	KnowledgeOwner kowner = new KnowledgeOwner(this);
	kowner.name = owner;
	postKnowledge(kowner);
    }

    public void handleKnowledge(ApplicationKnowledge k)
    {
	if(k instanceof KnowledgeOwner) handleOwner((KnowledgeOwner)k);
    }
    
    void handleOwner(KnowledgeOwner kowner)
    {
	log("handling Owner = " + kowner.name);
	owner = kowner.name;
    }

}

