/*

  Copyright 2010-2013 SRI International

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/


package cf.app.mason.artificialphysics; 

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.api.devices.Position2D;
import cf.api.devices.PositionApplication;
import cf.sim.mason.SimWorld2DMason;
import cf.util.Dist;
import cf.util.Log;
import sim.portrayal.simple.OvalPortrayal2D;

import java.util.Random; 

class PositionKnowledge extends ApplicationKnowledge
{
	double posx = Double.NaN;
	double posy = Double.NaN;

	PositionKnowledge(CyberApplication creator)
	{
		super(creator);
	}

	public String logRepLong()
	{
		return super.logRepLong() + " Position: " + "(" + posx + "," + posy + ")";
	}
}

class PhaseKnowledge extends ApplicationKnowledge
{
	int phase = 0;
	
	PhaseKnowledge(CyberApplication creator)
	{
		super(creator);
	}

	public String logRepLong()
	{
		return super.logRepLong() + " Phase " + phase;
	}
}

class ArtificialPhysicsApp extends CyberApp implements PositionApplication
{
	Position2D pos;

	int numnodes = 0;
	String node[] = new String[100];
	double posx[] = new double[100];
	double posy[] = new double[100];
	
	double velocity_x=0.0, velocity_y=0.0; 
	double dx[] = new double[100];
	double dy[] = new double[100];
	
	int phase = 0; 
	
	ArtificialPhysicsApp(CyberNode node, Position2D pos)
	{
		super(node);
		this.pos = pos;
		pos.addApplication(this);
	}    

	public double posx()
	{
		return pos.posx();
	}

	public double posy()
	{
		return pos.posy();
	}

	public void move(double x, double y)
	{
		pos.move(x,y);
	}

	public boolean known(String r)
	{
		for(int i = 0; i < numnodes; i++)
		{
			if(node[i].equals(r)) return true;
		}
		return false;
	}

	public String logRepLong()
	{
		return super.logRepLong() + " Position: " + "(" + posx() + "," + posy() + ")" + " KnownNodes: " + numnodes;
	}

	double distance(int i)
	{
		double dx = posx() - posx[i];
		double dy = posy() - posy[i];
		return Math.sqrt(dx*dx+dy*dy);
	}
	
	double goal_distance()
	{
		double dx = goal_x - posx();
		double dy = goal_y - posy();
		return Math.sqrt(dx*dx+dy*dy);
	}

	double goal_distance(int i)
	{
		double dx = goal_x - posx[i];
		double dy = goal_y - posy[i];
		return Math.sqrt(dx*dx+dy*dy);
	}

	double goal_distance(double x, double y)
	{
		double dx = goal_x - posx() - x;
		double dy = goal_y - posy() - y;
		return Math.sqrt(dx*dx+dy*dy);
	}
	
	boolean stable_formation()
	{
		if (Math.sqrt(velocity_x*velocity_x + velocity_y*velocity_y) > Cstable*R)
			return false;
		
		int n = 0;
		for (int i=0; i < numnodes; i++)
		{
			double d = Math.sqrt(dx[i]*dx[i] + dy[i]*dy[i]);
			if (d < Cstable*R)
				n++; 
		}
			
		if (n == numnodes)
			return true;
	
		return false;
	}
		
	boolean goal_reached(double x, double y)
	{
		int n = 0;
		
		double d = goal_distance(x, y);
		//System.out.println("my d  " + d +  " x " + (posx() + x) + " y " + (posy() + y));
		if (d < Ctarget*R)
			n++;
		
		for (int i=0; i < numnodes; i++)
		{
			d = goal_distance(i);
			if (d < Ctarget*R)
				n++;
			//System.out.println("i " + i + " d " + d + " n " + n + " x " + posx[i] + " y " + posy[i]);
		}
		
		if ((numnodes <= 6 && n == numnodes+1) || (numnodes > 6 && n > 6))
			return true;
		
		return false;
	}
	
	public void initialize()
	{
		super.initialize();
		log("initialization");	
		postPosition();
		postPhase(1);
		postEvent(0.0);
	}

	double R = 50.0;							// desired distance among nodes 
	double Fmax = 1.0;							// maximum force  
	double G = Fmax*R*R/(2.0*Math.sqrt(3.0));	// gravitational constant avoiding clusters
	
	double Vmax = 5.0;							// maximum velocity of nodes
	double viscosity_coeff = 0.5;				// viscosity coefficient for friction  
	
	double goal_x = SimWorld2DMason.XMAX*0.5;	// target position
	double goal_y = SimWorld2DMason.YMAX*0.5;	// target position
	
	double Fgoal = G/(R*R);						// goal force 

	double Tpstart;								// start time of a phase 
	double Tpmin = 100.0; 						// minimum duration of a phase 
	double Tpmax = 1000.0;						// maximum duration of a phase
	
	double Cstable = 0.05;						// constant for stability check
	double Ctarget = 1.5;						// constant for target check 
	
	void ap()
	{
		double force;
		
		for (int i=0; i < numnodes; i++)
		{
			double r = distance(i);
		
			if (r != 0)			force = G / (r * r);
			else 				force = Fmax;
			
			if (force > Fmax)  	force = Fmax;
			if (r > (R*1.5)) 	force = 0;
			
			double dx = posx() - posx[i];
			double dy = posy() - posy[i];
			
			if (r <= R)
			{
				velocity_x += (dx * force) / r;
				velocity_y += (dy * force) / r;
			}
			else if (r < (R*1.5))
			{
				velocity_x -= (dx * force) / r;
				velocity_y -= (dy * force) / r;
			}
		}
	}
	
	public void handleEvent()
	{
		if (numnodes == 0) 
		{
			move(posx(),posy());

			postEvent(1.0);
			return;
		}
		
		ap();
		
		// friction
		velocity_x *= viscosity_coeff;
		velocity_y *= viscosity_coeff;
		
		// goal force 
		double goal_dx = goal_x - center_x();
		double goal_dy = goal_y - center_y(); 
		double d = Math.sqrt(goal_dx*goal_dx + goal_dy*goal_dy);

		if (d != 0)
		{
			double f = Fgoal + (d/(4.0*R));		// weighted goal force 
						
			velocity_x += f * goal_dx / d;
			velocity_y += f * goal_dy / d; 
		}
		
		// cut-down
		double velocity = Math.sqrt(velocity_x * velocity_x + velocity_y * velocity_y);
		if (velocity > Vmax)
		{
			velocity_x *= (Vmax/velocity);
			velocity_y *= (Vmax/velocity);
		}

		// conditions for phase transition 
		if (time() < Tpstart+Tpmin)
		{
			//System.out.println("stay phase due to MIN time of a phase " + time() + " started at " + Tpstart);
			;
		}
		else if (time() > Tpstart+Tpmax)
		{
			//System.out.println("next phase due to MAX time of a phase " + time() + " started at " + Tpstart);
			postPhase(phase+1);
		}
		else if (stable_formation())
		{
			//System.out.println("formation at time " + time());
			if (goal_reached(velocity_x, velocity_y))
			{
				 Log.stdout.println("goal reached at time " + time());
				postPhase(phase+1);
			}
		}

		move(posx()+velocity_x,posy()+velocity_y);
		
		postEvent(1.0);
	}

	public void handlePositionUpdate()
	{	
		log("new position: (" + posx() + "," + posy() + ")");

		postPosition();
	}

	void postPosition()
	{
		log("posting position: (" + posx() + "," + posy() + ")");

		PositionKnowledge kposition = new PositionKnowledge(this);
		kposition.posx = posx();
		kposition.posy = posy();
		postKnowledge(kposition);
	}
	
	void postPhase(int p)
	{
		Log.stdout.println("posting phase: " + p);

		PhaseKnowledge kphase = new PhaseKnowledge(this);
		kphase.phase = p;
		postKnowledge(kphase);
	}
	
	public void handleKnowledge(ApplicationKnowledge knew)
	{
		if(knew instanceof PositionKnowledge) handlePosition((PositionKnowledge)knew);
		if(knew instanceof PhaseKnowledge)    handlePhase((PhaseKnowledge)knew);
	}

	void handlePosition(PositionKnowledge knew)
	{
		log("handling position: (" + knew.posx + "," + knew.posy + ")" + " from " + knew.creator());

		String n = knew.creator();
		if(n != name())
		{
			if(!known(n))
			{
				log("position for new node");
				node[numnodes] = knew.creator();
				dx[numnodes] = 0.0;
				dy[numnodes] = 0.0; 
				posx[numnodes] = knew.posx;
				posy[numnodes] = knew.posy;
				numnodes++;
			}
			else
			{
				for(int i = 0; i < numnodes; i++)
				{
					if(knew.creator().equals(node[i]))
					{
						log("position for known node");
						dx[i] = posx[i] - knew.posx;
						dy[i] = posy[i] - knew.posy;
						posx[i] = knew.posx;
						posy[i] = knew.posy;
					}
				}
			}
		}
	}

	double center_x()
	{
		double sx = posx();
		
		for(int i = 0; i < numnodes; i++)
			sx += posx[i];
		
		return sx/(double)(numnodes+1);
	}
	
	double center_y()
	{
		double sy = posy();
		
		for(int i = 0; i < numnodes; i++)
			sy += posy[i];
		
		return sy/(double)(numnodes+1);
	}
	
	void handlePhase(PhaseKnowledge knew)
	{
		// phase update 
		phase = knew.phase;
		Tpstart = time();
		
		// flight plan 
		if (phase % 4 == 1)
		{
			goal_x = SimWorld2DMason.XMAX*0.2;
			goal_y = SimWorld2DMason.YMAX*0.2;
		}
		else if (phase % 4 == 2)
		{
			goal_x = SimWorld2DMason.XMAX*0.8;
			goal_y = SimWorld2DMason.YMAX*0.8;
		}
		else if (phase % 4 == 3)
		{
			goal_x = SimWorld2DMason.XMAX*0.8;
			goal_y = SimWorld2DMason.YMAX*0.2;
		}
		else 
		{
			goal_x = SimWorld2DMason.XMAX*0.2;
			goal_y = SimWorld2DMason.YMAX*0.8;
		}
	}
}
