/*

  Copyright 2010-2013 SRI International  

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/


package cf.app.mason.simulatedannealing; 

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.api.devices.Position2D;
import cf.api.devices.PositionApplication;
import cf.sim.mason.SimWorld2DMason;
import cf.util.Log;

import java.util.Random; 

class PositionKnowledge extends ApplicationKnowledge
{
	double posx = Double.NaN;
	double posy = Double.NaN;

	PositionKnowledge(CyberApplication creator)
	{
		super(creator);
	}

	public String logRepLong()
	{
		return super.logRepLong() + " Position: " + "(" + posx + "," + posy + ")";
	}
}

class PhaseKnowledge extends ApplicationKnowledge
{
	int phase = 0;
	
	PhaseKnowledge(CyberApplication creator)
	{
		super(creator);
	}

	public String logRepLong()
	{
		return super.logRepLong() + " Phase " + phase;
	}
}

class SimulatedAnnealingApp extends CyberApp implements PositionApplication
{
	Position2D pos;

	int numnodes = 0;
	String node[] = new String[100];
	double posx[] = new double[100];
	double posy[] = new double[100];

	double velocity_x=0.0, velocity_y=0.0; 
	double dx[] = new double[100];
	double dy[] = new double[100];
	
	int phase = 0; 
	
	SimulatedAnnealingApp(CyberNode node, Position2D pos)
	{
		super(node);
		this.pos = pos;
		pos.addApplication(this);
	}    

	public double posx()
	{
		return pos.posx();
	}

	public double posy()
	{
		return pos.posy();
	}

	public void move(double x, double y)
	{
		pos.move(x,y);
	}

	public boolean known(String r)
	{
		for(int i = 0; i < numnodes; i++)
		{
			if(node[i].equals(r)) return true;
		}
		return false;
	}

	public String logRepLong()
	{
		return super.logRepLong() + " Position: " + "(" + posx() + "," + posy() + ")" + " KnownNodes: " + numnodes;
	}

	double distance(int i)
	{
		double dx = posx() - posx[i];
		double dy = posy() - posy[i];
		return Math.sqrt(dx*dx+dy*dy);
	}

	double distance(int i, double x, double y)
	{
		double dx = posx() + x - posx[i];
		double dy = posy() + y - posy[i];
		return Math.sqrt(dx*dx+dy*dy);
	}

	double goal_distance(int i)
	{
		double dx = goal_x - posx[i];
		double dy = goal_y - posy[i];
		return Math.sqrt(dx*dx+dy*dy);
	}

	double goal_distance(double x, double y)
	{
		double dx = goal_x - posx() - x;
		double dy = goal_y - posy() - y;
		return Math.sqrt(dx*dx+dy*dy);
	}
	
	boolean stable_formation(double x, double y)
	{
		if (Math.sqrt(x*x + y*y) > Cstable*R)
			return false;
		
		int n = 0;
		for (int i=0; i < numnodes; i++)
		{
			double d = Math.sqrt(dx[i]*dx[i] + dy[i]*dy[i]);
			if (d < Cstable*R)
				n++; 
		}
			
		if (n == numnodes)
			return true;
	
		return false;
	}
	
	boolean goal_reached(double x, double y)
	{
		int n = 0;
		
		double d = goal_distance(x, y);
		//System.out.println("my d  " + d +  " x " + (posx() + x) + " y " + (posy() + y));
		if (d < Ctarget*R)
			n++;
		
		for (int i=0; i < numnodes; i++)
		{
			d = goal_distance(i);
			if (d < Ctarget*R)
				n++;
			//System.out.println("i " + i + " d " + d + " n " + n + " x " + posx[i] + " y " + posy[i]);
		}
		
		if ((numnodes <= 6 && n == numnodes+1) || (numnodes > 6 && n > 6))
			return true;
		
		return false;
	}

	public void initialize()
	{
		super.initialize();
		log("initialization");	
		postPosition();
		postPhase(1);
		postEvent(0.0);
	}

	// lower utility is better 
	double R = 50.0;
	double G = R*R/(2.0*Math.sqrt(3.0));	// Fmax = 1.0
	double R1 = Math.sqrt(G);
	double U1 = 1.0; 						// Fmax 
	double C1 = U1 / (1.0/R1 - 1.0/R);
	double R2 = 100*R;
	double U2 = G / (R2 * R2);	// F at 1.5*R
	double C2 = U2 / (1.0/R - 1.0/R2);  
	
	double goal_x = SimWorld2DMason.XMAX*0.5;
	double goal_y = SimWorld2DMason.YMAX*0.5;

	double Wgoal = 0.00001;
		
	double Tpstart;								// start time of a phase 
	double Tpmin = 100.0; 						// minimum duration of a phase 
	double Tpmax = 1000.0;						// maximum duration of a phase
	
	double Cstable = 0.05;						// constant for stability check
	double Ctarget = 1.5;						// constant for target check 
	
	double utility(double x, double y) 
	{
		double utility = Double.POSITIVE_INFINITY;
	
		/*
		// use minimum   
		double min = Double.POSITIVE_INFINITY;
		for(int i = 0; i < numnodes; i++)
		{
			double r = distance(i);
			if (r < (R*1.5))
			{
				r = distance(i, x, y);
				if (r < min)
					min = r;
			}
		}
		utility = (min-R) * (min-R);				
		*/
		
		// use potential   
		double sum = 0;
		int n = 0;
		
		for(int i = 0; i < numnodes; i++)
		{	
			double r = distance(i);
			//if (r < (R*1.5))
			{
				r = distance(i, x, y);
				if (r <= R1)
					sum += R1 - r; 
				else if (r <= R)
					sum += C1/r - C1/R;
				else if (r <= R2)
					sum += C2/R - C2/r;
				else 
					sum += C2;
				n++;
			}
		}

		if (n != 0)
			utility = sum / (double)n;
		
		double goal_dx = posx() - goal_x;
		double goal_dy = posy() - goal_y;
		double goal_r = Math.sqrt(goal_dx*goal_dx + goal_dy*goal_dy);

		goal_dx = posx()+x - goal_x;
		goal_dy = posy()+y - goal_y;
		goal_r = Math.sqrt(goal_dx*goal_dx + goal_dy*goal_dy);

		utility += Wgoal*goal_r*goal_r;
		
		//System.out.println("utility " + utility);
		return utility;
	}
    	
	// Simulated Annealing Method 
    // L. Nolle et al. One Step Width Adaptation in SA for Continuous Params Opt.
    // LNCS 2206, pp589-598, 2001	

    // Exponential Cooling Scheme (ECS) 
    // cdq (cooling coef) = T_final / T_init
    // T_final * gamma^nIter = T_init 
	double cdq = 0.001;
	double beta = 6.0;
	double n_max = 100;
	double gamma = Math.pow (cdq, 1.0/(double)n_max);

	public void handleEvent()
	{
		double Vmax = 5.0;
		double move_x = 0, move_y = 0;
				
		// T_init = - dE/ln(P) 
	    // allows an cost increase of dE with a prob of P  
	    // e.g., -0.446/ln(0.8) ~~ 2.0
	    // ln(0.8) = -0.2231, ln(0.7) = -0.3516, ln(0.6) = -0.5108, ln(0.5) = -0.6931
	    double temperature = 0.446/0.6931;

		double temp_u, best_u = Double.POSITIVE_INFINITY;
		Random generator = new Random();
		
		if (numnodes == 0) 
		{
			move(posx(),posy());

			postEvent(1.0);
			return;
		}
		
		int nn = 1;
		while(true)
        {
            // s_max(n) = 2*s / (1 + exp(beta*n/n_max))
            double range = 2*SimWorld2DMason.XMAX / (1 + Math.exp((beta*(double)nn)/(double)n_max));

    		velocity_x = generator.nextDouble() * 2 * range - range;
    		velocity_y = generator.nextDouble() * 2 * range - range;
        	            
    		// cut-down
    		double velocity = Math.sqrt(velocity_x * velocity_x + velocity_y * velocity_y);
    		if (velocity > Vmax)
    		{
    			velocity_x *= (Vmax/velocity);
    			velocity_y *= (Vmax/velocity);
    		}
    		
            temp_u = utility(velocity_x, velocity_y);
               
            if (Math.exp ((best_u - temp_u) / temperature) > Math.random ())
            //if (best_u > temp_u) 
            {
               	best_u = temp_u; 
               	move_x = velocity_x;
               	move_y = velocity_y;
            }
            
            if (nn > n_max)
        		break;
    		
            temperature *= gamma;
            nn++; 
        }
		
		// conditions for phase transition
		if (time() < Tpstart+Tpmin)
		{
			//System.out.println("stay phase due to MIN time of a phase " + time() + " started at " + Tpstart);
			;
		}
		else if (time() > Tpstart+Tpmax)
		{
			//System.out.println("next phase due to MAX time of a phase " + time() + " started at " + Tpstart);
			postPhase(phase+1);
		}
		else if (stable_formation(move_x, move_y))
		{
			//System.out.println("formation at time " + time());
			if (goal_reached(move_x, move_y))
			{
				Log.stdout.println("goal reached at time " + time());
				postPhase(phase+1);
			}
		}
		
		move(posx()+move_x,posy()+move_y);
		
	    postEvent(1.0);
	}

	public void handlePositionUpdate()
	{	
		log("new position: (" + posx() + "," + posy() + ")");

		postPosition();
	}

	void postPosition()
	{
		log("posting position: (" + posx() + "," + posy() + ")");

		PositionKnowledge kposition = new PositionKnowledge(this);
		kposition.posx = posx();
		kposition.posy = posy();
		postKnowledge(kposition);
	}
	
	void postPhase(int p)
	{
		Log.stdout.println("posting phase: " + p);
		
		PhaseKnowledge kphase = new PhaseKnowledge(this);
		kphase.phase = p;
		postKnowledge(kphase);
	}
	
	public void handleKnowledge(ApplicationKnowledge knew)
	{
		if(knew instanceof PositionKnowledge) handlePosition((PositionKnowledge)knew);
		if(knew instanceof PhaseKnowledge)    handlePhase((PhaseKnowledge)knew);
	}

	void handlePosition(PositionKnowledge knew)
	{
		log("handling position: (" + knew.posx + "," + knew.posy + ")" + " from " + knew.creator());

		String n = knew.creator();
		if(n != name())
		{
			if(!known(n))
			{
				log("position for new node");
				node[numnodes] = knew.creator();
				dx[numnodes] = 0.0;
				dy[numnodes] = 0.0; 
				posx[numnodes] = knew.posx;
				posy[numnodes] = knew.posy;
				numnodes++;
			}
			else
			{
				for(int i = 0; i < numnodes; i++)
				{
					if(knew.creator().equals(node[i]))
					{
						log("position for known node");
						dx[i] = posx[i] - knew.posx;
						dy[i] = posy[i] - knew.posy;
						posx[i] = knew.posx;
						posy[i] = knew.posy;
					}
				}
			}
		}
	}

	void handlePhase(PhaseKnowledge knew)
	{
		// phase update 
		phase = knew.phase;
		Tpstart = time();

		// flight plan 
		if (phase % 4 == 1)
		{
			goal_x = SimWorld2DMason.XMAX*0.2;
			goal_y = SimWorld2DMason.YMAX*0.2;
		}
		else if (phase % 4 == 2)
		{
			goal_x = SimWorld2DMason.XMAX*0.8;
			goal_y = SimWorld2DMason.YMAX*0.8;
		}
		else if (phase % 4 == 3)
		{
			goal_x = SimWorld2DMason.XMAX*0.8;
			goal_y = SimWorld2DMason.YMAX*0.2;
		}
		else 
		{
			goal_x = SimWorld2DMason.XMAX*0.2;
			goal_y = SimWorld2DMason.YMAX*0.8;
		}
	}
}