/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

// Simple non-object-oriented dispatching of events and knowledge
// Events are distinguished based on local state (no subclasses for events)
// MutexKnowledge as an inner class (local knowledge)

// optimistic distributed mutex
// each node has a notion of available resources
// but node knowledge can be inconsistent
// mutual exclusion becomes a probabilistic property

package cf.app.mutex1;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberNode;
import cf.util.Dist;

// not a stateless node
// local sate encapsulates program state (in critical section or not, encoded in events)

// knowledge ordering depends on creator and timestamp
// furthermore time is compared across nodes (need of synchronization)

class MutexApp extends CyberApp
{
    static double AcquireDelay = 0; 
    // static double AcquireDelay = 10; // less optimistic version

    int numnodes = 0;
    String node[] = new String[100];
    int request[] = new int[100];
    int requests = 0;

    boolean requesting = false;
    boolean acquiring = false;
    boolean critical = false;

    MutexApp(CyberNode node)
    {	
	super(node);
    }

    public void initialize()
    {
	postMutexKnowledge(0);
	log("initial request event");
	requesting = true;
	postEvent(100 + Dist.dist.sampleUniform(0.0,10.0));
    }

    public boolean known(String n)
    {
	for(int i = 0; i < numnodes; i++)
	    {
		if(node[i].equals(n)) return true;
	    }
	return false;
    }

    public void handleEvent()
    {
	     if(requesting) handleRequestEvent();
	else if(acquiring) handleAcquireEvent();
	else if(critical) handleReleaseEvent();
    }
    
    public void handleRequestEvent()
    {
	log("attempting to request resource");
	if(requests == 0)
	    {
		log("request granted");
		postMutexKnowledge(1);
		requesting = false;
		acquiring = true;
		postEvent(AcquireDelay);
	    }
	else
	    {
		log("resource already requested");
		requesting = true;
		postEvent(100);
	    }
    }

    public void handleAcquireEvent()
    {
	log("attempting to acquire resource");
	if(requests <= 1)
	    {
		log("resource granted");
		acquiring = false;
		critical = true;
		postEvent(50);
	    }
	else
	    {
		log("resource already requested");
		postMutexKnowledge(0);
		acquiring = false;
		requesting = true;
		postEvent(100);
	    }
    }

    public void handleReleaseEvent()
    {
	log("request release");
	critical = false;
	postMutexKnowledge(0);
	// postEvent(0);
    }

    class MutexKnowledge extends ApplicationKnowledge
    {
	int request;

	public MutexKnowledge(MutexApp app)
	{
	    super(app);
	}       
	
	public String logRepLong()
	{
	    return super.logRepLong() + " Request: " + request;
	}
    }

    public void handleKnowledge(ApplicationKnowledge k)
    {
	if(k instanceof MutexKnowledge) handleMutexKnowledge((MutexKnowledge)k);
    }
    
    void postMutexKnowledge(int r)
    {
	log("posting request = " + r);
	MutexKnowledge krequest = new MutexKnowledge(this);
	krequest.request = r;
	postKnowledge(krequest);
    }

    public void handleMutexKnowledge(MutexKnowledge knew)
    {
	log("handling request = " + knew.request);

	String n = knew.creator();

	if(!known(n))
	    {
		log("request for new node");
		node[numnodes] = knew.creator();
		request[numnodes] = knew.request;
		numnodes++;
	    }
	else
	    {
		for(int i = 0; i < numnodes; i++)
		    {
			if(knew.creator().equals(node[i]))
			    {
				log("request for known node");
				request[i] = knew.request;
			    }
		    }
	    }

	log("counting requests");

	requests = 0;	
	for(int i = 0; i < numnodes; i++)
	    {
		log("processing node " + node[i] + " request = " + request[i]);

		if(request[i] >= 0)
		    {
			requests += request[i];
		    }
	    }
	
	log("requests = " + requests);
    }

}
