/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/
// distributed optimization example
// minimization of a value function

package cf.app.optimizing;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.sim.SimEngine;
import cf.util.Dist;

// creator and timestamp is not used in any essential way
// ordering is only based on value

class SolutionKnowledge extends ApplicationKnowledge
{
    double sample = Double.NaN;
    double value = Double.NaN;

    SolutionKnowledge(CyberApplication creator)
    {
	super(creator);
    }

    public boolean related(ApplicationKnowledge k)
    {
	return this.getClass() == k.getClass();
    }

    public boolean subsumes(ApplicationKnowledge k)
    {
	return ((SolutionKnowledge)k).value >= value;
    }

    public String logRepLong()
    {
	return super.logRepLong() + 
	    " Sample: " + sample +
	    " Value: " + value;
    }
}

// computation is invoked periodically (SampleEvent)
// no other events are used

class OptimizingApp extends CyberApp
{
    double sample = Double.NaN;
    double value = Double.NaN;

    public OptimizingApp(CyberNode node)
    {
	super(node);
    }

    double value(double s)
    {
	return s*s;
    }

    public void initialize()
    {
	log("initialization");
	postEvent(10);
    }

    public void handleEvent()
    {
	double s = Dist.dist.sampleUniform(0.0,1.0);
	double v = value(s);

	log("sampling - value(" + s + ") = " + v);

	if(Double.isNaN(value) || v < value)
	    {
		sample = s;
		value = v;
		postSolution(s,v);
	    }

	postEvent(1);

    }

    void postSolution(double s, double v)
    {
	log("posting solution - value(" + s + ") = " + v);

	SolutionKnowledge ksolution = new SolutionKnowledge(this);
	ksolution.sample = s;
	ksolution.value = v;
        postKnowledge(ksolution);
    }

    public void handleKnowledge(ApplicationKnowledge knew)
    {
	if(knew instanceof SolutionKnowledge) handleSolution((SolutionKnowledge)knew);
    }
    
    void handleSolution(SolutionKnowledge knew)
    {
	log("handling solution - value(" + knew.sample + ") = " + knew.value + " from " + knew.creator());
	log("current  solution - value(" + sample + ") = " + value);
	
	if(knew.value < value)
	    {
		sample = knew.sample; value = knew.value;
	    }
	else
	    {
		log("ignoring solution - no improvement"); // should not happen
	    }
    }
}
