/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.pn.sensorrobots;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.api.devices.Area;
import cf.api.devices.Camera;
import cf.api.devices.CameraApplication;
import cf.api.devices.Image;
import cf.api.devices.MotionSensor;
import cf.api.devices.MotionSensorApplication;
import cf.api.devices.NoiseSensor;
import cf.api.devices.NoiseSensorApplication;
import cf.api.devices.PositionApplication;
import cf.api.devices.PositionMobility;
import cf.sim.pn.Place;
import cf.util.Log;

class PositionKnowledge extends ApplicationKnowledge
{
  String robot;
  Area area;
  PositionKnowledge(CyberApplication creator)
  {
    super(creator);
  }
    public boolean replaces(ApplicationKnowledge k)
    {
  return false;
    }
  public String logRepLong()
  {
    return super.logRepLong() + " Robot: " + robot + ", "+ area.logRepLong();
  }
}

class MotionKnowledge extends ApplicationKnowledge
{
    String robot;
    Area area;
    
    MotionKnowledge(CyberApplication creator)
    {
	super(creator);
    }

    public String logRepLong()
    {
	return super.logRepLong() + " Robot: " + robot;
    }
}

class NoiseKnowledge extends ApplicationKnowledge
{
    String robot;
    Area area;
    
    NoiseKnowledge(CyberApplication creator)
    {
	super(creator);
    }

    public String logRepLong()
    {
	return super.logRepLong() + " Robot: " + robot;
    }
}

class ImageKnowledge extends ApplicationKnowledge
{
    String robot;
    Area area;
    Image image;
    
    ImageKnowledge(CyberApplication creator)
    {
	super(creator);
    }

    public boolean replaces(ApplicationKnowledge k)
    {
  return false;
    }

    public String logRepLong()
    {
	return super.logRepLong() + " Robot: " + robot + " Image: " + image.logRepLong();
    }
}

class RootApp extends CyberApp
{
    public RootApp(CyberNode node)
    {
	super(node);
    }

    public void handleKnowledge(ApplicationKnowledge k)
    {
	if(k instanceof ImageKnowledge) handleImage((ImageKnowledge)k);
    }
    
    void handleImage(ImageKnowledge kimage)
    {
	log("handling image");
    }
}

class MotionSensorApp extends CyberApp implements MotionSensorApplication
{    
    MotionSensor sensor;

    public MotionSensorApp(CyberNode node, MotionSensor sensor)
    {
	super(node);
	this.sensor = sensor;
	sensor.addApplication(this);
    }

    public void handleMotion()
    {
	log("handling motion");
	MotionKnowledge kmotion = new MotionKnowledge(this);
	kmotion.robot = sensor.parentname();
	postKnowledge(kmotion);
    }
}

class NoiseSensorApp extends CyberApp implements NoiseSensorApplication
{
    NoiseSensor sensor;

    public NoiseSensorApp(CyberNode node, NoiseSensor sensor)
    {
	super(node);
	this.sensor = sensor;
	sensor.addApplication(this);
    }

    public void handleNoise()
    {
	log("handling noise");
	NoiseKnowledge knoise = new NoiseKnowledge(this);
	knoise.robot = sensor.parentname();
	postKnowledge(knoise);
    }
}

class CameraApp extends CyberApp implements CameraApplication
{
    Camera camera;

    Area area;

    public CameraApp(CyberNode node, Camera camera)
    {
	super(node);
	this.camera = camera;
	camera.addApplication(this);
    }

    public void handleKnowledge(ApplicationKnowledge knew)
    {
	if(knew instanceof MotionKnowledge) handleMotionKnowledge((MotionKnowledge)knew);
	else if(knew instanceof NoiseKnowledge) handleNoiseKnowledge((NoiseKnowledge)knew);
  else if(knew instanceof PositionKnowledge) handlePositionKnowledge((PositionKnowledge)knew);
    }

    void handleMotionKnowledge(MotionKnowledge kmotion)
    {	
	log("handling motion knowledge");

	if(kmotion.robot.equals(camera.parentname()) && kmotion.area == area)
	    {
		Image image = camera.takeImage(area);
		ImageKnowledge kimage = new ImageKnowledge(this);
		kimage.robot = camera.parentname();
		kimage.area = area;
		kimage.image = image;
		postKnowledge(kimage);
	    }
    }

    void handleNoiseKnowledge(NoiseKnowledge knoise)
    {	
	log("handling noise knowledge");

	if(knoise.robot.equals(camera.parentname()) && knoise.area == area)
	    {
		Image image = camera.takeImage(area);
		ImageKnowledge kimage = new ImageKnowledge(this);
		kimage.robot = camera.parentname();
		kimage.area = area;
		kimage.image = image;
		postKnowledge(kimage);
	    }
    }

    void handlePositionKnowledge(PositionKnowledge k)
    {
      log ("handling position knowledge");
      if(k.robot.equals(camera.parentname()) && k.area.name().equals( "pd" ) )
      {
		Image image = camera.takeImage(area);
		ImageKnowledge kimage = new ImageKnowledge(this);
		kimage.robot = camera.parentname();
		kimage.area = k.area;
		kimage.image = image;
		postKnowledge(kimage);
      }
    }

}

class PositionApp extends CyberApp implements PositionApplication
{
    PositionMobility position;

    public PositionApp(CyberNode node, PositionMobility position)
    {
	super(node);
	this.position = position;
	position.addApplication(this);
    }

    public void initialize()
    {
      Place togop = position.move(  "pb" );
    }

    public void handleEvent()
    {
    }

    boolean arrive = false;

    public void handlePositionUpdate()
    {
      if(arrive)
        retreat();
      else
        forward();
    }

    void forward()
    {
      Place p = position.getPlace();
      Log.stdout.println("current position : " + p.logRepShort() + ", time: " + time() ); 
      if(p.name().equals( "pb" ))
      {
        log("handling position pb");
        makePositionKnowledge(p.name());
        Place togop = position.move(  "pe" );
        if(togop !=null)
          Log.stdout.println("togo position : " + togop.logRepShort() ); 
      }
      if(p.name().equals("pe") )
      {
        log("handling position pe");
        makePositionKnowledge(p.name());
        Place togop = position.move(  "pd" );
        if(togop !=null)
          Log.stdout.println("togo position : " + togop.logRepShort() ); 
      }
      if(p.name().equals("pd") )
      {
        log("handling position pd");
        makePositionKnowledge(p.name());
        Place togop = position.move(  "pe" );
        if(togop !=null)
          Log.stdout.println("togo position : " + togop.logRepShort() ); 
        arrive = true;
        log("Arrive");
        Log.stdout.println("Arrive");
      }
    }

    void retreat()
    {
      Place p = position.getPlace();
      Log.stdout.println("current position : " + p.logRepShort() +", time: " + time() ); 
      if(p.name().equals( "pb" ))
      {
        log("handling position pb");
        makePositionKnowledge(p.name());
        Place togop = position.move( "pa" );
        if(togop !=null)
          Log.stdout.println("togo position : " + togop.logRepShort() ); 
      }
      if(p.name().equals("pe") )
      {
        log("handling position pe");
        makePositionKnowledge(p.name());
        Place togop = position.move(  "pb" );
        if(togop !=null)
          Log.stdout.println("togo position : " + togop.logRepShort() ); 
      }
    }

    void makePositionKnowledge(String p_name)
    {
        PositionKnowledge pk  = new PositionKnowledge( this );
        pk.robot = position.parentname();
        pk.area = new Area(p_name);
        postKnowledge(pk);
    }
    

}

class RobotApp extends CyberApp
{
    public RobotApp(CyberNode node)
    {
	super(node);
    }
    
}

//Dummy class to keep ant happy
class SensorApp extends CyberApp
{
 public SensorApp(CyberNode node)
 {
	super(node);
 }
}