/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/
// Simple robot scattering example, e.g. to extend coverage of an area
// Perform random work while number of close neigbors is sufficiently high
// Robots are assumed to be equipped with GPS.

package cf.app.scattering;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.api.devices.Position;
import cf.api.devices.PositionApplication;
import cf.util.Dist;

class PositionKnowledge extends ApplicationKnowledge
{
    double posx = Double.NaN;
    double posy = Double.NaN;

    PositionKnowledge(CyberApplication creator)
    {
	super(creator);
    }

    public String logRepLong()
    {
	return super.logRepLong() + " Position: " + "(" + posx + "," + posy + ")";
    }
}

class ScatteringApp extends CyberApp implements PositionApplication
{
    Position pos;

    int numnodes = 0;
    String node[] = new String[100];
    double posx[] = new double[100];
    double posy[] = new double[100];

    ScatteringApp(CyberNode node, Position pos)
    {
	super(node);
	this.pos = pos;
	pos.addApplication(this);
    }    

    public double posx()
    {
	return pos.posx();
    }

    public double posy()
    {
	return pos.posy();
    }

    public void move(double x, double y)
    {
	pos.move(x,y);
    }

    public boolean known(String r)
    {
	for(int i = 0; i < numnodes; i++)
	    {
		if(node[i].equals(r)) return true;
	    }
	return false;
    }

    public String logRepLong()
    {
	return super.logRepLong() + " Position: " + "(" + posx() + "," + posy() + ")" + " KnownNodes: " + numnodes + " Neighbors: " + countNeighbors();
    }

    double distance(int i)
    {
	double dx = posx() - posx[i];
	double dy = posy() - posy[i];
	return Math.sqrt(dx*dx+dy*dy);
    }

    int countNeighbors()
    {
	int c = 0;
	for(int i = 0; i < numnodes; i++)
	    {
		if(distance(i) < 10) c++;
	    }	
	return c;
    }

    public void initialize()
    {
	super.initialize();
	log("initialization");
	postPosition();
	postEvent(0.0);
    }

    public void handleEvent()
    {
	double dx = Dist.dist.sampleUniform(-1.0,1.0);
	double dy = Dist.dist.sampleUniform(-1.0,1.0);
	
	int closeneighbors = countNeighbors();

	log("close neighbors: " + closeneighbors);
	
	if(closeneighbors >= 3)
	    {
		move(posx()+dx,posy()+dy);
	    }

	if(time() < 1000) postEvent(10.0);
    }

    public void handlePositionUpdate()
    {	
	log("new position: (" + posx() + "," + posy() + ")");
	
	postPosition();
    }

    void postPosition()
    {
	log("posting position: (" + posx() + "," + posy() + ")");

	PositionKnowledge kposition = new PositionKnowledge(this);
	kposition.posx = posx();
	kposition.posy = posy();
	postKnowledge(kposition);
    }

    public void handleKnowledge(ApplicationKnowledge knew)
    {
	if(knew instanceof PositionKnowledge) handlePosition((PositionKnowledge)knew);
    }
    
    void handlePosition(PositionKnowledge knew)
    {
	log("handling position: (" + knew.posx + "," + knew.posy + ")" + " from " + knew.creator());

	String n = knew.creator();
	if(n != name())
	    {
		if(!known(n))
		    {
			log("position for new node");
			node[numnodes] = knew.creator();
			posx[numnodes] = knew.posx;
			posy[numnodes] = knew.posy;
			numnodes++;
		    }
		else
		    {
			for(int i = 0; i < numnodes; i++)
			    {
				if(knew.creator().equals(node[i]))
				    {
					log("position for known node");
					posx[i] = knew.posx;
					posy[i] = knew.posy;
				    }
			    }
		    }
	    }

    }
}
