/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.common;

import java.io.IOException;
import java.util.PriorityQueue;
import java.util.concurrent.ExecutorService;

import cf.sim.mason.SimWorldMason;
import cf.sim.stage.Event2StageCallback;
import cf.sim.stage.SimulationInterface;
import cf.util.Dist;
import cf.util.Log;

/** simple discrete event simulator using a single priority queue */
public abstract class Engine
{
  public static Thread myThread;

  public static Dist dist;
  public static Log log, rep;

  protected static double time;
  protected static PriorityQueue<Event> q;

  protected static long start_time = 0;

  public static ExecutorService cyberThreadExecutor;
  protected static int round = 0;

  protected static long myEngineId;

  public static SimWorldMason simworld;

  //PADO
  protected static int terminated_node = 0;

  public static synchronized void terminate()
  {
    terminated_node++;
  }
  public static int numTerminated()
  {
    return terminated_node;
  }
  //END of PADO
  
  public static long myEngineId()
    {
	return myEngineId;
    }

  public static void initialize()
    {
	myThread = Thread.currentThread();
	
	long seed = System.currentTimeMillis();
	Engine.dist = new Dist(seed);
	KnowledgeManager.dist = new Dist(seed);

	myEngineId = Engine.dist.rand().nextLong();
	if(myEngineId < 0) myEngineId *= -1;
	if(Param.RealMode || Param.ShowRounds) Log.stdout.println("engine id: " + myEngineId);
    }

  public static double time() // unit: second
  {
	if(Param.EnableMason)
	{
		double t = simworld.simstate.schedule.time();
		return t >= 0 ? t : 0;		
	}
	else if(Param.EnableStage)
    {
      return (double)(SimulationInterface.TimeNow()/1e6);
    }
    else if(Param.RealTime)
    {
      double delta = (double)(System.currentTimeMillis() - start_time)/1000.0*Param.SpeedUp;
      if(!Param.UseComputerTime) return delta;
      else return (double)(start_time/1000.0) + delta;
    }
    else
      return time;
  }
  
  public static void add(Event e)
  {
	if(Param.EnableMason)
	  {		
		log.println("add to mason q " + e.logRepLong());
		simworld.simstate.schedule.scheduleOnce(e.time(), e);		  
	  }
	else if(Param.EnableStage)
    {
      double t = e.time() - time(); // get offset time from now
      SimulationInterface.add_ap_callback(t*1e6/*us*/, new Event2StageCallback(e));
    }
    else
	{
	    if(Param.CyberThread || Param.UseRealNetwork)
		{
		    synchronized(q) {
			q.add(e);
		    }
		}
	    else
		{
		    q.add(e);
		}
	  	    
	    myThread.interrupt();
	}
  }

  abstract public void run(World world, boolean display);
  abstract public void run()
    throws IOException;

  public static void setEndTime(double time)
  {
    Param.MaxTime = time;
  }

  public static void setRound(int round)
  {
    Param.Rounds = round;
  }

  public static void fixTime(double difference) //s
  {
    start_time -= difference*1000;
    Param.MaxTime += difference;
    Log.stdout.println("--- time synchronization - adjustment: " + difference + " s ---");
    if(!Param.SILENT) log.println("--- time synchronization - adjustment: " + difference + " s ---");
  }

  static PriorityQueue<Event> getEventQueue()
  {
    return q;
  }

  protected static void showStatus(World n)
  {
    Log.stdout.println("--- Simulation Snapshot at Time " + Log.tf.format(time()) + " ---");
    // Log.stdout.println("Future Events Scheduled: " + q.size());			   
  }

  public static class TimeThread extends Thread
  {
    World world;
    boolean done = false;

    public TimeThread(World world)
    {
      this.world = world;
    }

    public void run()
    {
      while(!done) 
      {
        try { Thread.sleep(5000); }
        catch(InterruptedException e) { done = true; }						
        if(!done) showStatus(world);
      }
      world = null;
    }

    public void terminate()
    {
      done = true;
    }
  }
}
