/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.common;

import java.io.Serializable;

import cf.api.ApplicationKnowledge;
import cf.util.Dist;
import cf.util.Log;

/** Naive/dummy implementation of distributed information layer (all available
    knowledge is flooded through the network in regular intervals). */

public abstract class KnowledgeManager implements Serializable
{

    /** CONSTRUCTOR **/
    public KnowledgeManager(World world, Node node)
    {
        this.node = node;
        this.world = world;
        world.allKM().add(this);
        if(!Param.SharedKnowledgeBase)
            {
                kb  = new KnowledgeBase();
            }
    }

    /** MEMBER VARS **/
    public static Dist dist;
    public Node node;

    protected World world;
    protected KnowledgeBase kb;
    protected int directUpdates, indirectUpdates;

    /** MEMBER FUNCS **/
    public synchronized KnowledgeBase kb()  
    {
        if(Param.SharedKnowledgeBase)
            return world.kb_shared();
        else
            return kb; 
    }

    public String logRepShort()
    {
        if(Param.SILENT){ return ""; }
        return Log.getClassName(this) + "(" + node().name() + ")";
    }
    
    public String logRepLong()
    {
        if(Param.SILENT){ return ""; }
        return logRepShort();
    }

    public Node node()
    {
        return node.node();
    }

    public void log(int level, String msg)
    {
        if(Param.SILENT){ return; }

        node().log(level, logRepShort() + " > " + msg);
    }

    abstract public void initialize();
    abstract public void manageKnowledge(ApplicationKnowledge e);
    /* this function only called when indirect( coming from other node)  knowledge comes */
    abstract public  boolean handleKnowledgeUpdate(Node src, ApplicationKnowledge e);

    abstract public void print();

    public static void printAll(World n)
    {
        for (KnowledgeManager km : n.allKM()) {
            Engine.rep.println("");
            km.print();
        }
    }

    public static int avgKnowledgeItems(World n)
    {
        int r = 0;
        for (KnowledgeManager km : n.allKM()) r += km.kb().size();
        return r;
    }

    public static int totDirectUpdates(World n)
    {
        int r = 0;
        for (KnowledgeManager km : n.allKM()) r += km.directUpdates;
        return r;
    }

    public static int totIndirectUpdates(World n)
    {
        int r = 0;
        for (KnowledgeManager km : n.allKM()) r += km.indirectUpdates;
        return r;
    }

    public boolean requestUpdate(ApplicationKnowledge k)
    {
        log(2, "requesting direct knowledge update: "+ k.logRepShort());

        if( fresh(k) )
            {
                //		long start = System.nanoTime(); 
                update(k);
                //		long end = System.nanoTime(); 
                //		System.out.println("------ update takes " + (end-start)/1000);
                //		System.out.flush();
                directUpdates++;

                k.engineId = world.myEngineId(); // direct update case, mark engine ID
                //		start = System.nanoTime(); 
                manageKnowledge(k);
                //		end = System.nanoTime(); 
                //		System.out.println("------ manageKnowledge takes " + (end-start)/1000);
                return true;
            }
        return false;
    }

    /** Only fresh knowledge received will overwrite current knowledge. */
    protected void update(ApplicationKnowledge k)
    {
	String info = "";
        k.aware.set(node().index());

        if(Param.UseRealNetwork)
            {
                k.aware_engine().add(world.myEngineId());
                if(Param.LogAwareness) info = " - engine awareness set: " + k.aware_engine().toString();
            }

        kb().update(k);
    
        log(1, "update knowledge: " + k.logRepShort() + info);
    
        double expiration = k.expirationTime() - Engine.time();
        if(expiration != Double.POSITIVE_INFINITY) {
            node().postEvent(new KnowledgeExpiredEvent(k.expirationTime() - Engine.time(), this, k));
	}
    }

    protected void handleExpiredKnowledge(ApplicationKnowledge k)
    {
        kb().removeExpiredKnowledge(k);
        if(!Param.SILENT) log(1, "discard knowledge (expired): " + k.logRepShort());
    }

    protected void removeExpiredKnowledge()
    {
        kb().removeExpiredKnowledge();
    }

    protected boolean fresh(ApplicationKnowledge k)
    {
        return kb().fresh(k);
    }

    /** INNER CLASSES **/

    protected class KnowledgeExpiredEvent extends Event
    {
        KnowledgeManager km;
        ApplicationKnowledge k;
        KnowledgeExpiredEvent(double delay, KnowledgeManager km, ApplicationKnowledge k) {
            super(delay);
            this.km = km;
            this.k = k;
        }
        public void handle() {
            km.handleExpiredKnowledge(k);
	    node().handleKnowledgeExpired(k);
        }
    } 

}
