/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.common;

import java.util.LinkedList;
import java.util.PriorityQueue;

import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.real.RealNode;
import cf.util.Log;

abstract public class Model
{
    /** CONSTRUCTOR **/
    public Model(World world)
    {
        this.world = world;
        this.index = world.freshModelIndex();
        world.addModel(this);
    }

    public Model(World world, String name)
    {
        this(world);
        this.name = name;
    }

    public Model(World world, Model parent)
    {
        this(world);
        this.parent = parent;
    }

    public Model(World world, Model parent, String name)
    {
        this(world);
        this.name = name;
        this.parent = parent;
    }

    /** MEMBER VARS **/
    public World world; // world this node belongs to in the simulator

    protected PriorityQueue<Event> queue = new PriorityQueue<Event>();
    Thread myThread; // associated/waiting thread if any

    private int index; // globally unique index
    private String name; // globally unique name
    private Model parent; // parent if any
    private LinkedList<CyberApplication> allApplications = new LinkedList<CyberApplication>();
  
    //PADO
    protected boolean explicit_terminated = false;
    //END of PADO

    /** MEMBER FUNCS **/
    public PriorityQueue<Event> queue() 
    { 
        return queue; 
    }

    public int index() 
    {
        return index;
    }

    public String logRepLong()
    {
        return "";
    }

    public final String name()
    {
        if(name == null) 
            {
                return "N" + index;
            }
        else return name;
    }

    public final Model parent()
    {
        return parent;
    }

    public final String parentname()
    {
        return parent.name();
    }

    public final String logRepShort()
    {
        //if(Param.SILENT){ return ""; }
        return Log.getClassName(this) + "(" + name() + ")";
    }

    public final void log(String msg)
    {
    	if(Param.SILENT){ return; }
    	Engine.log.println(1, logRepShort() + " > " + msg);
    }
    
    public final void log(int level, String msg)
    {
        if(Param.SILENT){ return; }
        Engine.log.println(level, logRepShort() + " > " + msg);
    }

    public final double time()
    {
        return Engine.time();
    }

    public final void addApplication(CyberApplication app) 
    {
        allApplications.add(app);
        world.addApplication(app);
    }

    public final LinkedList<CyberApplication> allApplications()
    {
        return allApplications;
    }

    public void initialize()
    {
        //CHECK
        if( this instanceof Node) // prohibit calling application init from device
            {
                for(CyberApplication app : allApplications)
                    {
                        app.initialize();
                    }
            }
    }

    public final void postEvent(Event e)
    {
        log("posting event: " + e.logRepShort());
        if(Param.CyberThread && Param.MultiQ)
            {
                //synchronized(queue) {
                //      queue.add(e);
                //}
                //PADO
                if( !explicit_terminated )
                    {
                        synchronized(queue) {
                            queue.add(e);
                        }
                    }
                else
                    {
                        if(this instanceof RealNode)
                            {
                                RealNode n = (RealNode)this;
                                if(n.isNetworkNode())
                                    {
                                        synchronized(queue) {
                                            queue.add(e);
                                        }
                                    }
                            }
                    }
                //END of PADO
                if(myThread != null) myThread.interrupt();
            }
        else
            {
                Engine.add(e);
            }
    }

    public final void postEvent(double delay)
    {
        postEvent(new SimpleEvent(this,delay));
    }

    public void print()
    {
        Engine.rep.println("Model " + logRepShort());
        Engine.rep.println("State:");
        Engine.rep.begin();
        Engine.rep.println(logRepLong());
        Engine.rep.end();
    } 
  
    public static void printAll(World world)
    {
        for(Model node : world.allModels()) { Engine.rep.println(""); node.print(); }
    }

    public void setName(String name)
    {
        this.name = name;
    }

    /* Get related node(first if app has many nodes) from Cyber Devices */
    protected Node getNode()
    {
        Node node;
        for(CyberApplication app : allApplications())
            {
                if( app instanceof CyberApp)
                    {
                        CyberApp ca = (CyberApp) app;
                        if(ca.node instanceof Node)
                            {
                                node = (Node)ca.node;
                                return node;
                            }
                    }
            }
        return null;
    }

    protected void handleEvent()
    {
    }

    /** INNER CLASSES **/
    class SimpleEvent extends Event
    {
        Model model;

        SimpleEvent(Model model, double delay)
        { 
            super(delay);
            this.model = model;
        }

        public void handle()
        {
            model.handleEvent();
        }
    }
}
