/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.common;

import java.util.LinkedList;

import cf.api.ApplicationEvent;
import cf.api.ApplicationKnowledge;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.sim.SimEngine;
import cf.util.Dist;
import cf.util.Log;

public abstract class Node extends Model implements CyberNode
{
  private static final String TAG = Node.class.getSimpleName();

  /** CONSTRUCTOR **/
	
  public Node(World world)
  {
    super(world);
  }

  public Node(World world, String name)
  {
    super(world,name);
  }

  public Node(World world, Model parent)
  {
    super(world, parent);
  }
  public Node(World world, Model parent, String name)
  {
    super(world, parent, name);
  }

  /** MEMBER VARS **/
  
  public boolean threadAllocated = false;
  protected KnowledgeManager km;

  public LinkedList<ApplicationEvent> skip_event_list = new LinkedList<ApplicationEvent>();

  /** MEMBER FUNCS **/

  //PADO
  public boolean explicit_terminated()
  {
    return explicit_terminated;
  }
  //END of PADO
  
  public KnowledgeManager km() 
  { 
    return km; 
  }

  public boolean threadAllocated() 
  { 
    return threadAllocated; 
  }

  public Node node()
  {
    return this;
  }

  public void postKnowledge(ApplicationKnowledge k) // request knowledge update (and publish knowledge)
  {
    log("posting knowledge: " + k.logRepShort());

    if(Param.CyberThread)
    {
      //local update directly
      //km.requestUpdate(k); 
      synchronized (km.kb()) {
        if(!km.requestUpdate(k))
        {
          return;
        }
      }
      
      if(Param.SharedKnowledgeBase)
      {
        for(Node n : world.allNodes())
        {
          if(n != this) 
          {
            n.postEvent( new KnowledgeHandleEvent(0, n, k) ); // job will be done in local thread 
          }
        }
      }
      else if(!Param.UseSimNetwork) 
      {
        for(Node n : world.allNodes())
        {
          if(n != this) 
          {
            n.postEvent( new KnowledgeUpdateEvent(0, n, k) );// job will be done in local thread 
          }
        }
      }
    }
    else
    {
      // Local Update 
      //km.requestUpdate(k); 
      if(!km.requestUpdate(k))
      {
        return ;
      }

      if(Param.UseSimBroadcast)
      {
    		for(Node n : world.allNodes())
    		    {
    			if(n != this) 
    				{
    				double delay = Param.AvgDelay;   	 				
    				if(Param.ExpDelay) delay = Dist.dist.sampleExpWithMean(Param.AvgDelay);
    				else if(Param.UniformDelay) delay = Dist.dist.sampleUniform(0,2*Param.AvgDelay);
    				n.postEvent(new KnowledgeUpdateEvent(delay, n, k));
    				}
    		    }
      }
      // handle Knowledge on each Cyber node directly, when SharedKnowledge Base 
      else if(Param.SharedKnowledgeBase)
      {
        for(Node n : world.allNodes())
        {
          if(n != this) 
          {
            n.postEvent( new KnowledgeHandleEvent(0, n, k) ); // job will be done in local thread 
          }
        }
      }
      // No link, update directly  to every node in world
      else if(!Param.UseSimNetwork) 
      {
        for(Node n : world.allNodes())
        {
          if(n != this) {
            n.postEvent( new KnowledgeUpdateEvent(0, n ,k) );// job will be done in local thread 
          }
        }
      }
    }
  }

  public void postEvent(ApplicationEvent e)
  {
    e.node = this;
    postEvent((Event)e);
  }


  public void print()
  {
    Engine.rep.println("CyberNode " + logRepShort());

    Engine.rep.begin();

    Engine.rep.println("Local State:");
    Engine.rep.begin();
    Engine.rep.println(logRepShort());
    Engine.rep.end();
    //    Engine.rep.println("Local Knowledge:");
    //    Engine.rep.begin();
    //    km.kb.print();
    //    Engine.rep.end();
    Engine.rep.println("Applications:");
    Engine.rep.begin();

    for(CyberApplication app : allApplications())
    {
      if(app instanceof CyberApplication) 
      {
        Engine.rep.println("Application " + app.name());
        Engine.rep.begin();
        Engine.rep.println(app.logRepShort());
        Engine.rep.end();
      }
    }
    Engine.rep.end();
    Engine.rep.println("KnowledgeManager:");
    Engine.rep.begin();
    km.print();
    Engine.rep.end();
    Engine.rep.end();
  }

  public static void printAll(World world)
  {
	for(Node node : world.allNodes()) { SimEngine.rep.println(""); node.print(); }
  }

  /** INNER CLASSES **/

  class KnowledgeReceivedEvent extends Event
  {
	Node node;
	ApplicationKnowledge k;
	
	KnowledgeReceivedEvent(Node node, double delay, ApplicationKnowledge k)
	{ 
	    super(delay);
	    this.node = node;
	    this.k = k;
	}
	
	public void handle()
	{
	    node.handleKnowledgeReceivedEvent(this);
	}
  }

    /*    
  class KnowledgeExpiredEvent extends Event
  {
      Node node;
      ApplicationKnowledge k;

      KnowledgeExpiredEvent(Node node, double delay, ApplicationKnowledge k){ 
	  super(delay);
	  this.node = node;
	  this.k = k;
      }
        
      public void handle()
      {
	  km.handleExpiredKnowledge(k);
	  node.handleKnowledgeExpired(k);
      }
   }
    */


  void handleKnowledgeReceivedEvent(KnowledgeReceivedEvent e)
  {
    ApplicationKnowledge k = e.k;

    log("handling knowledge: " + k.logRepShort());

    if(!k.expired())
    {
      //ApplicationKnowledge kold = km.kb().subsumed(k);
      ApplicationKnowledge kold = km.kb().ignore(k);
      if(kold == null) 
      {
        km.kb().update(k);

        for(CyberApplication app : allApplications())
        {
          if(app instanceof CyberApplication) 
          {
            k.handle((CyberApplication)app);
          }
        }
      }
      else
      {
        log("ignoring obsolete knowledge - subsumed by: " + kold.toString());
      }
    }
    else
    {
      log("ignoring expired knowledge: " + k.toString());
    }
  }

  //Maneesh's handleExpiredKnowledge
  public void handleKnowledgeExpired(ApplicationKnowledge k) { 
      for(CyberApplication app : allApplications())
      {
	  k.handleKnowledgeExpired((CyberApplication)app);
      }
  }

  public class KnowledgeHandleEvent extends ApplicationEvent
  {
    Node node;
    ApplicationKnowledge k;
    public KnowledgeHandleEvent(double delay, Node node, ApplicationKnowledge k)
    {
      super(null, delay); // non-creator application event
      this.node = node;
      this.k = k;
    }
    public void handle()
    {
      node.km().manageKnowledge(k);
    }
  }

  public class KnowledgeUpdateEvent extends ApplicationEvent
  {
    Node node;
    ApplicationKnowledge k;
    public KnowledgeUpdateEvent(double delay, Node node, ApplicationKnowledge k)
    {
      super(null,delay); // non-creator application event
      this.node = node;
      this.k = k;
    }
    public void handle()
    {
      node.km().handleKnowledgeUpdate(node, k);
    }
  }
}
