/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.common;

/* 

   The cyber-application framework supports two modes of operation 
   depending on which top-level class is used to set up the world. 

   We say that the cyber-framework is running 
   in *** simulation mode *** if SimWorld is used, and 
   in *** real mode *** if the RealWorld class is used.

*/

public class Param
{       
    /* hprof: for disabiling logging (easy) and the accompanying string appending (painful) */
    public static final boolean SILENT = false;
    
    /* if LoggingDirectory exists , logs are stored under LoggingDirectory. 
       Otherwise logs are stored under java.io.tmpdir  */
    public static final String LoggingDirectory = "logging";
    
    /* use short names for logs like "trace.log" and "report.log" to overwrite logs */
    public static final boolean OverwriteLog = false;

    /* use framework in simulation mode (false) or real mode (true) */
    public static boolean RealMode = false;
    
    /* use Mason as simulation engine (simulation mode only) */
    public static boolean EnableMason = false; 

    /* number of simulation rounds (simulation mode) */
    public static int Rounds = 1;

    /* display information for each round (simulation mode) */
    public static boolean ShowRounds = true;

    /* time limit of simulation/execution (can be Double.POSITIVE_INFINITY) */
    public static double MaxTime = 100000000;  // in sec from start time

    /* seeds for random number generators (0 = derive from system time) */
    public static long SimulationSeed = 1; 
    
    /* use GUI if available, e.g. if used together with Mason */
    public static boolean GUI = false; 
    
    /* verbosity of traces (0-3) - traces are written to logging/trace-n.log */
    public static int Verbosity = 3;
    
    /* log knowledge awareness sets if logging enabled */
    public static boolean LogAwareness = false;
    
    /* level of detail for reports (0-2) - reports are written to logging/report-n.log */
    public static int Report = 0; 
    
    /* processing of events takes place in or simulation time (false) or real time (true) 
       (will be set automatically depending on use of SimWorld/RealWorld class) */
    public static boolean RealTime = true; 

    /* max time to sleep before checking for next event (real mode only) */
    public static int MaxSleepTime = 100; // msec

    /* time is determined by offset from startup time (false) or 
       by computer time in seconds relative to midnight, January 1, 1970, UTC (true) */
    public static boolean UseComputerTime = false;

    /* simulation or execution is accelerated by this factor */
    public static int SpeedUp = 1;  

    /* parameters for real mode */
    
    /* enable logical time synchronization */
    public static boolean LogicalTimeSync = true;
    
    /* synchronize only if error larger than this */
    public static double MaxTimeError = 0.0; //s
    
    /* concurrent execution of cyber-applications (currently only supported in real mode) */
    public static boolean CyberThread = true;
    
    /* multiple event queues with one thread per cyber-node (true), or
       single event queue with fixed size thread pool (false) */
    public static boolean MultiQ = true;  
    
    /* number of threads in the pool for the thread pool version */
    public static int ThreadPoolSize = 10;
    
    /* enable knowledge dissemination through real network (real mode only) */
    public static boolean UseRealNetwork = true; 
    
    /* port number shared by all cyber-engines running on a single cyber-host */
    public static int SharedPort = 7777;  
      
    /* on a given cyber-host each cyber-engine selects a unique local port in this range */
    public static int LocalPortMin = 7000;  
    public static int LocalPortMax = 7010;
      
    /* period of hello broadcast message */
    public static int HelloBroadcastDelay = 1;//s
      
    /* life time of neighbor entry (if not refreshed) */
    public static int NeighborExpirationTime = 5;//s
      
    /* support for multi-hop neighbor discovery (if a cyber-engine receives hello message from another 
       subnet then it will be forwarded it to its neighbors if hops have not reached this limit */
    public static int MultiHopDiscovery = 2; 

    /* enable broadcast for neighbor discovery (subnet broadcast is default, e.g. via X.Y.Z.255) */
    public static boolean EnableBroadcast = true;
    /* use unrestricted broadcast via 255.255.255.255 instead of subnet broadcast (may not work on all systems) */
    public static boolean EnableUnrestrictedBroadcast = false;

    /* enable broadcast for knowledge dissemination (requires EnableBroadcast = true to enable physical broadcast) */
    public static boolean BroadcastKnowledge = true;

    /* additional neighbors can be specified explicitly (serving as discovery facilitators) */
    public static String UserDefinedNeighbors[] = { "" }; //{ "anode.csl.sri.com" };

    /* enable knowledge dissemination through simulated broadcast (simulation mode only) */
    public static boolean UseSimBroadcast = false;

    /* delay distribution for simulated broadcast */
    public static double AvgDelay = 1.0;
    public static boolean ExpDelay = true;
    public static boolean UniformDelay = false;
    
    /* enable knowledge dissemination through simulated network (simulation mode only) */
    public static boolean UseSimNetwork = true; 

    /* upper bound for simulated network size */
    public static int MaxNodes = 300;

    /* upper bound for number of bundles */
    public static int MaxBundles = 10000;
        
    /* default link characteristics */

    public static double LinkCost = 0.0;
    public static double LinkBandwidth = 2000000;
    public static double LinkLatency = 1.0;
    public static double LinkErrorRate = 0.0;

    /* simulated links can be dynamic, i.e. linkstate (up/down) changes over time */
    public static boolean DynamicLinks = false;

    /* dynamic links use average link up/down times that are randomly 
       selected from these intervals upon link creation with a uniform distribution */

    public static double AvgLinkUpTimeMin = 1;
    public static double AvgLinkUpTimeMax = 50;
    public static double AvgLinkDownTimeMin = 1;
    public static double AvgLinkDownTimeMax = 50;

    /* up/down times can be deterministic (true) or neg. exponential distributed (false) */
    public static boolean DeterministicLinks = false;

    /* link sensing introduces a delay between physical link state (up/down) and sensed link state */
    public static boolean LinkSensing = true;

    /* the delay is randomly selected from the following interval with a uniform distribution */
    public static double LinkUpSenseDelay = 0.0;
    public static double LinkDownSenseDelay = 0.0;

    /* bound the number of transmissions queued on a link (drop transmission if exceeded) */
    public static int LinkQueueLength = Integer.MAX_VALUE;

    /* abstract mobility model based on annotated Petri nets (see readme_simmobility for details) */ 

    /* delay to move from one place to another */
    public static int MobilityTransitionDelay = 30; //s

    /* parameters for knowledge dissemination in simulated network (simulation mode only) */

    /* keep track of bundles for diagnosis */
    public static boolean KeepBundles = false;

    /* keep track of data bundles only (not used) */
    public static boolean KeepDataBundlesOnly = true; 

    /* resources available to a bundle (currently not used) */
    public static double BundleResources = Double.POSITIVE_INFINITY;

    /* average header overhead in bytes (approximate) */
    public static double BundleHeaderSize = 100;

    /* average size of a unit of knowledge (approximate) */
    public static int AvgKnowledgeItemSize = 100;

    /* consider link busy if queue length reaches this number */
    public static int BusyQueueLength = Integer.MAX_VALUE;

    /* discard bundle upon expiration */
    public static boolean DiscardExpiredBundles = true;
  
    /* discard duplicate bundles */
    public static boolean DiscardIncomingDuplicate = true;

    /* randomized waiting for data bundles (not used) */
    public static double MinBundleWaitingTime = 0;
    public static double MaxBundleWaitingTime = 0;

    /* max number of hops for data bundles (not used) */
    public static int StopAfterHops = Integer.MAX_VALUE;

    /* retransmission delay for data bundles (not used) */
    public static double RetransmissionDelay = Double.POSITIVE_INFINITY;

    /* expiration time for status bundles */
    public static double StatusBundleExpiration = Double.POSITIVE_INFINITY;

    /* maintaining network-related knowledge (not used) */

    public static boolean MaintainNodeKnowledge = false;
        
    public static boolean MaintainNodeStatusKnowledge = false;
    public static boolean RefreshNodeStatusKnowledge = false;
    public static double NodeStatusRefreshInterval = 100.0;

    public static boolean MaintainNodeStatisticsKnowledge = false;
    public static boolean RefreshNodeStatisticsKnowledge = false;
    public static double NodeStatisticsRefreshInterval = 1000.0;

    public static boolean MaintainLinkKnowledge = false;
        
    public static boolean MaintainLinkStatusKnowledge = false;
    public static boolean RefreshLinkStatusKnowledge = false;
    public static double LinkStatusRefreshInterval = 100.0;

    public static boolean MaintainLinkStatisticsKnowledge = false;
    public static boolean RefreshLinkStatisticsKnowledge = false;
    public static double LinkStatisticsRefreshInterval = 1000;

    public static double BundleStatusDisseminationIntervalMin = 100.0;
    public static double BundleStatusDisseminationIntervalMax = 100.0;

    /* parameters for knowledge dissemination in real and simulated network (for real and simulation mode) */

    /* period for knowledge dissemination (uniformly random in this range) */
    public static double KnowledgeDisseminationIntervalMin = 1.0;//s
    public static double KnowledgeDisseminationIntervalMax = 2.0;//s

    /* bound amount of knowledge per bundle */
    public static int MaxKnowledgeItemsPerBundle = 1000;

    /* dissemination policies for knowledge dissemination (optimized flooding is default) */
    public static enum DisseminationPolicyEnum {
        FLOOD,			// Deterministic flooding only
        PROB_REF,		// Deterministic flooding + Coarse-Grained Probabilistic Reflection
        PROB_REF_FINE	// Deterministic flooding + Fine-Grained Probabilistic Reflection
    }
    
    /* probability for Coarse-Grained Probabilistic Reflection */
    public static double ReflectionParameter = 0.6;

    public static DisseminationPolicyEnum DisseminationPolicy = DisseminationPolicyEnum.PROB_REF;

    /* type of knowledge base used for cyber-nodes in same cyber-engine (no knowledge base is a possible option) */
    public static boolean SharedKnowledgeBase = false; // knowledge base shared between cyber-threads

    /* use Stage multi-robot simulator for the network (simulation mode only) */
    public static boolean EnableStage = false; 

    /* use Stage Wifi network model for simulated links */
    public static boolean UseStageLinks = true; 

    /* enable transmission through Stage Wifi Network model 
       (currently messages are transmitted by reference only) */
    public static boolean UseStageTransfer = false; 

    /* parameters for cyber-ping (simple diagnosis tool) ) */

    public static double PingKnowledgeExpireTime = 20;//s
    public static double PingResponseKnowledgeExpireTime =  20;//s
    public static boolean PingUrgentResponse = true;
   
    /* parameters for PADO */
    public static long EngineIdSeed = 0;
}
