/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.real;

import java.io.File;
import java.io.IOException;
import java.util.PriorityQueue;

import cf.api.ApplicationEvent;
import cf.common.CyberThreadForPool;
import cf.common.Engine;
import cf.common.Event;
import cf.common.Model;
import cf.common.Node;
import cf.common.Param;
import cf.common.World;
import cf.common.Engine.TimeThread;
import cf.util.Log;

/** simple discrete event simulator using a single priority queue */
public abstract class RealEngine extends Engine
{
  abstract public RealWorld createRealWorld();
  public static void initialize() 
  {
    Engine.initialize();

    time = 0;
    q = new PriorityQueue<Event>();

    if(Param.EnableStage)
    {
      Log.stderr.println("EnableStage should be off");
      System.exit(1);
    }

    Param.RealTime = true;

    Param.SharedKnowledgeBase = true; // FIXME: if Distributed Simulation mode , this option should be selectable

    Param.Rounds = 1;

    if(Param.SharedKnowledgeBase)
      Param.UseSimNetwork = false;
    Param.UseSimNetwork = false;

    if(Param.UseComputerTime)
    {
       Param.MaxTime += ((double)System.currentTimeMillis())/1000.0;
    }
  }

  public void run(World world, boolean display)
  {
    int count = 0;
    TimeThread tt = null;
    //LinkedList<ApplicationEvent> skip_event_list = new LinkedList<ApplicationEvent>();

    //PADO
    int node_num = world.allNodes().size();
    //END of PADO
    
    if(display)
    {
      // tt = new TimeThread(world);
      // tt.start();
    }
    try
    {
      while(true)
      {		
        if(time() > Param.MaxTime)
        {
          Log.stdout.println("Events processed: " + count + " - maximum time reached");
          break;
        }
        
        //PADO
        if(terminated_node == node_num)
        {
          Log.stdout.println("All node terminated ");
          break;
        }
        //END of PADO

        Event e;
        if(Param.CyberThread || Param.UseRealNetwork)
        {
          synchronized(q) {
            e = q.peek();
          }
        }
        else
        {
          e = q.peek();
        }

        if(e == null) 
        {
	    // Log.stdout.println("Events processed: " + count + " - no further events");
          continue;
        }

        int maxsleep = (int)(1000.0 * (e.time() - time()));
        if(maxsleep > 0)
	    {
	        if(maxsleep > Param.MaxSleepTime) maxsleep = Param.MaxSleepTime;
	        try { Thread.sleep(maxsleep); } catch(InterruptedException ex) {}
	    }

        while( time() >= e.time() )
        {
          //PADO
          if(terminated_node == node_num)
          {
              Log.stdout.println("All node terminated ");
              break;
          }
          //END of PADO
          
          if(!e.cancelled())
          {
	        if(!Param.SILENT) RealEngine.log.println(1, "--- Real Time: " + Log.tf.format(time()) + " --- Event: " + e.logRepShort());

            if(Param.CyberThread && !Param.MultiQ)
            {
              if(e instanceof ApplicationEvent)
              {
                ApplicationEvent ae = (ApplicationEvent)e;
                Node node = (Node)(ae.node);
                synchronized(node)
                {
                  if(!node.threadAllocated())
                  {
                    node.threadAllocated = true;
                    if(ae.application()!=null) {
			if(!Param.SILENT) RealEngine.log.println(1, "node: " + node.name() + " event: "  + ae);
		    }
                    cyberThreadExecutor.execute(new CyberThreadForPool(ae));
                  }
                  else
                  {
                    node.skip_event_list.add(ae);
                  }
                }
              }
              else
              {
                e.trigger(); // not Application event, world's event.
              }
            }
            else
            {
              e.trigger();
            }
          }
          count++;

          if(Param.CyberThread || Param.UseRealNetwork)
          {
            synchronized(q) {
              //q.poll();
              q.remove(e);
              e = q.peek();
            }
          }
          else
          {
            q.remove(e);
            e = q.peek();
          }

          if(e == null) 
            break;
        }
        //if(Param.CyberThread && !Param.MultiQ && skip_event_list.size() > 0)
        //{
        //  synchronized(q)
        //  {
        //    q.addAll(skip_event_list);
        //  }
        //  skip_event_list.clear();
        //}
      }
    }
    finally
    {
      if(tt != null) 
      {
        tt.terminate();
      }

    }
  }

  public void run()
    throws IOException
    {

      long StartTime = System.currentTimeMillis();

      Log.stdout.println("running in real mode ...");

      initialize();
      start_time = System.currentTimeMillis();
      
      RealWorld world = createRealWorld();
      beginLogging(world);      
      world.initialize();
            
      run(world,true);
      
      rep.println("final world state");
      world.close();
      Model.printAll(world);
      
      endLogging();

      long EndTime = System.currentTimeMillis();

      Log.stdout.println("total computation time: " + Log.tf.format((double)(EndTime-StartTime)/1000.0));

      Log.stdout.println("done");
      Log.stdout.println("");
    }

  public static void beginLogging(RealWorld world)
    throws IOException
    {
	  String logHome = Param.LoggingDirectory;
	  
	  if(logHome.equals("") || !(new File(logHome).isDirectory())){
	  	  logHome = System.getProperty("java.io.tmpdir");
	  	  
	  	  File logHomeDir = new File(logHome);
	  	  if(!logHomeDir.isDirectory())
	  		  logHomeDir.mkdirs();
	  	Log.stdout.println("\""+Param.LoggingDirectory+"\" doesn't exist. Logs are stored under \"" + logHome+"\"");
	    }
	  
	  // use short name for log files to overwrite logs
	  if(Param.OverwriteLog){
		  RealEngine.log = new Log(logHome + File.separator + "trace.log");
	      RealEngine.rep = new Log(logHome + File.separator + "report.log");
	  }
	  
	  else{
		  RealEngine.log = new Log(logHome + File.separator + "trace-" + myEngineId() + ".log");
	      RealEngine.rep = new Log(logHome + File.separator + "report-" + myEngineId() + ".log");
	  }
    }

  public static void endLogging()
    throws IOException
    {
	if(!Param.SILENT) RealEngine.log.close();		
      round++;
    }
}
