/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.real;

import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Iterator;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApplication;
import cf.common.Engine;
import cf.common.Event;
import cf.common.KnowledgeBase;
import cf.common.KnowledgeManager;
import cf.common.Node;
import cf.common.Param;
import cf.util.Log;

/** Naive/dummy implementation of distributed information layer (all available
  knowledge is flooded through the network in regular intervals). */

public class RealKnowledgeManager extends KnowledgeManager  implements Serializable
{
	private static final long serialVersionUID = 1L;
	
	// Size for spliting KnowledgeBase to transmit. 
	// Maximum udp payload size is 65528 bytes (2^16(total datagram size) - 8(header size))
	public static final int PAYLOAD_SIZ = 65000;	//65000 bytes	

  /** CONSTRUCTOR **/
  public RealKnowledgeManager(RealWorld world, RealNode node)
  {
    super(world, node);
    this.node = node;
    this.world = world;
  }

  /** MEMBER VARS **/
  public RealNode node;
  private RealWorld world;

  /** MEMBER FUNCS **/
  public void initialize()
  {
    if(Param.UseRealNetwork &&  world.allNetworkNodes().contains( node ) ) 
    {
      double delay = dist.sampleUniform(Param.KnowledgeDisseminationIntervalMin, Param.KnowledgeDisseminationIntervalMax);
      node.postEvent(  new UDPKnowledgeTransmitEvent(delay, node() ) );
    }
  }

  public RealNode node()
  {
    return node.node();
  }

  public void manageKnowledge(ApplicationKnowledge e)
  {
    /* case: common ApplicationKnowledge */
    for(CyberApplication app : node().allApplications())
    {
      if(app instanceof CyberApplication) 
      {
        e.handle((CyberApplication)app);
      }
    }

    if(Param.UseRealNetwork)
    {
      if( e instanceof PingKnowledge || e instanceof PingResponseKnowledge)
      {
        if(Param.PingUrgentResponse)
          UDPKnowledgeTransmit(false); // send immediately
      }
    }
  }

  /* this function only called when indirect( coming from other node)  knowledge comes */
  public  boolean handleKnowledgeUpdate(Node src, ApplicationKnowledge e)
  {
    if (!e.expired())
    {
      if (fresh(e))
      {
	    log(1, "received fresh knowledge: "+ e.logRepShort());
	    //System.out.println("received fresh knowledge: "+ e.logRepShort());
        e = e.copy();
        //e.receivedfrom.set(src.index());

        if (!e.aware.get(node().index())) e.aware.set(node().index());
        update(e);
        indirectUpdates++;

        manageKnowledge(e);

        /* handle fresh PingKnowledge */
        if (e instanceof PingKnowledge) 
        {
          if(Param.UseRealNetwork && world.allNetworkNodes().contains(node()) )
            handlePingKnowledge((PingKnowledge)e);
        }
        return true;
      }
      else
      {
	    log(1, "received known knowledge: " + e.logRepShort() );
        synchronized(kb()) {
          for (ApplicationKnowledge k : kb())
          {
            //if (e.equal(k) && e.related(k) )
            if (e.related(k) && (k.replaces(e) || k.subsumes(e)))
            {
              //if (!e.aware.get(node().index())) k.receivedfrom.set(src.index());
              k.aware.or(e.aware);
              if(Param.LogAwareness) log(2, "node awareness set updated: "  + k.aware.toString());
              k.aware_engine().addAll(e.aware_engine());
              if(Param.UseRealNetwork && Param.LogAwareness) log(2, "engine awareness set updated: "  + k.aware_engine().toString());
              //k.processedfor = new BitSet();
              
            }
          }
        }
        return false;
      }
    }
    return false;
  }

  public void UDPKnowledgeTransmit(boolean periodic) 
  {
    if( !world.allNetworkNodes().contains(node()) )
    {
      return;
    }
    try
    {
      DatagramSocket socket = world.socket_local();
      if(!Param.BroadcastKnowledge) 
      {
        //System.out.println(logRepShort()+ " Start Sending UDP Knowledge " + k.logRepShort() );

        //NEIGHBOR ITER 
        //Iterator<NeighborInfo> neighbor_iter = world.neighbors().iterator();
        //while(neighbor_iter.hasNext())
        //{
        //  NeighborInfo neighbor_info = neighbor_iter.next();
        //PADO 
    	NeighborList temp_neighbor = (NeighborList)(world.neighbors().clone());
        for(NeighborInfo neighbor_info : temp_neighbor)
        {
          KnowledgeBase totransmit = new KnowledgeBase();
        //END of PADO
          
          //KNOWLEDGE BASE ITER
          // Random Pick up 
          int kb_size = kb().size();
          int count = 0;
          int n_sequence[] = randomSequence(kb_size);
          while(count < kb_size)
          {
            ApplicationKnowledge k = kb().get(n_sequence[count]);
            count ++;

            //System.out.println(logRepShort() + " check Knowledge " + k.logRepShort() +"'s sentto ip: " + k.sentto_engine().toString() + "'s aware ip:  " + k.aware_engine().toString() + " for neighbor : "+neighbor_info.toString());

            if(!k.aware_engine().contains(neighbor_info.CyberEngineId()) && !k.sentto_engine().contains(neighbor_info.CyberEngineId()))
            {
              log(2, "sending knowledge " + k.logRepShort() + " to " + neighbor_info.toString());
              k.sentto_engine().add(neighbor_info.CyberEngineId());

              totransmit.add(k);
            }
            
            
            else if( Param.DisseminationPolicy == Param.DisseminationPolicyEnum.PROB_REF ||
            		Param.DisseminationPolicy == Param.DisseminationPolicyEnum.PROB_REF_FINE) 
            {
         	  // JS -------
              if(!k.aware_engine().contains(neighbor_info.CyberEngineId())){
          	    log(2, "sending knowledge " + k.logRepShort() + " to " + neighbor_info.toString());
          	    k.sentto_engine().add(neighbor_info.CyberEngineId());
                totransmit.add(k);
              }
              // ------- JS
              else if( k.aware_engine().contains(neighbor_info.CyberEngineId()) && !k.sentto_engine().contains(neighbor_info.CyberEngineId()) )
              {
            	// JS: Support Maneesh's Probabilistic Reflection -------
            	double r = Param.ReflectionParameter;
            	if(Param.DisseminationPolicyEnum.PROB_REF_FINE == Param.DisseminationPolicy)
            	  r = k.transmissionProbability;
            	// ------- JS
            		
                double n = world.neighbors().size();
                //double r = Param.ReflectionParameter;
                if( dist.sampleUniform(0, n) < r )
                {
   		          log(2, "sending knowledge " + k.logRepShort() + " to " + neighbor_info.toString() + " - probabilistic reflection ["+r/n+"]");
                  k.sentto_engine().add(neighbor_info.CyberEngineId());
                  totransmit.add(k);
                }
              }
            }

            if(count == kb_size && totransmit.size() > 0 || totransmit.size() >= Param.MaxKnowledgeItemsPerBundle)
            {
              // Application knowledge to byte array
              //byte [] data = kb2byte(totransmit);									// JS : comment out
            	// JS : Split totransmit into knowledge bases if it's over the PAYLOAD_SIZ -------
            	ArrayList<KnowledgeBase> kbs = splitKb(totransmit, PAYLOAD_SIZ);
            	Iterator kbs_it = kbs.iterator();
            	while(kbs_it.hasNext()){
            		byte[] data = kb2byte((KnowledgeBase)kbs_it.next());
            	// -------- JS
	              /* iteration of IPList in Neighbor Info */
	              Iterator<IPInfo> ip_iter = neighbor_info.ipList().iterator();
	              while(ip_iter.hasNext())
	              {
	                IPInfo n_ip = ip_iter.next();
	                
	                DatagramPacket o_packet = new DatagramPacket(data, data.length, n_ip.ip(), n_ip.port());
	                try {
			          log(2, "sending UDP datagram to " + n_ip.ip().getHostAddress() + ":" + n_ip.port() + " at time " + Engine.time());
	                  socket.send(o_packet);
	                } catch(Exception e) { /* do nothing, it's ok not possible to send */ }
	              }
            	}	// JS

              totransmit = new KnowledgeBase();
              //totransmit.clear();
            }
          }
          //System.out.println(logRepShort() + " Knowledge " + k.logRepShort()+ " try to send to " + neighbor_ip.toString()+ "'s aware ip:  " + k.aware_engine().toString() );
        }
      }

      else if(Param.BroadcastKnowledge)
      {
        KnowledgeBase totransmit = new KnowledgeBase();

        //KNOWLEDGE BASE ITER
        // Random Pick up 
        int kb_size = kb().size();
        int count = 0;
        int n_sequence[] = randomSequence(kb_size);
        while(count < kb_size)
        {

          ApplicationKnowledge k = kb().get(n_sequence[count]);
          count ++;

          if( !k.aware_engine().containsAll( world.neighbors() ) &&
              ! k.sentto_engine().containsAll( world.neighbors() ) )
          {
            log(2, "sending knowledge " + k.logRepShort() + " via broadcast");
            k.sentto_engine().addAll(world.neighbors() );
            totransmit.add(k);
          }
          //else if( Param.ProbabilisticReflection )
          else if( Param.DisseminationPolicy == Param.DisseminationPolicyEnum.PROB_REF ||
          		Param.DisseminationPolicy == Param.DisseminationPolicyEnum.PROB_REF_FINE)
          {
        	// JS --------
            if(!k.aware_engine().containsAll(world.neighbors())){
              log(2, "sending knowledge " + k.logRepShort());
              k.sentto_engine().addAll(world.neighbors());
              totransmit.add(k);
            }
            // ------- JS
            else if( k.aware_engine().containsAll(world.neighbors()) && !k.sentto_engine().containsAll(world.neighbors()) )
            {
              // JS: Support Maneesh's Probabilistic Reflection -------
              double r = Param.ReflectionParameter;
              if(Param.DisseminationPolicyEnum.PROB_REF_FINE == Param.DisseminationPolicy)
                r = k.transmissionProbability;
              // ------- JS
            	
              double n = world.neighbors().size();
              //double r = Param.ReflectionParameter;
              if( dist.sampleUniform(0, n) < r )
              {
                log(2, "sending knowledge " + k.logRepShort() + " via broadcast - probabilistic refection ["+r/n+"]");
                k.sentto_engine().addAll(world.neighbors());
                totransmit.add(k);
              }
            }
          }

          if(count == kb_size && totransmit.size() > 0 || totransmit.size() >= Param.MaxKnowledgeItemsPerBundle)
          {
            // Application knowledge to byte array
        	//byte [] data = kb2byte(totransmit);									// JS : comment out
        	// JS : Split totransmit into knowledge bases if it's over the PAYLOAD_SIZ -------
        	ArrayList<KnowledgeBase> kbs = splitKb(totransmit, PAYLOAD_SIZ);	
        	Iterator kbs_it = kbs.iterator();
        	while(kbs_it.hasNext()){
        		byte[] data = kb2byte((KnowledgeBase)kbs_it.next());
        	// ------- JS
	            DatagramPacket o_packet;
	
	            //System.out.println(logRepShort() + " KnowledgeBase Size " + totransmit.size() + " sentto" + "Broadcast");
	            Iterator<InetAddress> bli = world.broadcastIP().iterator();
	            while(bli.hasNext())
	            {
	              InetAddress broadIP = bli.next();
	              o_packet = new DatagramPacket(data, data.length, broadIP, Param.SharedPort);
	              try {
			        log(2, "sending UDP datagram to " + broadIP.getHostAddress() + ":" + Param.SharedPort + " at time " + Engine.time());
	                socket.send(o_packet);
	              } catch(Exception e) { /* do nothing, it's ok not possible to send */ }
	            }
        	

	            /* Neighbor List iter */
	            Iterator<NeighborInfo> i = world.neighbors().iterator();
	            while(i.hasNext())
	            {
	              NeighborInfo ni = i.next();
	
	              /* iteration of IPList in Neighbor Info */
	              Iterator<IPInfo> ip_iter = ni.ipList().iterator();
	              while(ip_iter.hasNext())
	              {
	                IPInfo n_ip = ip_iter.next();
	                /* only send to outer subnet, because already braodcast above */
	                if(n_ip.isForeign() || !Param.EnableBroadcast)
	                {
	                  o_packet = new DatagramPacket(data, data.length, n_ip.ip(), n_ip.port());
	                  try {
			            log(2, "sending UDP datagram to " + n_ip.ip().getHostAddress() + ":" + n_ip.port() + " at time " + Engine.time());
	                    socket.send(o_packet);
	                  } catch(Exception e) { /* do nothing, it's ok not possible to send */ }
	                }
	              }
	            }
        	}	// JS
            totransmit = new KnowledgeBase();
          }
        }
      }
    }
    catch(Exception e) 
    {
      Log.stderr.println("KnowledgeManager.UDPKnowledgeTransmit - " + e);
      e.printStackTrace();
    }

    if(periodic )
    {
      double delay = dist.sampleUniform(Param.KnowledgeDisseminationIntervalMin, Param.KnowledgeDisseminationIntervalMax);
      node().postEvent(  new UDPKnowledgeTransmitEvent(delay, node() ) );
    }
  }

  public void print()
  {
    RealEngine.rep.println("KnowledgeManager " + logRepShort());
    RealEngine.rep.begin();
    RealEngine.rep.println("KnowledgeItems: " + kb().size());
    RealEngine.rep.println("DirectUpdates: " + directUpdates);
    RealEngine.rep.println("IndirectUpdates: " + indirectUpdates);
    RealEngine.rep.println("KnownDestinations: ");
    RealEngine.rep.begin();
    RealEngine.rep.end();
    if(Param.SharedKnowledgeBase)
      RealEngine.rep.println("SharedKnowledgeBase: " );
    else
      RealEngine.rep.println("LocalKnowledgeBase: ");
    RealEngine.rep.begin();
    RealEngine.rep.println("size: " + kb().size());
    kb().print();
    RealEngine.rep.end();
    RealEngine.rep.end();
  }

  /* handle Ping Knowledge */
  private void handlePingKnowledge(PingKnowledge k)
  {
    PingResponseKnowledge prk = new PingResponseKnowledge(node().name(), node().time()  , node().time() + Param.PingResponseKnowledgeExpireTime, k.seq());
    requestUpdate(prk);
  }
  
  // JS : Returned each KnowledgeBase could be a little bigger 
  //      than payloadSize for members of KnowledgeBase class.
  private ArrayList<KnowledgeBase> splitKb(KnowledgeBase kb, final int payloadSize){
	  ArrayList<KnowledgeBase> kblist = new ArrayList<KnowledgeBase>();
	  KnowledgeBase currentKb = new KnowledgeBase();
	  
	  int currentKbSize = 0;
	  for(ApplicationKnowledge k: kb){
		  int knowledgeSize = knowledgeSize(k); 
		  
		  // If the size of knowledge is bigger than payload size, drop the knowledge.
		  if(knowledgeSize > payloadSize){
			  Log.stderr.println("KnowledgeManager.splitKb: drop " + k);
			  continue;
		  }
		  
		  currentKbSize += knowledgeSize;
		  
		  if(currentKbSize < payloadSize)
			  currentKb.update(k);
		  
		  else{
			  kblist.add(currentKb);
			  currentKb = new KnowledgeBase();
			  currentKb.update(k);
			  currentKbSize = knowledgeSize;
		  }
	  }
	  
	  kblist.add(currentKb);
	  
	  return kblist;
  }
  
  // JS
  private int knowledgeSize(ApplicationKnowledge k){
	  try 
	    {
	      ByteArrayOutputStream b_out = new ByteArrayOutputStream();
	      ObjectOutputStream o_out = new ObjectOutputStream(b_out);
	      o_out.writeObject(k);
	      o_out.close();
	      return b_out.toByteArray().length;
	    }
	    catch(Exception e)
	    {
	      Log.stderr.println("KnowledeManager.kb2byte - " + e);
	      e.printStackTrace();
	      return -1;
	    } 
  }

  private byte [] kb2byte(KnowledgeBase kb)
  {
    try 
    {
      ByteArrayOutputStream b_out = new ByteArrayOutputStream();
      ObjectOutputStream o_out = new ObjectOutputStream(b_out);
      o_out.writeObject(kb);
      o_out.close();
      return b_out.toByteArray();
    }
    catch(Exception e)
    {
      Log.stderr.println("KnowledeManager.kb2byte - " + e);
      e.printStackTrace();
      return null;
    }
  }

  private int [] randomSequence(int size) 
  {
    BitSet mark = new BitSet(size);
    int [] s = new int[size];
    mark.clear();
    int count=0;
    while(count < size)
    {
      int num = dist.sampleUniform(0, size);
      if( mark.get(num) ) continue;
      mark.set(num);
      s[count] = num;
      count ++;
    }
    return s;
  }

  /** INNER CLASSES **/

  class UDPKnowledgeTransmitEvent extends Event
  {
    RealNode node;
    UDPKnowledgeTransmitEvent(double delay, RealNode node)
    {
      super(delay);
      this.node = node;
    }
    public void handle() 
    {
      node.km().UDPKnowledgeTransmit(true);
    }
  }
}
