/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.real;

import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.InterfaceAddress;
import java.net.NetworkInterface;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import cf.api.ApplicationKnowledge;
import cf.common.CyberThread;
import cf.common.Engine;
import cf.common.Event;
import cf.common.KnowledgeBase;
import cf.common.Model;
import cf.common.Node;
import cf.common.NodeList;
import cf.common.Param;
import cf.common.World;
import cf.util.Log;

public class RealWorld extends World
{
  /** CONSTRUCTOR **/
	
  public RealWorld()
  {		
    super();
  }

  /** MEMBER VARS **/
  
  private NodeList networkNodes = new NodeList(); // SimNodes which send/receives knowledge to/from UDP networks

  private String myHostname; // hostname of this cyber-host
  private String myShortHostname; // short name of this cyber-host
  private int myPort; // my unique Port number in the cyber-host
  private int myIndex; // my unique index in the cyber-host ( Port - local minimum Port number ) 

  private LinkedList<InetAddress> broadcastIP = new LinkedList<InetAddress>(); // Store all broadcast IP addresses of network interfaces which can be accessed by this process
  private IPList myIP = new IPList(); // store all IP addresses of network interface of this machine

  private IPList userDefinedNeighbors = new IPList(); // store IP information of neighbors which are defined by user explicitly
  private NeighborList neighbors = new NeighborList();


  private DatagramSocket socket; // socket for Shared Port
  private DatagramSocket socket_local; // socket for unique local port

  private KnowledgeReceiver kr;  // knowledge receiver listening on shared port
  private KnowledgeReceiver kr_local_socket;  // knowledge receiver listening on unique local port

  private LinkedList<NeighborIPExpireEvent> neighbor_ip_expire_events = new LinkedList<NeighborIPExpireEvent>();

  /** MEMBER FUNCS **/
  
  //PADO
  public int numTerminated()
  {
    return RealEngine.numTerminated();
  }
  //END of PADO
  
  public NodeList allNetworkNodes() 
  {
    return networkNodes; 
  }
  public int myIndex() 
  {
    return myIndex; 
  }
  public String myHostname() 
  {
    return myHostname; 
  }
  public String myShortHostname() 
  {
    return myShortHostname; 
  }
  public LinkedList<InetAddress> broadcastIP() 
  {
    return broadcastIP; 
  }
  public IPList myIP() 
  {
    return myIP; 
  }
  public IPList userDefinedNeighbors() 
  {
    return userDefinedNeighbors; 
  }
  public NeighborList neighbors() 
  {
    return neighbors; 
  }
  public DatagramSocket socket() 
  { 
    return socket; 
  }
  public DatagramSocket socket_local() 
  {
    return socket_local; 
  }
  /* determine Node to communicate through Networks */
  public void setNetworkNode(RealNode n)
  {
    networkNodes.add(n);
  }
  /* add Neighbor IP + port  to user defined list explicitly */
  public void addNeighbor(String ip, int port)
  {
    try {
      InetAddress address = InetAddress.getByName(ip);
      IPInfo ipinfo = new IPInfo(address,  port , ip );      
      boolean foreign = !isMySubnet(InetAddress.getByName(ip));
      ipinfo.setForeign(foreign);
      userDefinedNeighbors.add( ipinfo );
      // RealEngine.log.println(0, "adding user-defined neighbor: " + ipinfo.toString()); // log file not yet open
    } catch ( Exception e ) {
      Log.stderr.println("cannot resolve neighbor " + ip);
    }
  }

  /* add Neighbor IP + Param.SharedPort to user defined list explicitly */
  public void addNeighbor(String ip)
  {
      addNeighbor(ip, Param.SharedPort);
  }

  public void initialize()
  {
    if(Param.UseRealNetwork)
    {
      try { 
        UDPInit(); 
      } catch (Exception e ) { 
     	Log.stderr.println("RealWorld.initialize - " + e);
        e.printStackTrace();
      }

      RealEngine.add(new UDPBroadcastMyIPEvent(0.0, myIP));
      changeNames(); 

      for(int i = 0; i < Param.UserDefinedNeighbors.length; i++)
	  {
	      addNeighbor(Param.UserDefinedNeighbors[i]); 
	  }
    }

    if(Param.CyberThread && Param.MultiQ)
    {
      RealEngine.cyberThreadExecutor = Executors.newFixedThreadPool(allNodes.size());
 	  for(Model n : allModels)  
      { 
        RealEngine.cyberThreadExecutor.execute(new CyberThread(this, n));
      }
      synchronized(this) {
        cyberThreadStart = true;
      }
    }
    else {
      if(Param.CyberThread && !Param.MultiQ)
      {
        RealEngine.cyberThreadExecutor = Executors.newFixedThreadPool(Param.ThreadPoolSize);
      }
   
      for(Model n : allModels) 
      { 
    	if(n instanceof Node) 
        {
          Node node = (Node) n;
          node.km().initialize();
        }
       
        n.initialize();
      }
    }

    //for(CyberApplication app : allApplications)
    //{
    //  app.initialize();
    //}
  }

  /* will be called when finish the world */
  public void close() 
  {
    try
    {
      if(Param.UseRealNetwork)
      {
        kr.halt();
        kr_local_socket.halt();
        try {
          socket.close();
          socket_local.close();
        }catch(Exception e ) {
          Log.stderr.println("RealWorld.clsoe - " + e);
          e.printStackTrace();
        }
	/*
        Iterator<NeighborInfo> i = neighbors.iterator();
        while(i.hasNext())
        {
         Log.stdout.println("neighbor: " + i.next().toString());
        }
	*/
      }
      if(Param.CyberThread)
      {
        RealEngine.cyberThreadExecutor.shutdown();
        try 
        {
          while(!RealEngine.cyberThreadExecutor.awaitTermination(2000,TimeUnit.MILLISECONDS));
        }
        catch(InterruptedException e)
        {
          e.printStackTrace();
        }
        Log.stdout.println("all threads terminated");
      }
    }
    catch(Exception e)
    {
      Log.stderr.println("RealWorld.close - " + e);
      e.printStackTrace();
    }
  }

  void handleIncomingIPKnowledge( IPKnowledge ip_k, DatagramPacket dp)
  {
    if(ip_k.engineId() == Engine.myEngineId())
      return;

    /* make sure that there's no future created knowledge */
    if(Param.LogicalTimeSync)
    {
      adjustTime(ip_k);
    }

    /* coming from same subnet */
    if(isMySubnet(ip_k.ip_list()))
    {

      boolean sameSubnet = true;
      NeighborInfo ni = new NeighborInfo( ip_k.engineId(), ip_k.ip_list(), sameSubnet);
      add_neighbor(ni);
    }
    /* coming from other subnet */
    else
    {
      //System.out.println("IP knowledge from other subnet: "+ip_k.ip_list().toString());
      /* get  public IP Address, attach it to HelloMessage 
         only in case of direct hop */
      if(ip_k.hop() == 1)
      {
        IPInfo public_ip = new IPInfo(dp.getAddress(), dp.getPort(), dp.getAddress().toString().substring(0,dp.getAddress().toString().indexOf('/')));

        /*attach to hello message to forward */
        ip_k.ip_list().add(public_ip);
      }

      /* add neighbor as other subnet */
      boolean sameSubnet = false;
      NeighborInfo ni = new NeighborInfo( ip_k.engineId(), ip_k.ip_list(), sameSubnet);
      add_neighbor(ni);

      /* forward incoming IP knowledge to other, when hop count is still valid */
      int cur_hop = ip_k.hop();
      if(Param.MultiHopDiscovery > cur_hop )
      {
        try
        {
          /* change Hop Count and send again */
          ip_k.setHop(cur_hop + 1);
          ip_k.aware_engineId().add(Engine.myEngineId());
          byte [] data = ipKnowledge2byte(ip_k);

          /* broadcast to local subnet */
          Iterator<InetAddress> bli = broadcastIP.iterator();
          while(bli.hasNext())
          {
            InetAddress broadIP = bli.next();
            DatagramPacket o_packet = new DatagramPacket(data, data.length, broadIP, Param.SharedPort);
            try {
              socket_local.send(o_packet);
            } catch(Exception e) { /* do nothing, it's ok not possible to send */ }
          }

          /* send to outer neighbors */

          /* Neighbor List iter */
          Iterator<NeighborInfo> i = neighbors.iterator();
          while(i.hasNext())
          {
            NeighborInfo ni_ = i.next();

            if( !ip_k.aware_engineId().contains( ni_.CyberEngineId() ) )
            {
              /* iteration of IPList in Neighbor Info */
              Iterator<IPInfo> ip_iter = ni_.ipList().iterator();
              while(ip_iter.hasNext())
              {
                IPInfo n_ip = ip_iter.next();
                /* only send to outer subnet, because already braodcasted above */
                if(n_ip.isForeign() || !Param.EnableBroadcast)
                {
                  DatagramPacket o_packet = new DatagramPacket(data, data.length, n_ip.ip(), n_ip.port());
                  try {
                    socket_local.send(o_packet);
                  } catch(Exception e) { /* do nothing, it's ok not possible to send */ }
                }
              }
            }
          }
        }
        catch (Exception e)
        {
          Log.stderr.println("RealWorld.handleIncomingIPKnowledge - " + e);
          e.printStackTrace();
        }
      } // end of if(hop count < 

      /* add IP information of hello message(IP) from other subnet as user defined neighbor to send hello message back */
      IPList neighbor_ip_list = ip_k.ip_list();
      Iterator<IPInfo> i = neighbor_ip_list.iterator();
      while(i.hasNext())
      {
        IPInfo n_ip = i.next();
        boolean foreign = !isMySubnet(n_ip.ip());
        n_ip.setForeign(foreign);
        userDefinedNeighbors.add( n_ip );
      }
    }// end of if(mySubnet) else
  }

  /* Make an IPKnowledge(HelloMessage) from ip and broadcast */
  void UDPBroadcastIP(IPList ip)
  {
    try
    {
      IPKnowledge toSend = new IPKnowledge("helloMessage", RealEngine.time(), 10/*menaingless value*/); 
      toSend.ip_list().addAll(ip);
      toSend.setHop(1); // makr Hop as 1
      toSend.engineId = Engine.myEngineId();
      toSend.aware_engineId.add(Engine.myEngineId());
      byte [] data = ipKnowledge2byte(toSend);

      /* Broadcast */
      Iterator<InetAddress> bli = broadcastIP.iterator();
      while(bli.hasNext())
      {
        InetAddress broadIP = bli.next();
        DatagramPacket o_packet = new DatagramPacket(data, data.length, broadIP, Param.SharedPort);
        try{
          socket_local.send(o_packet);
        } catch(Exception e) { /* do nothing, it's ok not possible to send */ }
      }

      /* try to
         Send hello message to User defined Neighbor */

      /* send IP information to user defined neighbor set */
      Iterator<IPInfo> user_neighbor_iter = userDefinedNeighbors.iterator();
      while(user_neighbor_iter.hasNext())
      {
        IPInfo un = user_neighbor_iter.next();
        if(un.isForeign() || !Param.EnableBroadcast)
        {
          DatagramPacket o_packet = new DatagramPacket(data, data.length, un.ip(), un.port());
          try{
            socket_local.send(o_packet);
          } catch(Exception e) { /* do nothing, it's ok not possible to send */ }
        }
      }

    }
    catch (Exception e)
    {
      Log.stderr.println("RealWorld.UDPBroadcastIP - " + e);
    }

    RealEngine.add(new UDPBroadcastMyIPEvent(Param.HelloBroadcastDelay, myIP));
  }

  /* determine given addr is same subnet or not */
  boolean isMySubnet(InetAddress addr) 
  {
    int addr_int = InetAddress2int(addr);
    //System.out.println(Integer.toHexString(addr_int));
    Iterator<IPInfo> ip_iter = myIP.iterator();
    while(ip_iter.hasNext())
    {
      IPInfo ip = ip_iter.next();

      int mask;
      int myaddr_int;

      mask = 0xFFFFFFFF << (32 - ip.prefixLength());
      myaddr_int = InetAddress2int(ip.ip());

      //System.out.println(Integer.toHexString(myaddr_int));
      //System.out.println(Integer.toHexString(mask));
      //System.out.println(Integer.toHexString(mask&myaddr_int));
      //System.out.println(Integer.toHexString(mask&addr_int));

      if( (mask & myaddr_int) == (mask & addr_int) )
        return true;
    }
    return false;
  }

  void changeForeignState(IPList ip_list)
  {
    Iterator<IPInfo> i = ip_list.iterator();
    while(i.hasNext())
    {
      IPInfo ip = i.next();
      boolean foreign = !isMySubnet(ip.ip());
      ip.setForeign(foreign);
    }
  }

  /* true when one of ip is in my subnet */
  boolean isMySubnet(IPList ip_list)
  {
    Iterator<IPInfo> i = ip_list.iterator();
    while(i.hasNext())
    {
      IPInfo ip = i.next();
      /* check if same subnet */
      if(!isMySubnet(ip.ip()))
        return false;
    }
    return true;
  }

  /* check knowledge received queue which is filled with knowledge from UDP */
  void checkIncomingKnowledge(KnowledgeBase kb)
  {
    try {
   	if(!Param.SILENT)  RealEngine.log.println("check incoming knowledge"); 
      ListIterator<ApplicationKnowledge> k_iter = kb.listIterator();
      while(k_iter.hasNext()) 
      {
        ApplicationKnowledge k = k_iter.next();

        /* make sure that there's no knowledge created in the future */
        if(Param.LogicalTimeSync)
        {
          adjustTime(k);
        }

        k.clear_sets(); // clear sentto, local aware node sets and others

        for(Node n : allNetworkNodes())
        {
          RealNode node = (RealNode)n;
          node.postKnowledgeFromUDP( k );
          if(Param.SharedKnowledgeBase || !Param.UseSimNetwork)
            break;
        }
      }
    }catch (Exception e)
    {
      Log.stderr.println("RealWorld.checkIncomingKnowledge - " + e);
      e.printStackTrace();
    }
  }

  void add_neighbor(NeighborInfo ni)
  {
    changeForeignState(ni.ipList());
    /* check if already known neighbor */
    if(!neighbors.contains(ni)) 
    {
      neighbors.add(ni);
      /* handle Neighbor valid life time */
      NeighborIPExpireEvent niee = new NeighborIPExpireEvent(Param.NeighborExpirationTime, this, ni);
      RealEngine.add(niee);
      neighbor_ip_expire_events.add(niee); // event history list
      if(!Param.SILENT) RealEngine.log.println(0, "neighbor added - " + ni.toString());
    }

    /* already exist  then postpone expire time */
    else 
    {
      Iterator<NeighborIPExpireEvent> niee_iter = neighbor_ip_expire_events.iterator();
      while(niee_iter.hasNext())
      {
        NeighborIPExpireEvent niee = niee_iter.next();
        if(niee.ni().CyberEngineId() == ni.CyberEngineId())
        {
          niee.cancel();
		  niee_iter.remove();
          niee = new NeighborIPExpireEvent(Param.NeighborExpirationTime, this, ni);
          RealEngine.add( niee);
          neighbor_ip_expire_events.add(niee);
          break;
        }
      }
    }
  }

  private int InetAddress2int(InetAddress addr)
  {
    byte [ ] temp = addr.getAddress();
    int result;
    result =
      ( ( (int) temp[0] & 0xFF ) << 24 ) |
      ( ( (int) temp[1] & 0xFF ) << 16 ) |
      ( ( (int) temp[2] & 0xFF ) << 8 ) |
      ( ( (int) temp[3] & 0xFF )  ) ;
    return result;
  }

  /* change CyberNode's name with host name + processID(port offset) */
  private void changeNames() 
  {
    for(Node n : allNodes)
    {
	n.setName(n.name() + "@" + myShortHostname + ":" + myIndex);
    }
  }

  /* check if port number is available,
     return Socket with that port if available 
     if not  return null */
  private DatagramSocket available_port(int port) 
  {
    DatagramSocket ds = null;
    try {
      ds = new DatagramSocket(port); //  wild card address 0.0.0.0
      return ds;
    } catch (Exception e) 
    {
      return null;
    }
  }

  private void adjustTime(ApplicationKnowledge k)
  {
    /* make sure that there's no future created knowledge */
    if(Param.LogicalTimeSync)
    {
      double time_diff = k.creationTime()  - RealEngine.time();
      if( time_diff > Param.MaxTimeError && time_diff > 0 ) // this means knowledge sender's time goes faster at least of time_diff amount
      {
        RealEngine.fixTime( time_diff );
      }
    }

  }

  /* initialize necessary things for UDP network */
  private void UDPInit() throws Exception
  {
    try {
      /* socket for Broadcast with Shared Port */
      socket = new DatagramSocket(null); // wild card address 0.0.0.0
      socket.setReuseAddress(true);
      socket.setBroadcast(true);
      //socket.setLoopbackMode(false);
      socket.bind( new InetSocketAddress(Param.SharedPort) );

      /* socket for unicast purpose with unique port */
      for(int i=Param.LocalPortMin; i <= Param.LocalPortMax; i++ )
      {
        socket_local = available_port(i);
        if(socket_local != null){
          break;
        }
      }
      if(socket_local == null) {
    	Log.stderr.println("no available local port number in range");
        System.exit(0);
      }
    }
    catch (Exception e)
    {
      Log.stderr.println("RealWorld.UDPInit - " + e);
      e.printStackTrace();
    }
    //System.out.println(socket.getLocalAddress().toString());

    myPort = socket_local.getLocalPort();

    /* real host name of this machine */
    myHostname = InetAddress.getLocalHost().getHostName();    

    int doti = myHostname.indexOf('.');
    if(doti >= 0) myShortHostname = myHostname.substring(0,doti);
    else myShortHostname = myHostname;

    /* process Id is offset of mimimum unique port number */
    myIndex = myPort - Param.LocalPortMin ;

    if(!Param.SILENT) RealEngine.log.println(0, "host name: " + myHostname + " => short host name: " + myShortHostname);
    if(!Param.SILENT) RealEngine.log.println(0, "shared port: " + socket.getLocalPort() + " local port: " + myPort + " => engine index: " + myIndex);

    /* get network interface, and extract IP address and each of broadcast address and real IP address of this machine */
    Enumeration<NetworkInterface> interfaces = NetworkInterface.getNetworkInterfaces();
    while (interfaces.hasMoreElements()) {
      NetworkInterface networkInterface = interfaces.nextElement();
      //      if (networkInterface.isLoopback())
      //        continue;    // Don't want to broadcast to the loopback interface
      for (InterfaceAddress interfaceAddress : networkInterface.getInterfaceAddresses()) {
        InetAddress broadcast = interfaceAddress.getBroadcast();
        InetAddress myaddress = interfaceAddress.getAddress();
        if(Param.EnableBroadcast)
        {
          if (broadcast != null && !(broadcast instanceof Inet6Address) && !(broadcast.isLoopbackAddress()))
          {
            broadcastIP.add(broadcast);
          }
        }
        if ( !myaddress.isLoopbackAddress() && !(myaddress instanceof Inet6Address))
        {
          short prefixLength;
          prefixLength = interfaceAddress.getNetworkPrefixLength();
          myIP.add(new IPInfo(myaddress,myPort, myHostname, prefixLength) );
        }
      }
    }

    /* Default Broadcast Address */
    if(Param.EnableBroadcast && Param.EnableUnrestrictedBroadcast)
    {
      broadcastIP.clear();
      broadcastIP.add(InetAddress.getByName("255.255.255.255"));
    }


    /* Print IP Information */
    Iterator<IPInfo> ip_iter = myIP.iterator();
    while(ip_iter.hasNext())
    {
      IPInfo ip = ip_iter.next();
      if(!Param.SILENT) RealEngine.log.println(0, "local IP address: " + ip.ip().getHostAddress() );
    }
    Iterator<InetAddress> ip_iter_ = broadcastIP.iterator();
    while(ip_iter_.hasNext())
    {
      InetAddress ip = ip_iter_.next();
      if(!Param.SILENT) RealEngine.log.println(0, "broadcast IP address: " + ip.getHostAddress() );
    }

    /* start two thread for listening knowledge incomming  with each shared and unique socket */
    kr = new KnowledgeReceiver(socket, this);
    kr_local_socket = new KnowledgeReceiver(socket_local, this);
    kr.start();
    kr_local_socket.start();
  }
  /* function for IPKnowledge to Byte */
  private byte [] ipKnowledge2byte(IPKnowledge k)
  {
    try 
    {
      ByteArrayOutputStream b_out = new ByteArrayOutputStream();
      ObjectOutputStream o_out = new ObjectOutputStream(b_out);
      o_out.writeObject(k);
      o_out.close();
      return b_out.toByteArray();
    }
    catch(Exception e)
    {
      Log.stderr.println("RealWorld.ipKnowledge2byte - " + e);
      e.printStackTrace();
      return null;
    }
  }


  /** INNER CLASSES **/

  /* broad cast my information Periodically */
  class UDPBroadcastMyIPEvent extends Event
  {
    IPList ip_list;
    UDPBroadcastMyIPEvent(double delay, IPList ip_list)
    {
      super(delay);
      this.ip_list = ip_list;
    }
    public void handle()
    {
      UDPBroadcastIP(ip_list);
    }
  }

  private class NeighborIPExpireEvent extends Event
  {
    RealWorld world;
    NeighborInfo ni;
    NeighborIPExpireEvent(double delay, RealWorld world, NeighborInfo ni)
    {
      super(delay);
      this.world = world;
      this.ni = ni;
    }

    NeighborInfo ni()
    {
      return ni;
    }

    public void handle(){
      if(!Param.SILENT) RealEngine.log.println(0, "neighbor removed - "+ ni.toString() );
      NeighborList nl = world.neighbors();
      nl.remove(ni);
      neighbor_ip_expire_events.remove(this);
    }
  }
}
