/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.NumberFormat;
import java.util.LinkedList;
import java.util.PriorityQueue;

// for approximation 
import org.apache.commons.math.distribution.*; 
import org.apache.commons.math.stat.*; 

import cf.api.ApplicationEvent;
import cf.common.Engine;
import cf.common.Event;
import cf.common.Model;
import cf.common.Node;
import cf.common.Param;
import cf.common.World;
import cf.common.Engine.TimeThread;
import cf.real.RealEngine;
import cf.sim.mason.SimWorldMason;
import cf.util.Log;

// for mason
import sim.engine.Schedule;
import sim.engine.SimState;

/** simple discrete event simulator using a single priority queue */
public abstract class SimEngine extends Engine
{
	/** class initializer */
	abstract public SimWorld createSimWorld();

	public static void initialize() 
	{
		Engine.initialize();

		time = 0;

		q = new PriorityQueue<Event>();

		if(Param.EnableStage)
		{
			Log.stderr.println("EnableStage Should Be off");
			System.exit(1);
		}

		Param.RealTime = false;

		if(!Param.RealTime) //FIXME: Distributed & Multi-thread SimEngine Mode ?
		{
			Param.UseRealNetwork = false;
			Param.CyberThread = false;
		}

		if(Param.SharedKnowledgeBase)
			Param.UseSimNetwork = false;

		if(Param.CyberThread)
		{
			Param.UseSimNetwork = false; // FIXME:temporarily !! SimLink Synchronization necessary
		}

		if(Param.UseComputerTime)
		{
			Param.MaxTime += ((double)System.currentTimeMillis())/1000.0;
		}
	}

	public boolean invariantProperty(SimWorld world)
	{
		return true;
	}

	public boolean terminationProperty(SimWorld world)
	{
		return true;
	}

	private static long tdiff=0;

	public void run(World world, boolean display)
	{
		int count = 0;
		TimeThread tt = null;
		
		if(display)
		{
			// tt = new TimeThread(world);
			// tt.start();
		}
		try
		{
			while(true)
			{		
				boolean ip = invariantProperty((SimWorld)world);
				if(!ip) 
				{
                                    if(!Param.SILENT){ SimEngine.log.println(1, "events processed: " + count + " - invariant violation");  }
                                    break;
				}

				Event e;
				if(Param.CyberThread)
				{
					synchronized(q) {
						e = q.poll();
					}
				}
				else
				{
					e = q.poll();
				}

				if(e == null) 
				{
                                    if(!Param.SILENT){ SimEngine.log.println(1, "events processed: " + count + " - no further events"); }
                                    break;
				}

				if(e.time() > Param.MaxTime)
				{
                                    if(!Param.SILENT){ SimEngine.log.println(1, "events processed: " + count + " - maximum time reached"); }
                                    break;
				}

				time = e.time();
				if(!Param.SILENT){ SimEngine.log.println(1, "--- SimEngine Time " + Log.tf.format(time()) + " ---"); }
				//long start = System.nanoTime(); 
				if(!e.cancelled()) e.trigger();
				//long end = System.nanoTime(); 
				//tdiff += (end-start)/1000;
				//System.out.println("--- trigger diff " + (end-start)/1000 + " acc " + tdiff );
				//System.out.flush();
				count++;
			}
		}
		finally
		{
			if(tt != null) 
			{
				tt.terminate();
			}

		}
	}

	public void runWithGUI() throws IOException
	{
		Log.stdout.println("running in simulation mode (mason with GUI) ...");
		beginRound();
		initialize();
		simworld = (SimWorldMason) createSimWorld();   
		simworld.initialize();
		simworld.createController(); 
	}

	public static void quit()
	{
		try { endRound(); } 
		catch (IOException e) { e.printStackTrace(); }

		rep.println("Final World State");
		Node.printAll(simworld);

		Log.stdout.println("Done");
		Log.stdout.println("");
	}

	public void runWithoutGUI()
	throws IOException
	{
		int ipsatcount = 0;
		int tpsatcount = 0;

		long simStartTime = System.currentTimeMillis();

		Log.stdout.println("running in simulation mode (mason w/o GUI) ...");

		for(int i = 0; i < Param.Rounds; i++)
		{	
			beginRound();

			initialize();

			simworld = (SimWorldMason) createSimWorld();
			simworld.initialize();

			runWithoutGUI(simworld,true);

			rep.println("Final World State");
			Node.printAll(simworld);

			boolean ip = invariantProperty(simworld);
			rep.println("");
			rep.println("Invariant Property satisfied: " + ip);
			Log.stdout.println("Invariant Property satisfied: " + ip);
			if(ip) ipsatcount++;

			boolean tp = terminationProperty(simworld);
			rep.println("");
			rep.println("Termination Property satisfied: " + tp);
			Log.stdout.println("Termination Property satisfied: " + tp);
			if(tp) tpsatcount++;

			endRound();
		}

		long simEndTime = System.currentTimeMillis();

		Log.stdout.println("Total Computation Time: " + Log.tf.format((double)(simEndTime-simStartTime)/1000.0));
		if(Param.Rounds > 1)
		{
			Log.stdout.println("Invariant Property satisfied in " + ipsatcount + " out of " + Param.Rounds + " executions - p = " + ((double)ipsatcount/(double)Param.Rounds));
			Log.stdout.println("Termination Property satisfied in " + tpsatcount + " out of " + Param.Rounds + " executions - p = " + ((double)tpsatcount/(double)Param.Rounds));
		}

		Log.stdout.println("Done");
		Log.stdout.println("");
	}

	public void runWithoutGUI(SimWorldMason world, boolean display)
	{
		int count = 0;
		TimeThread tt = null;

		if(display)
		{
			tt = new TimeThread(world);
			tt.start();
		}

		try
		{
			SimState.nameThread(simworld.simstate);
			Log.stdout.println("Simulation Seed: " + simworld.simstate.seed());
			//simworld.simstate.start();
			
			Schedule schedule = simworld.simstate.schedule;

			while(true)
			{
				if(schedule.time() > Param.MaxTime)
				{
					Log.stdout.println("Events processed: " + count + " - maximum simulation time reached");
					break;
				}

				if (!schedule.step(simworld.simstate))
				{
					Log.stdout.println("Events processed: " + count + " - no further events");
					break;
				}
				count++;
			}

			simworld.simstate.finish();
		}

		finally
		{
			if(tt != null) 
			{
				tt.terminate();
			}

		}
	}

	// Generic approximation by Lassaigne et al., 
	// Ann. Pure Appl. Logic 2008
	// "Probabilistic Verification and Approximation Schemes" 
	// 0 < epsilon <= 1, 0 < delta <= 1   
	public void generic_approx_run(double epsilon, double delta)
	throws IOException
	{
		int ipsatcount = 0;
		int tpsatcount = 0;

		long StartTime = System.currentTimeMillis();

		Log.stdout.println("running in simulation mode (generic approximation algorithm) ...");

		int Rounds = (int)Math.ceil((double)(Math.log(2.0/delta)/(2.0 * epsilon * epsilon))); 
		Log.stdout.println("epsilon: " + epsilon + ", delta: " + delta + " => rounds: " + Rounds);

		for(int i = 1; i <= Rounds; i++)
		{	
			beginRound();

			initialize();
			start_time = System.currentTimeMillis();

			SimWorld world = createSimWorld();
			world.initialize();
			run(world,true);
			world.close();

			boolean ip = invariantProperty(world);
			rep.println("");
			rep.println("invariant property satisfied: " + ip);
			if(ip) ipsatcount++;

			boolean tp = terminationProperty(world);
			rep.println("");
			rep.println("termination property satisfied: " + tp);
			if(tp) tpsatcount++;

			endRound();
		}

		long EndTime = System.currentTimeMillis();

		Log.stdout.println("total computation time: " + Log.tf.format((double)(EndTime-StartTime)/1000.0));
		Log.stdout.println("invariant property satisfied in " + ipsatcount + " out of " + Rounds + " executions - p = " + ((double)ipsatcount/(double)Rounds));
		Log.stdout.println("termination property satisfied in " + tpsatcount + " out of " + Rounds + " executions - p = " + ((double)tpsatcount/(double)Rounds));

		Log.stdout.println("done");
		Log.stdout.println("");
	}

	public double stopping_rule_algo(double epsilon, double delta, int z)
	throws IOException
	{
		int satcount = 0;
		int Rounds = 0;

		// ramda = (e - 2) ~ 0.72 
		double Gamma = 1.0 + (1.0 + epsilon) * 4.0 * 0.72 * ((double)(Math.log(2.0/delta)/(epsilon * epsilon))); 
		while (satcount < Gamma)
		{	
			Rounds++;

			beginRound();

			initialize();
			start_time = System.currentTimeMillis();

			SimWorld world = createSimWorld();
			world.initialize();
			run(world,true);
			world.close();

			boolean p; 
			if (z == 0) // ip
			p = invariantProperty(world);
			else // tp
				p = terminationProperty(world);

			if(p) satcount++;

			endRound();
		}

		Log.stdout.println("stopping rule algorithm - rounds: " + Rounds + ", gamma: " + Gamma);

		return ((double)satcount/(double)Rounds); 	  
	}

	// Optimal approximation by Dagum et al., 
	// SIAM Journal on Computing 2000
	// "An Optimal Algorithm for Monte Carlo Estimation" 
	// 0 < epsilon <= 1, 0 < delta <= 1, c > 1, z: property (0 means ip, 1 means tp)   
	public void optimal_approx_run(double epsilon, double delta, double c, int z)
	throws IOException
	{
		int satcount = 0;
		int sat = 0;
		int psat = 0;
		double sum = 0.0; 

		long StartTime = System.currentTimeMillis();

		Log.stdout.println("running in simulation mode (optimal approximation algorithm) ...");

		Log.stdout.println("epsilon: " + epsilon + ", delta: " + delta);

		Log.stdout.println("*** 1st step - stopping rule (0.5, " + delta/3.0 + ")");

		double mu = stopping_rule_algo(0.5, delta/3.0, z); 
		Log.stdout.println("mu: " + mu);

		double gamma = 4.0 * 0.72 * (double)(Math.log(2.0/delta)/(epsilon * epsilon)); 
		// ramda = (e - 2) ~ 0.72 
		int Rounds = (int)Math.ceil(c * gamma * epsilon / mu); 
		Log.stdout.println("*** 2nd step - rounds: " + Rounds);

		for(int i = 1; i <= Rounds; i++)
		{	
			beginRound();

			initialize();
			start_time = System.currentTimeMillis();

			SimWorld world = createSimWorld();
			world.initialize();
			run(world,true);
			world.close();

			boolean p; 
			if (z == 0) // ip
			p = invariantProperty(world);
			else // tp
				p = terminationProperty(world);

			if(p) sat = 1;
			else sat = 0; 
			if(i > 1)
				sum += (double)(sat - psat) * (double)(sat - psat) / 2.0;  
			psat = sat; 

			endRound();
		}

		double rho = Math.max (sum / (double)Rounds, epsilon * mu);  
		Log.stdout.println("rho: " + rho + " = max(std_dev/n,epsilon*mu) = max(" + sum + "/" + Rounds + ", " +
				epsilon + "*" + mu + ")");

		// ramda = (e - 2) ~ 0.72 
		Rounds = (int)Math.ceil(c * gamma * rho / (mu * mu)); 
		Log.stdout.println("*** 3rd step - rounds: " + Rounds);

		for(int i = 1; i <= Rounds; i++)
		{	
			beginRound();

			initialize();
			start_time = System.currentTimeMillis();

			SimWorld world = createSimWorld();
			world.initialize();
			run(world,true);
			world.close();

			boolean p; 
			if (z == 0) // ip
			p = invariantProperty(world);
			else // tp
				p = terminationProperty(world);

			if(p) satcount++;

			endRound();
		}	

		long EndTime = System.currentTimeMillis();

		Log.stdout.println("total computation time: " + Log.tf.format((double)(EndTime-StartTime)/1000.0));
		if (z == 0) 
			Log.stdout.println("invariant property satisfied in " + satcount + " out of " + Rounds + " executions - p = " + ((double)satcount/(double)Rounds));
		else 
			Log.stdout.println("termination property satisfied in " + satcount + " out of " + Rounds + " executions - p = " + ((double)satcount/(double)Rounds));

		Log.stdout.println("done");
		Log.stdout.println("");
	}

	// Vesta black-box approach by Sen et al., 
	// CAV 2004 
	// "Statistical Model-checking of Black-box Probabilistic Systems Vesta" 
	// th : threshold
	public void blackbox_approx_run(double th, int Rounds)
	throws IOException
	{
		try {
			int ipsatcount = 0;      
			int tpsatcount = 0;

			long StartTime = System.currentTimeMillis();

			Log.stdout.println("running in simulation mode (blackbox hypothesis testing) ...");

			Log.stdout.println("rounds: " + Rounds + ", threshold: " + th);

			for(int i = 1; i <= Rounds; i++)      
			{
				beginRound();

				initialize();
				start_time = System.currentTimeMillis();

				SimWorld world = createSimWorld();
				world.initialize();
				run(world,true);
				world.close();

				boolean ip = invariantProperty(world);
				rep.println("");
				rep.println("invariant property satisfied: " + ip);
				if(ip) ipsatcount++;

				boolean tp = terminationProperty(world);
				rep.println("");
				rep.println("termination property satisfied: " + tp);
				if(tp) tpsatcount++;

				endRound();
			}

			long EndTime = System.currentTimeMillis();

			Log.stdout.println("total computation time: " + Log.tf.format((double)(EndTime-StartTime)/1000.0));

			Log.stdout.println("invariant property satisfied in " + ipsatcount + " out of " + Rounds + " executions - p = " + ((double)ipsatcount/(double)Rounds));

			BinomialDistribution binomial = new BinomialDistributionImpl(Rounds, th);
			double p = binomial.cumulativeProbability(ipsatcount); 
			if (((double) ipsatcount / (double) Rounds) > th)
				Log.stdout.println (" property (p < " + th + ") violates, p-value: " + (1.0-p));
			else if (((double) ipsatcount / (double) Rounds) < th) 
				Log.stdout.println (" property (p < " + th + ") holds, p-value: " + p);
			else 
				Log.stdout.println (" property undecidable ");

			Log.stdout.println("termination property satisfied in " + tpsatcount + " out of " + Rounds + " executions - p = " + ((double)tpsatcount/(double)Rounds));
			p = binomial.cumulativeProbability(tpsatcount); 
			if (((double) tpsatcount / (double) Rounds) > th)
				Log.stdout.println (" property (p < " + th + ") violates, p-value: " + (1.0-p));
			else if (((double) tpsatcount / (double) Rounds) < th) 
				Log.stdout.println (" property (p < " + th + ") holds, p-value: " + p);
			else  
				Log.stdout.println (" property undecidable ");

			Log.stdout.println("done");
			Log.stdout.println("");
		}
		catch (Exception e)
		{
			Log.stdout.println("exception");
		}
	}

	// Sequential probability ratio test by Younes et al., 
	// Journal on Software Tools for Technology Transfer 2006
	// "Numerial vs. Statistical Probabilistic Model Checking" 
	// alpha: false negative (TYPE I error), beta: false positive (TYPE II error) 
	// theta: threshold, delta: indifference region
	// z: property (0 means ip, 1 means tp)  
	public void seqtest_approx_run(double alpha, double beta, double theta, double delta, int z)
	throws IOException
	{
		int satcount = 0;      
		int Rounds = 0; 
		int h = 0;

		double p0 = theta + delta;
		double p1 = theta - delta;
		double A = (1 - beta) / alpha;
		double B = beta / (1 - alpha);

		long StartTime = System.currentTimeMillis();

		Log.stdout.println("running in simulation mode (sequential hypothesis testing) ...");

		Log.stdout.println("alpha: " + alpha + ", beta: " + beta + ", theta: " + theta + ", delta: " + delta);

		while (true)      
		{
			Rounds++; 

			beginRound();

			initialize();
			start_time = System.currentTimeMillis();

			SimWorld world = createSimWorld();
			world.initialize();
			run(world,true);
			world.close();

			boolean p; 
			if (z == 0) // ip
			p = invariantProperty(world);
			else // tp
				p = terminationProperty(world);

			if(p) satcount++;

			double condition = (Math.pow (p1, satcount) * Math.pow (1 - p1, Rounds - satcount)) / (Math.pow (p0, satcount) * Math.pow (1 - p0, Rounds - satcount));

			if (condition <= B) {
				h = 0;
				break;
			}
			if (condition >= A) {
				h = 1;
				break;
			}

			endRound();
		}

		long EndTime = System.currentTimeMillis();

		Log.stdout.println("total computation time: " + Log.tf.format((double)(EndTime-StartTime)/1000.0));
		if (z == 0) 
			Log.stdout.println("invariant property satisfied in " + satcount + " out of " + Rounds + " executions - p = " + ((double)satcount/(double)Rounds));
		else 
			Log.stdout.println("termination property satisfied in " + satcount + " out of " + Rounds + " executions - p = " + ((double)satcount/(double)Rounds));
		if (h == 0) 
			Log.stdout.println(" accept H0: (theta+delta) < p");   
		else 
			Log.stdout.println(" accept H1: p < (theta-delta)");   

		Log.stdout.println("done");
		Log.stdout.println("");
	}

	public boolean jb_test(double[] observed, int n, double p)
	throws IOException
	{
		try {
			double mean = StatUtils.mean(observed,0,n);
			double var = StatUtils.variance(observed,0,n);
			double stddev = Math.sqrt(var);
			double skew, kurto, jb;
			double temp;

			temp = 0.0;
			for (int i=0;i<n;i++){
				temp += Math.pow((observed[i]-mean),3.0);
			}
			skew = temp/(Math.pow(stddev, 3.0)*n);

			temp = 0.0;
			for (int i=0;i<n;i++){
				temp += Math.pow((observed[i]-mean),4.0);
			}
			kurto = temp/(Math.pow(stddev, 4.0)*n);

			jb = n*((Math.pow(skew, 2.0)/6.0)+(Math.pow(kurto-3.0, 2.0)/24.0));

			/*
        System.out.println("mean " + mean);
        System.out.println("var  " + var);
        System.out.println("dev  " + stddev);
        System.out.println("skew " + skew);
        System.out.println("k    " + kurto);
        System.out.println("jb   " + jb);
			 */

			ChiSquaredDistribution chisquare = new ChiSquaredDistributionImpl(2.0);
			double pval = 1.0 - chisquare.cumulativeProbability(jb);

			if (pval >= p) {
				Log.stdout.println ("rounds: " + n + ", H0: data is normally distributed -> fail to reject - p-value: " + pval);
				return true;
			}
			else {
				Log.stdout.println ("rounds: " + n + ", H0: data is normally distributed -> reject - p-value: " + pval);

				if(var == 0)
					Log.stdout.println (" std dev is zero, expected value = " + observed[0] + " without statistical guarantee");

				return false;
			}
		}
		catch (Exception e)
		{
			Log.stdout.println("exception");
		}
		return false;
	} 

	// Quantitative analysis by Kim et al., 
	// FMOODS 2007 
	// "A Probabilistic Formal Approach to Cross-layer Optimization in Distributed Embedded Systems" 
	// alpha: error bound, delta: confidence interval, pval: pvalue for JB test
	// minsize: minimum rounds, maxsize: maximum rounds, rate: % of sample (=rounds) increase  
	// z: property (0 means ip, 1 means tp)   
	public void qa_approx_run(double alpha, double delta, double pval, int minsize, int maxsize, double rate, int z)
	throws IOException
	{
		try {
			int satcount = 0;      
			int Rounds = 0; 

			double val[] = new double[maxsize]; 	
			double bval[] = new double[maxsize]; 	// bootstrapped val 
			long StartTime = System.currentTimeMillis();

			Log.stdout.println("running in simulation mode (quantitative analsysis) ...");

			Log.stdout.println("alpha: " + alpha + ", delta: " + delta + ", p-value(JBtest): " + pval);

			Log.stdout.println("*** 1st step - JB normality test");

			for (int i=0; i<minsize; i++) // initial samples      
			{
				beginRound();

				initialize();
				start_time = System.currentTimeMillis();

				SimWorld world = createSimWorld();
				world.initialize();
				run(world,true);
				world.close();

				boolean p; 
				if (z == 0) // ip
				p = invariantProperty(world);
				else // tp
					p = terminationProperty(world);

				if(p) 
				{
					satcount++;
					val[Rounds] = 1.0;
				}
				else
					val[Rounds] = 0.0; 

				Rounds++; 

				endRound();
			}

			// bootstraping 
			for (int i=0; i<Rounds; i++) 
			{
				double temp=0; 

				for (int j=0; j<Rounds; j++) 
					temp += val[(int) (Math.random () * Rounds)];

				bval[i] = temp / (double) Rounds; 
			} 

			while (!jb_test(bval, Rounds, pval))
			{  
				if (Rounds >= maxsize)
				{
					Log.stdout.println("JB test fails - reached max sample size " + maxsize);

					long EndTime = System.currentTimeMillis();

					Log.stdout.println("total computation time: " + Log.tf.format((double)(EndTime-StartTime)/1000.0));

					Log.stdout.println("done");
					Log.stdout.println("");
					return; 
				}

				// (Rounds * rate) more samples (at least 1 more sample)  
				int n; 
				if ((int)(Rounds * rate) > 1) n = (int)(Rounds * rate); 
				else n = 1; 	
				for (int i=0; i<n; i++)       
				{
					if (Rounds >= maxsize)
						break; 

					beginRound();

					initialize();
					start_time = System.currentTimeMillis();

					SimWorld world = createSimWorld();
					world.initialize();
					run(world,true);
					world.close();

					boolean p;
					if (z == 0) // ip
					p = invariantProperty(world);
					else // tp
					p = terminationProperty(world);

					if(p)
					{
						satcount++;
						val[Rounds] = 1.0;
					}
					else
						val[Rounds] = 0.0;

					Rounds++;

					endRound();
				}

				// bootstraping 
				for (int i=0; i<Rounds; i++) 
				{
					double temp=0; 

					for (int j=0; j<Rounds; j++) 
						temp += val[(int) (Math.random () * Rounds)];

					bval[i] = temp / (double) Rounds; 
				} 
			}

			Log.stdout.println("*** 2nd step - z-score");

			double sum=0, sumSquare=0;
			int count = Rounds;
			for (int i=0; i<Rounds; i++)
			{
				sum += bval[i];
				sumSquare += (bval[i] * bval[i]);
			}

			NumberFormat format = NumberFormat.getInstance ();
			format.setMaximumFractionDigits (5);
			NormalDistribution normaldist = new NormalDistributionImpl();
			double zScore = normaldist.inverseCumulativeProbability (1 - (alpha / (double) 2));
			double s = Math.sqrt ((sumSquare - ((sum * sum) / (double) count)) / (double) (count - 1));
			double avg = sum / (double) count;
			double d = zScore * s / Math.sqrt (count) / avg;
			while (d > delta) {
				Log.stdout.println ("rounds: " + count + ", " + "(d = " + format.format (d) + ") > (delta = " + delta + ")");

				if (Rounds >= maxsize)
				{
					Log.stdout.println("z-score fails - reached max sample size " +
							maxsize);
					break;
				}

				// (Rounds * rate) more samples (at least 1 more sample)  
				int n; 
				if ((int)(Rounds * rate) > 1) n = (int)(Rounds * rate); 
				else n = 1; 	
				for (int i=0; i<n; i++)       
				{
					if (Rounds >= maxsize)
						break; 

					beginRound();

					initialize();
					start_time = System.currentTimeMillis();

					SimWorld world = createSimWorld();
					world.initialize();
					run(world,true);
					world.close();

					boolean p;
					if (z == 0) // ip
					p = invariantProperty(world);
					else // tp
					p = terminationProperty(world);

					if(p)
					{
						satcount++;
						val[Rounds] = 1.0;
					}
					else
						val[Rounds] = 0.0;

					Rounds++;

					endRound();
					if (Rounds >= maxsize)
						break;
				}

				// bootstraping 
				for (int i=0; i<Rounds; i++) 
				{
					double temp=0; 

					for (int j=0; j<Rounds; j++) 
						temp += val[(int) (Math.random () * Rounds)];

					bval[i] = temp / (double) Rounds; 
				} 

				sum=0; sumSquare=0; 
				for (int i=0; i<Rounds; i++) 
				{
					sum += bval[i];
					sumSquare += (bval[i] * bval[i]); 
				}

				count = Rounds; 
				s = Math.sqrt ((sumSquare - ((sum * sum) / (double) count)) / (double) (count - 1));
				avg = sum / (double) count;
				d = zScore * s / Math.sqrt (count) / avg;
			}
			if (d <= delta)
				Log.stdout.println ("rounds: " + count + ", " + "(d = " + format.format (d) + ") <= (delta = " + delta + ")");

			long EndTime = System.currentTimeMillis();

			Log.stdout.println("total computation time: " + Log.tf.format((double)(EndTime-StartTime)/1000.0));
			if (z == 0) 
				Log.stdout.println("invariant property satisfied in " + satcount + " out of " + Rounds + " executions - p = " + avg);
			else 
				Log.stdout.println("termination property satisfied in " + satcount + " out of " + Rounds + " executions - p = " + avg);

			Log.stdout.println (" with error: " + alpha + ", confidence interval " + format.format (d));

			Log.stdout.println("done");
			Log.stdout.println("");
		}
		catch (Exception e) 
		{
			Log.stdout.println("exception");
		}
	}

	public void run()
	throws IOException
	{
		if(Param.EnableMason)
		{
			if(Param.GUI) runWithGUI();
			else runWithoutGUI();	
		}
		else
		{
			int ipsatcount = 0;
			int tpsatcount = 0;

			long StartTime = System.currentTimeMillis();

			Log.stdout.println("running in simulation mode ...");

			for(int i = 1; i <= Param.Rounds; i++)
			{	
				beginRound();

				initialize();
				start_time = System.currentTimeMillis();

				SimEngine.log.println(1, "--- Simulation Time: " + Log.tf.format(time()) + " ---");

				SimEngine.log.println(1, "=== Creating Simulation Instance ===");

				SimWorld world = createSimWorld();

				SimEngine.log.println(1, "=== Initializing Simulation Instance ===");
				world.initialize();

				SimEngine.log.println(1, "=== Processing Simulation Events ===");

				run(world,true);

				rep.println("final world state");
				world.close();
				Model.printAll(world);

				boolean ip = invariantProperty(world);
				rep.println("");
				rep.println("invariant property satisfied: " + ip);
				if(Param.ShowRounds) Log.stdout.println("invariant property satisfied: " + ip);

				if(ip) ipsatcount++;

				boolean tp = terminationProperty(world);
				rep.println("");
				rep.println("termination property satisfied: " + tp);
				if(Param.ShowRounds) Log.stdout.println("termination property satisfied: " + tp);

				if(tp) tpsatcount++;

				endRound();
			}

			long EndTime = System.currentTimeMillis();
                        if(Param.SILENT){
                        	Log.stderr.println("total computation time: " + Log.tf.format((double)(EndTime-StartTime)/1000.0));
                        } else {
                            Log.stdout.println("total computation time: " + Log.tf.format((double)(EndTime-StartTime)/1000.0));
                        }
			Log.stdout.println("invariant property satisfied in " + ipsatcount + " out of " + Param.Rounds + " executions - p = " + ((double)ipsatcount/(double)Param.Rounds));
			Log.stdout.println("termination property satisfied in " + tpsatcount + " out of " + Param.Rounds + " executions - p = " + ((double)tpsatcount/(double)Param.Rounds));

			Log.stdout.println("done");
			Log.stdout.println("");
		}
	}

	public static void beginRound()
	throws IOException
	{
		if(Param.ShowRounds) 
		{
			Log.stdout.println("-------------- Simulation Round " + round + " -------------------");		
		}
		
		String logHome = Param.LoggingDirectory;
		  
	    if(logHome.equals("") || !(new File(logHome).isDirectory())){
	  	  logHome = System.getProperty("java.io.tmpdir");
	  	  
	  	  File logHomeDir = new File(logHome);
	  	  if(!logHomeDir.isDirectory())
	  		  logHomeDir.mkdirs();
	  	  Log.stdout.println("\""+Param.LoggingDirectory+"\" doesn't exist. Logs are stored under \"" + logHome+"\"");
	    }
	  
	    // use short name for log files to overwrite logs
		if(Param.OverwriteLog){
			RealEngine.log = new Log(logHome + File.separator + "trace.log");
			RealEngine.rep = new Log(logHome + File.separator + "report.log");
		}
		  
		else{
			RealEngine.log = new Log(logHome + File.separator + "trace-" + round + ".log");
			RealEngine.rep = new Log(logHome + File.separator + "report-" + round + ".log");
		}
	}

	public static void endRound()
	throws IOException
	{
		SimEngine.log.close();		
		round++;
	}

	public static void fork_stage(String java_stage_ctrl_class_name, String java_stage_ctrl_class_path, String world_file_name, boolean is_stage_gui)
	{
		for(int i = 0; i < Param.Rounds; i++)
		{	
			String gui;
			if(is_stage_gui) gui = "";
			else gui = "--gui";
			String[] cmd = { "/bin/sh", "-c", "stage " + gui  +" --ctrlclassname "+ java_stage_ctrl_class_name+ " --ctrlclasspath "+ java_stage_ctrl_class_path + " " + world_file_name+ " | /usr/bin/tee logging/stage_output_"+i };
			Log.stdout.println(cmd[2]);

			try {
				String s;
				Process oProcess = new ProcessBuilder(cmd).start();

				BufferedReader stdOut   = new BufferedReader(new InputStreamReader(oProcess.getInputStream()));
				BufferedReader stdError = new BufferedReader(new InputStreamReader(oProcess.getErrorStream()));

				while ((s =   stdOut.readLine()) != null) {Log.stdout.println(s);}
				while ((s = stdError.readLine()) != null) {Log.stderr.println(s);}

			} catch (IOException e1) {
				Log.stderr.println(e1);
				e1.printStackTrace();
				System.exit(1);
			}

			// keep log files
			copy_log(i);
		}
	}

	private static void copy_log(int num)
	{
		try {
			String s;
			String [] cmd_1 = {"/bin/sh", "-c", "/bin/mv logging/report-0.log logging/report-stage-"+num+".log"};
			String [] cmd_2 = {"/bin/sh", "-c", "/bin/mv logging/trace-0.log logging/trace-stage-"+num+".log"};
			Process oProcess_1 = new ProcessBuilder(cmd_1).start();
			Process oProcess_2 = new ProcessBuilder(cmd_2).start();

			//      BufferedReader stdOut   = new BufferedReader(new InputStreamReader(oProcess.getInputStream()));
			//      BufferedReader stdError = new BufferedReader(new InputStreamReader(oProcess.getErrorStream()));
			//
			//      while ((s =   stdOut.readLine()) != null) System.out.println(s);
			//      while ((s = stdError.readLine()) != null) System.err.println(s);

		} catch (IOException e1) {
			Log.stderr.println(e1);
			e1.printStackTrace();
			System.exit(1);
		}

	}


	/* Generate Simple world file with robot number */

	public static void stage_world_gen(int robot_num, String world_file_name, int link_success_rate, int end_time) 
	{
		String [] cmd ={ "/bin/sh", "-c" , "$CF/sri/cf/sim/stage/genworld.py " + robot_num + " " +world_file_name + " " + link_success_rate + " " + end_time};
		try {
			String s;
			Process oProcess = new ProcessBuilder(cmd).start();

			BufferedReader stdOut   = new BufferedReader(new InputStreamReader(oProcess.getInputStream()));
			BufferedReader stdError = new BufferedReader(new InputStreamReader(oProcess.getErrorStream()));

			while ((s =   stdOut.readLine()) != null) Log.stdout.println(s);
			while ((s = stdError.readLine()) != null) Log.stderr.println(s);

		} catch (IOException e1) {
			Log.stderr.println(e1);
			e1.printStackTrace();
			System.exit(1);
		}

	}

	public static void setStage() 
	{
		Param.EnableStage = true;
	}
}
