/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim;

import java.util.BitSet;
import java.util.ListIterator;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.common.Event;
import cf.common.KnowledgeBase;
import cf.common.Model;
import cf.common.Node;
import cf.common.NodeList;
import cf.common.Param;
import cf.sim.dtn.Address;
import cf.sim.dtn.Bundle;
import cf.sim.dtn.BundleList;
import cf.sim.dtn.DataPayload;
import cf.sim.dtn.KnowledgePayload;
import cf.sim.dtn.Payload;
import cf.sim.dtn.SimKnowledgeManager;
import cf.sim.dtn.SimLink;
import cf.sim.dtn.SimLinkList;
import cf.sim.dtn.Transmission;
import cf.sim.dtn.TransmissionPayload;
import cf.sim.stage.JavaStageCtrlCyberFramework;
import cf.util.Log;

public class SimNode extends Node implements CyberNode
{
  /** CONSTRUCTOR **/
 
  public SimNode(SimWorld world)
  {
    super(world);
    this.world = world;
    km = new SimKnowledgeManager(world, this);
    super.km = this.km;
    addr = new Address(this);
    world.allAddr().add(addr);
  }

  public SimNode(SimWorld world, String name)
  {
    super(world,name);
    this.world = world;
    km = new SimKnowledgeManager(world, this);
    super.km = this.km;
    addr = new Address(this);
    world.allAddr().add(addr);
  }

  public SimNode(SimWorld world, Model parent)
  {
    super(world, parent);
    this.world = world;
    km = new SimKnowledgeManager(world, this);
    super.km = this.km;
    addr = new Address(this);
    world.allAddr().add(addr);
  }
  public SimNode(SimWorld world, Model parent, String name)
  {
    super(world, parent, name);
    this.world = world;
    km = new SimKnowledgeManager(world, this);
    super.km = this.km;
    addr = new Address(this);
    world.allAddr().add(addr);
  }

  /*
     for JavaStage 
   */

  public SimNode(JavaStageCtrlCyberFramework ctrl, SimWorld world)
  {
    this(world);
    this.ctrl = ctrl;
  }

  public SimNode(JavaStageCtrlCyberFramework ctrl, SimWorld world, String name)
  {
    this(world,name);
    this.ctrl = ctrl;
  }

  public SimNode(JavaStageCtrlCyberFramework ctrl, SimWorld world, Model parent)
  {
    this(world, parent);
    this.ctrl = ctrl;
  }
  public SimNode(JavaStageCtrlCyberFramework ctrl, SimWorld world, Model parent, String name)
  {
    this(world, parent, name);
    this.ctrl = ctrl;
  }

  /** MEMBER VARS **/
  
  private JavaStageCtrlCyberFramework ctrl; // JavaStage Controller reference
  private SimLinkList outgoingLinks = new SimLinkList(); // outgoing links potentially connected to this node
  private SimLinkList connectedLinks = new SimLinkList(); // outgoing links currently connected to this node
  private SimLinkList sensedLinks = new SimLinkList(); // outgoing links sensed to be connected to this node
  private SimLinkList availableLinks = new SimLinkList(); // outgoing links sensed to be connected to this node and not busy
  private SimLinkList incomingLinks = new SimLinkList(); // incoming links potentially connected to this node
  private int contacts, sensedContacts;

  private BundleList pending = new BundleList();
  private BundleList delivered = new BundleList();
  private BitSet delSeqSet = new BitSet(Param.MaxBundles);
  private Address addr;
  private SimWorld world;
  private SimKnowledgeManager km;


  /** MEMBER FUNCS **/

  public SimKnowledgeManager km() 
  { 
    return km; 
  }

  public SimNode node()
  {
    return this;
  }
  
  public JavaStageCtrlCyberFramework ctrl() 
  {
    return ctrl;
  }

  public  Address addr()
  {
    return addr;
  }

  public  SimLinkList outgoingLinks()
  {
    return outgoingLinks;
  }

  public SimLinkList  incomingLinks()
  {
    return incomingLinks;
  }

  public SimLinkList sensedLinks()
  {
    return sensedLinks;
  }

  public SimLinkList availableLinks()
  {
    return availableLinks;
  }

  /** Add link potentially outgoing from this node. */
  public void addOutgoingLink(SimLink l)
  {
    outgoingLinks.add(l);
  }

  /** Add link potentially incoming to this node. */
  public void addIncomingLink(SimLink l)
  {
    incomingLinks.add(l);
  }

  /** Called if link is opened. */
  public void handleLinkUp(SimLink l)
  {
    contacts++;
    connectedLinks.add(l);
  }

  /** Called if link is closed. */
  public void handleLinkDown(SimLink l)
  {
    connectedLinks.remove(l);
  }

  /** Called if link is sensed open. */
  public void handleLinkSensedUp(SimLink l)
  {
    sensedLinks.add(l);
    sensedContacts++;
  }

  /** Called if link is sensed closed. */
  public void handleLinkSensedDown(SimLink l)
  {
    sensedLinks.remove(l);
  }

  /** Called if link is available. */
  public void handleLinkAvailable(SimLink l)
  {
    availableLinks.add(l);
    //  if(pending.isEmpty()) return;
    //  for(int i=0;i<pending.size();i++) 
    for (ListIterator<Bundle> li = pending.listIterator(); li.hasNext();)
    {
      Bundle b = li.next();
      //Bundle b = pending.get(i);
      if (b.isKnowledge()||b.isData()) // send status bundles first
      {
        performRouting(b);
        if (b.discarded()) li.remove();
        //if (b.discarded()) pending.remove(b);
      }

    }
    //  for (ListIterator<Bundle> li = pending.listIterator(); li.hasNext();)
    //  {
    //    Bundle b = li.next();
    //    if (b.isData()) // then data bundles
    //    {
    //      performRouting(b);
    //      if (b.discarded()) li.remove();
    //    }
    //  }
  }

  public void handleLinkUnavailable(SimLink l)
  {
    availableLinks.remove(l);
  }

  /** Request transmission over specific link. */
  public int queued(SimLink l) 
  {
    return l.queued();
  }

  /** Client requests transmission of status to a given
    destination. Status bundles are lightweight and only sent to
    immediate neighbors without acknowledgments or
    retransmissions. */
  public boolean requestStatusTransmission(Address dest, KnowledgePayload status, double expiration)
  {
    Bundle b = new Bundle(world ,dest,this,status, expiration);
    //        statusBundlesGenerated++;
    if (requestTransmission(b))
    {
      //statusBundlesGeneratedAndAccepted++;
      //statusBundlesGeneratedSize += status.realSize();
      return true;
    }
    return false;
  }

  public void requestTransmission(SimLink l, TransmissionPayload m)
  {
    assert l.transmitter() == this;
    l.requestTransmission(m);
  }

  /** Choose random outgoing next hop link. */
  public boolean transmitBundle(SimLink l, Bundle b)
  {
    assert b.created() || b.received();

    log(2, "link " + l.logRepShort() + " transmitBundle " + b.logRepShort());

    if (l.queued() < Param.BusyQueueLength)
    {
      Bundle cb = b.newTransmission();

      if (b.isKnowledge())
      {
    	log(2, "b.isKnowledge() " + b.logRepLong());
    	boolean retransmit = false; // status bundles are never retransmitted
        // if(retransmit) statusBundlesRetransmitted++; else statusBundlesTransmitted++;
      }
      else
      {
        //SimEngine.add(new BundleRetransmissionEvent(Param.RetransmissionDelay, b));
        postEvent(new BundleRetransmissionEvent(Param.RetransmissionDelay, b));
        // if(retransmit) dataBundlesRetransmitted++; else dataBundlesTransmitted++;
      }

      requestTransmission(l,cb);
      return true;
    }
    else
    {
      log(2, "link " + l.logRepShort() + " busy for " + logRepShort());
      return false;
    }
  }

  public void handleBundleRetransmission(Bundle b)
  {
    if (!b.discarded())
    {
      log(2, "checking for potential retransmission of " + b.logRepLong());
      performRouting(b);
    }
  }

  public void performRouting(Bundle b)
  {
	log("performRouting " + b.logRepLong());  
    if (b.isKnowledge() )
    {
      SimNode dest = b.dest().node(); // null if symbolic address

      if (dest != null)
      {
        SimLink l = selectDirectLink(dest, availableLinks); // needs to be improved if multiple links available
        if (l != null)
        {
          if (transmitBundle(l,b)) b.handleDiscard();
        }
        else 
        {
        }
      }
    }
    else if (b.isData()) // data bundles involve routing
    {
      Log.stderr.println("data bundle routing not supported in this version");
      System.exit(0);
    }
  }

  public void handleReception(SimLink l, Transmission m)
  {
    log(2,"received " + m.logRepLong());
    if (m.payload() instanceof Bundle)
    {
      Bundle b = (Bundle)m.payload();
      b.handleReception(l);

      log(2,"received and accepted " + b.logRepLong());
      processIncomingBundle(b);
    }
  }

  private boolean requestTransmission(Bundle b)
  {
    processIncomingBundle(b);
    return true;
  }

  private void handleKnowledgeReceivedEvent(KnowledgeReceivedEvent e)
  {
    ApplicationKnowledge k = e.k;

    log("handling knowledge: " + k.logRepShort());

    if(!k.expired())
    {
      //ApplicationKnowledge kold = km.kb().subsumed(k);
      ApplicationKnowledge kold = km.kb().ignore(k);
      if(kold == null) 
      {
        km.kb().update(k);

        for(CyberApplication app : allApplications())
        {
          if(app instanceof CyberApplication) 
          {
            k.handle((CyberApplication)app);
          }
        }
      }
      else
      {
        log("ignoring obsolete knowledge - subsumed by: " + kold.toString());
      }
    }
    else
    {
      log("ignoring expired knowledge: " + k.toString());
    }
  }

  private void deliverLocally(Bundle b)
  {
	log("deliver locally " + b.logRepLong());
    if (b.dest().node() == node())
    {
      Payload p = b.payload();
      if (p instanceof KnowledgePayload) deliverKnowledgeBundle(b);
      else assert false;
    }
  }

  private void deliverKnowledgeBundle(Bundle b)
  {
    if (!delSeqSet.get(b.seq()))
    {
      delSeqSet.set(b.seq());
      log(b.logRepLong() + " delivered");
      //KnowledgeManagerHandlers kmh = network.km[index()];
      //KnowledgeManagerHandlers kmh = km;
      //if (kmh != null) kmh.handleStatusDelivery(b.src(),(Status)b.payload());
      if (km != null) km.handleStatusDelivery(b.src(),(KnowledgePayload)b.payload());

      b.handleDiscard(); // new
      pending.remove(b); // new

      handleKnowledgeBundle();
    }
    else
    {
      log(2,"duplicate bundle " + b.logRepLong() + " discarded (already delivered)");
    }
  }

  /** Discard delivered bundles. */
  private void handleKnowledgeBundle()
  {
    log(2,"handle new knowledge bundle");
    //    if(pending.isEmpty()) return ;
    //    for(int i =0; i < pending.size() ; i++ )
    for (ListIterator<Bundle> li = pending.listIterator(); li.hasNext();)
    {
      Bundle b = li.next();
      //      Bundle b = pending.get(i);

      if (!b.discarded())
      {
        performRouting(b);
        if (b.discarded()) li.remove();
        //        if (b.discarded()) pending.remove(b);
      }
    }
  }

  /** Process incoming bundles. There can either come directly from the application or
    they have been received from another node. */
  private void processIncomingBundle(Bundle b)
  {
    Bundle pb = pending.findSeq(b.seq());

    if (pb != null)
    {
      if (Param.DiscardIncomingDuplicate)
      {
        log(2,"incoming bundle discarded " + b.logRepLong() + "(duplicate of pending bundle)");
        b.handleDiscard();
        return;
      }
      else
      {
        log(2,"pending bundle discarded " + pb.logRepLong() + "(to be replaced by incoming duplicate)");
        pb.handleDiscard();
        pending.remove(pb);
      }
    }

    Payload p = b.payload();
    //if(p instanceof Data) updateBundleReceived(b);

    if (p instanceof DataPayload) b.earliestTransmissionTime = SimEngine.time() + SimEngine.dist.sampleUniform(Param.MinBundleWaitingTime,Param.MaxBundleWaitingTime);

    log(2, "queueing " + b.logRepLong());
    pending.add(b);

    double expirationTime = b.timeToExpire() - SimEngine.time();

    if (expirationTime <= 0)
    {
      handleBundleExpiration(b);
    }
    else
    {
      if(expirationTime <= Param.MaxTime)
    	  postEvent(new BundleExpirationEvent(expirationTime, b));

      deliverLocally(b);

      if (b.hops() >= Param.StopAfterHops)
      {
        log(2,"incoming bundle terminated " + b.logRepLong() + "(max hops reached)");
      }
      else
      {
        performRouting(b);
        if (b.discarded()) pending.remove(b);
      }
    }
  }

  private SimLink selectRandomNextHopLink(SimLinkList available)
  {
    int choices = available.size();
    if (choices == 0) 
    {
      return null;
    }
    int selection = SimEngine.dist.sampleUniform(0,choices);
    return available.get(selection);
  }

  /** Select random direct link to destination. */
  private SimLink selectDirectLink(SimNode dest, SimLinkList available)
  {
    return selectRandomNextHopLink(available.selectLinksTo(dest));
  }

  private NodeList outgoingNeighbors()
  {
    NodeList r = new NodeList();
    SimLinkList ol = node().outgoingLinks;
    for (SimLink l : ol)
    {
      SimNode receiver = l.receiver();
      if (!r.contains(receiver)) r.add(receiver);
    }
    return r;
  }     

  private void handleBundleExpiration(Bundle b)
  {
    if (Param.DiscardExpiredBundles && !b.discarded())
    {
      if (b.expired())
      {
        pending.remove(b);
        log(b.logRepLong() + " discarded (global expiration)");
      }
      else
      {
        pending.remove(b);
        log(b.logRepLong() + " discarded (local expiration)");
      }
    }
  }

  public void terminate()
  {
  }
  
  /** INNER CLASSES **/

  private class BundleRetransmissionEvent extends Event
  {
    Bundle b;
    BundleRetransmissionEvent(double delay, Bundle b) {
      super(delay);
      this.b = b;
    }
    public void handle() {
      handleBundleRetransmission(b);
    }
  }

  private class KnowledgeReceivedEvent extends Event
  {
    SimNode node;
    ApplicationKnowledge k;

    KnowledgeReceivedEvent(SimNode node, double delay, ApplicationKnowledge k)
    { 
      super(delay);
      this.node = node;
      this.k = k;
    }

    public void handle()
    {
      node.handleKnowledgeReceivedEvent(this);
    }
  }

  private class BundleExpirationEvent extends Event
  {
    Bundle b;
    BundleExpirationEvent(double delay, Bundle b) {
      super(delay);
      this.b = b;
    }
    public void handle() {
      handleBundleExpiration(b);
    }
  }

  public KnowledgeBase kb() {
	return super.km().kb();
  }
}
