/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim;

import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import cf.api.CyberApplication;
import cf.common.CyberThread;
import cf.common.Event;
import cf.common.Model;
import cf.common.Node;
import cf.common.Param;
import cf.common.World;
import cf.sim.dtn.AddressList;
import cf.sim.dtn.BundleList;
import cf.sim.dtn.LinkType;
import cf.sim.dtn.SimLink;
import cf.sim.dtn.SimLinkList;
import cf.sim.pn.SimMobility;
import cf.util.Log;

public class SimWorld extends World
{
  /** CONSTRUCTOR **/
	
  public SimWorld()
  {		
    super();
  }

  /** MEMBER VARS **/
 
  private SimLinkList allLinks = new SimLinkList(); 
  private AddressList allAddr = new AddressList(); 
  private BundleList allBundles = new BundleList(); 
  private SimMobility sm;
  private int freshLinkIndex=0; // for unique SimLink index in the world

  /** MEMBER FUNCS **/
  
  public void initialize()
  {
	if(Param.UseSimNetwork)
    {
    	initLinks();
    }

    if(Param.CyberThread && Param.MultiQ)
    {
      SimEngine.cyberThreadExecutor = Executors.newFixedThreadPool(allNodes.size());
      for(Model n : allModels) 
      { 
        SimEngine.cyberThreadExecutor.execute(new CyberThread(this, n));
      }
      synchronized(this) {
        cyberThreadStart = true;
      }
    }
    else {
      if(Param.CyberThread && !Param.MultiQ)
      {
        SimEngine.cyberThreadExecutor = Executors.newFixedThreadPool(Param.ThreadPoolSize);
      }
      for(Model n : allModels) 
      { 
        if(n instanceof Node) 
        {
          Node node = (Node) n;
          node.km().initialize();
        }
        n.initialize();
      }
    }

    if(Param.UseSimNetwork)
    {
    	if(Param.DynamicLinks)
    	{
    		for(SimLink l : allLinks)
    		{
    			handleRandomLinkChange(true,l);
    		}
    	}
    }
    
    //for(CyberApplication app : allApplications)
    //{
    //  app.initialize();
    //}
  }

  /* add persistent linke btw parent & child */
  public void initLinks() 
  {
    for(Model model : allModels() )
    {
      if(model instanceof SimNode)
      {
        Model p = model.parent();
        if(p != null) 
        {
          SimLink l = 
            addSymmetricLink((SimNode)(p), (SimNode)model, LinkType.ParentChild, Param.LinkCost, Param.LinkLatency, Param.LinkBandwidth, Param.LinkErrorRate);
          handleLinkChange(true, l);
        }
      }
    }
  }

  /* will be called when finish the world */
  public void close() 
  {
    try
    {
      if(Param.CyberThread)
      {
        SimEngine.cyberThreadExecutor.shutdown();
        try 
        {
          while(!SimEngine.cyberThreadExecutor.awaitTermination(2000,TimeUnit.MILLISECONDS));
        }
        catch(InterruptedException e)
        {
          e.printStackTrace();
        }
        Log.stdout.println("all threads terminated");
      }
    }
    catch(Exception e)
    {
      Log.stderr.println("SimWorld.close - " + e);
      e.printStackTrace();
    }
    //System.out.println("World End");
  }

  public int freshLinkIndex() 
  { 
    return freshLinkIndex++; 
  }
  public SimLinkList allLinks() 
  { 
    return allLinks; 
  }
  public AddressList allAddr() 
  {
    return allAddr; 
  }
  public BundleList allBundles() 
  {
    return allBundles; 
  }
  public SimMobility sm() 
  { 
    return sm; 
  }
  public void setSimMobility(SimMobility sm) 
  { 
    this.sm = sm; 
  }
  public SimNode firstNode() 
  {
    return (SimNode)(allNodes.getFirst());
  }
  public SimNode nextNode(SimNode n) 
  {
    return (SimNode)(allNodes.getNode(n.index() + 1 ));
  }

  public int totLinks()
  { 
    return allLinks.size(); 
  }

  /* function for managing Links,
     comes from CPSSim */
  public SimLink addLink(SimNode transmitter, SimNode receiver, LinkType type,
      double cost, double latency, double bandwidth, double errorRate)
  {
    int index = freshLinkIndex();
    SimLink l = new SimLink(this, index, transmitter, receiver, type, cost, latency, bandwidth, errorRate);
    allLinks.add(l);
    //    link[index] = l;
    return l;
  }

  public SimLink addLink(SimNode transmitter, SimNode receiver, LinkType type,
      double cost, double latency, double bandwidth, double errorRate,
      double avgUpTime, double avgDownTime)
  {
    int index = freshLinkIndex();
    SimLink l = new SimLink(this, index, transmitter, receiver, type, cost, latency, bandwidth, errorRate, avgUpTime, avgDownTime);
    allLinks.add(l);
    //    link[index] = l;
    return l;
  }

  public SimLink addSymmetricLink(SimNode transmitter, SimNode receiver, LinkType type
      )
  {
    return   addSymmetricLink(transmitter, receiver, type, Param.LinkCost, Param.LinkLatency, Param.LinkBandwidth, Param.LinkErrorRate);

  }
  public SimLink addSymmetricLink(SimNode transmitter, SimNode receiver, LinkType type,
      double cost, double latency, double bandwidth, double errorRate)
  {
    SimLink l = addLink(transmitter, receiver, type, cost, latency, bandwidth, errorRate);
    SimLink rl = addLink(receiver, transmitter, type, cost, latency, bandwidth, errorRate);
    l.setReverse(rl);
    rl.setReverse(l);
    return l;
  }

  public SimLink addSymmetricLink(SimNode transmitter, SimNode receiver, LinkType type,
      double cost, double latency, double bandwidth, double errorRate,
      double avgUpTime, double avgDownTime)
  {
    SimLink l = addLink(transmitter, receiver, type, cost, latency, bandwidth, errorRate, avgUpTime, avgDownTime);
    SimLink rl = addLink(receiver, transmitter, type, cost, latency, bandwidth, errorRate, avgUpTime, avgDownTime);
    l.setReverse(rl);
    rl.setReverse(l);
    return l;
  }

  public void handleLinkChange(boolean up, SimLink l)
  {
	if(Param.UseSimNetwork)
	{
		l.handleLinkChange(up);
		SimLink rl = l.reverse();
		if (rl != null) rl.handleLinkChange(up);
	}
  }

  private class RandomLinkChangeEvent extends Event
  {
    SimWorld world;
    boolean up;
    SimLink l;
    RandomLinkChangeEvent(SimWorld world, double delay, boolean up, SimLink l)
    {
      super(delay);
      this.world = world;
      this.up = up;
      this.l = l;
    }
    public void handle() {
      world.handleRandomLinkChange(up,l);
    }
  }

  public void handleRandomLinkChange(boolean up, SimLink l)
  {
    handleLinkChange(up,l);
    if (Param.DynamicLinks)
    {
      if (Param.DeterministicLinks)
      {
        if (up) SimEngine.add(new RandomLinkChangeEvent(this,l.avgUpTime,false,l));
        else SimEngine.add(new RandomLinkChangeEvent(this,l.avgDownTime,true,l));
      }
      else
      {
        if (up) SimEngine.add(new RandomLinkChangeEvent(this,SimEngine.dist.sampleExpWithMean(l.avgUpTime),false,l));
        else SimEngine.add(new RandomLinkChangeEvent(this,SimEngine.dist.sampleExpWithMean(l.avgDownTime),true,l));
      }
    }
  }

  /** INNER CLASSES **/
  
}
