/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim.dtn;

import cf.common.Param;
import cf.sim.SimEngine;
import cf.sim.SimNode;
import cf.sim.SimWorld;
import cf.util.Log;

public class Bundle extends TransmissionPayload
{
  private SimWorld world;

  // Bundle sequence number (uniquely idenfies content of bundle).

  private int seq;
  private static int freshSeq;

  // Ultimate source of bundle.

  private SimNode src;

  // Final destination of bundle.

  private Address dest;

  // Application or administrative payload.

  private Payload payload;

  // State of bundle.

  public BundleState status;

  // Data bundles are distinguished from leightweight knowledge bundles.
  // Knowledge bundles are only used by the knowledge manager and sent to immediate neighbors.
  // They do not involve planning or acknowledgements.

  public boolean isKnowledge() {
    return payload instanceof KnowledgePayload;
  }
  public boolean isData() {
    return payload instanceof DataPayload;
  }
  public boolean isBackground() {
    if (payload instanceof DataPayload) return ((DataPayload)payload).background();
    else return false;
  }
  public boolean isApp() {
    if (payload instanceof DataPayload) return !((DataPayload)payload).background();
    else return false;
  }

  public double earliestTransmissionTime = 0;

  // Bundle statistics.

  private int transmissions;
  private double lastTimeTransmitted;

  private int hops;
  public double resources;

  private double timeCreated;
  private double timeToExpire;

  public double timeDelivered;

  public double timeCreated() {
    return timeCreated;
  }
  public double timeDelivered() {
    return timeDelivered;
  }

  // Observers.

  public int seq() {
    return seq;
  }

  public SimNode src() {
    return src;
  }
  public Address dest() {
    return dest;
  }

  public Payload payload() {
    return payload;
  }
  public double realSize() {
    return payload.realSize();
  }

  public boolean created() {
    return status == BundleState.Created;
  }
  public boolean received() {
    return status == BundleState.Received;
  }
  public boolean discarded() {
    return status == BundleState.Discarded || status == BundleState.Lost;
  }

  public double timeToExpire() {
    return timeToExpire;
  }
  public boolean expired() {
    return SimEngine.time() >= timeToExpire;
  }

  public int hops() {
    return hops;
  }
  public int transmissions() {
    return transmissions;
  }
  public double lastTimeTransmitted() {
    assert transmissions > 0;
    return lastTimeTransmitted;
  }

  // Bundle constructors.

  private Bundle()
  {
  }

  public Bundle(SimWorld world, Address dest, SimNode src, Payload payload, double expiration)
  {
    this.world = world;
    //this.seq = world.freshBundleSeq();
    this.seq = freshSeq++;
    this.dest = dest;
    this.src = src;
    this.timeCreated = SimEngine.time();
    this.timeToExpire = timeCreated + expiration;
    this.payload = payload;
    this.status = BundleState.Created;
    this.resources = Param.BundleResources;
    if (Param.KeepBundles)
    {
      if (!Param.KeepDataBundlesOnly || isData()) world.allBundles().add(this);
    }
  }

  // Make exact copy of bundle and change its plan to the given one.

  public Bundle copy()
  {
    Bundle b = new Bundle();
    b.world = this.world;
    b.seq = this.seq;
    b.src = this.src;
    b.dest = this.dest;
    b.hops = this.hops;
    b.resources = this.resources;
    b.timeCreated = this.timeCreated;
    b.timeToExpire = this.timeToExpire;
    b.payload = this.payload;
    b.status = this.status;
    if (Param.KeepBundles)
    {
      if (!Param.KeepDataBundlesOnly || b.isData()) world.allBundles().add(b);
    }
    return b;
  }

  // Informative abbreviation.

  public String logRepShort()
  {
      if(Param.SILENT){ return ""; }
    return "Bundle(" + seq + ")";
  }

  public String logRepLong()
  {
      if(Param.SILENT){ return ""; }
    return logRepShort() + " : " + src.logRepShort() + " -> " + dest.logRepShort();
  }

  // Called if bundle is discarded.

  public void handleDiscard()
  {
    assert status == BundleState.Created || status == BundleState.Received;
    status = BundleState.Discarded;
  }

  // Called if Bundle is lost

  public void handleLoss()
  {
    assert status == BundleState.InTransit;
    status = BundleState.Lost;
  }

  public boolean lost()
  {
    return status == BundleState.Lost;
  }

  // Called if transmission is disrupted due to link cut off.

  public void handleCutOff(double cutOffSize)
  {
    if (cutOffSize < Param.BundleHeaderSize)
    {
      handleLoss();
    }
    else
    {
      handleLoss(); // currently we assume worst case
    }
  }

  // Called if transmission is disrupted due to transmission error.

  public void handleError()
  {
    handleLoss(); // currently we assume worst case
  }

  // Called just before a bundle is transmitted by some node.

  public Bundle newTransmission() 
  {
    this.transmissions += 1;
    lastTimeTransmitted = SimEngine.time();
    Bundle b = copy();
    b.status = BundleState.InTransit;
    return b;
  }

  // Called if bundle has been received at some node.

  public void handleReception(SimLink l)
  {
    assert status == BundleState.InTransit;
    status = BundleState.Received;
    hops += 1;
    resources -= 1;
  }

  // Methods for individual bundle statistics.

  public void print()
  {
      if(Param.SILENT){ return; }
    SimEngine.rep.println(logRepLong());
    SimEngine.rep.begin();
    SimEngine.rep.println("BundlePayload: " + payload.toString());
    SimEngine.rep.println(" RealSize: " + realSize() +
        " Status: " + status.toString() +
        " Hops: " + hops +
        " TimeCreated: " + Log.tf.format(timeCreated) +
        " TimeToExpire: " + Log.tf.format(timeToExpire) +
        " Transmissions: " + transmissions +
        " LastTimeTransmitted: " + Log.tf.format(lastTimeTransmitted) +
        " TimeDelivered: " + Log.tf.format(timeDelivered) +
        ((timeDelivered == 0) ? "" : " TransitTime: " + Log.tf.format(timeDelivered-timeCreated)));
    SimEngine.rep.end();
  }

  public static void printAll(SimWorld n)
  {
      if(Param.SILENT){ return; }
      for (Bundle b : n.allBundles()) b.print();
  }
}

