/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim.dtn;

import java.util.BitSet;
import java.util.ListIterator;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApplication;
import cf.common.Event;
import cf.common.KnowledgeBase;
import cf.common.KnowledgeManager;
import cf.common.Node;
import cf.common.NodeList;
import cf.common.Param;
import cf.sim.SimEngine;
import cf.sim.SimNode;
import cf.sim.SimWorld;
import cf.util.Log;

/** Naive/dummy implementation of distributed information layer (all available
    knowledge is flooded through the network in regular intervals). */

public class SimKnowledgeManager extends KnowledgeManager
{

    /** CONSTRUCTOR **/
    public SimKnowledgeManager(SimWorld world, SimNode node)
    {
        super(world, node);
        this.node = node;
        this.world = world;
    }

    /** MEMBER VARS **/
    public SimNode node;
    private AddressList knownDestinations = new AddressList();
    private SimWorld world;
    private int statusTransmissions, statusDeliveries;
    private int knowledgeItemsTransmitted;
    private double nextBundleStatusDissemination = 0;
    private double lastNodeStatusRefresh = 0;
    private double lastNodeStatisticsRefresh = 0;
    private double lastLinkStatusRefresh = 0;
    private double lastLinkStatisticsRefresh = 0;

    /** MEMBER FUNCS **/

    public SimNode node()
    {
        return node.node();
    }

    public void initialize()
    {
        if (Param.MaintainNodeKnowledge)
            {
                requestUpdate(new NodeParamKnowledge(node(),node()));
                requestUpdateNode();
            }

        if (Param.MaintainLinkKnowledge)
            {
                SimLinkList ol = node().outgoingLinks();
                for (SimLink l : ol) {
                    requestUpdate(new LinkParamKnowledge(node(),l));
                }
                requestUpdateLinks();
            }

        if(Param.UseSimNetwork) 
            node.postEvent(new StatusTransmissionEvent(dist.sampleUniform(Param.KnowledgeDisseminationIntervalMin,Param.KnowledgeDisseminationIntervalMax)));
    }

    /* Called if status update received. */
    public void manageKnowledge(ApplicationKnowledge e)
    {
        if (e instanceof NodeParamKnowledge) addNodeKnowledge(e);
        else if (e instanceof BundleDeliveredKnowledge) addDeliveredKnowledge(e);
        else if (e instanceof LinkParamKnowledge ||
                 e instanceof NodeStatusKnowledge ||
                 e instanceof NodeStatisticsKnowledge ||
                 e instanceof LinkStatusKnowledge ||  
                 e instanceof LinkStatisticsKnowledge ||  
                 e instanceof BundleDeliveredKnowledge ||  
                 e instanceof BundleReceivedKnowledge
                 ) {}

        /* case: common ApplicationKnowledge */
        else 
            {
                for(CyberApplication app : node().allApplications())
                    {
                        if(app instanceof CyberApplication) 
                            {
                                //		long start = System.nanoTime(); 
                                e.handle((CyberApplication)app);
                                //		long end = System.nanoTime(); 
                                //		System.out.println("--------- app handle takes " + (end-start)/1000);
                                //		System.out.flush();
                            }
                    }
            }
    }

    /* this function only called when indirect( coming from other node)  knowledge comes */
    public  boolean handleKnowledgeUpdate(Node src, ApplicationKnowledge e)
    {
        if (!e.expired())
            {
                if (fresh(e))
                    {
                        //System.out.println(logRepShort() + "New Knowledge " + e.logRepShort() +" Updated");
                        if(!Param.SILENT){  log(1, "received fresh knowledge: "+ e.logRepShort()); }
                        e = e.copy();
                        //e.receivedfrom.set(src.index());
                        if (!e.aware.get(node().index())) e.aware.set(node().index());
                        update(e);
                        indirectUpdates++;

                        manageKnowledge(e);
                        return true;
                    }
                else
                    {
                        if(!Param.SILENT){  log(1, "received known knowledge: " + e.toString() ); }
                        synchronized(kb()) 
                            {
                                for (ApplicationKnowledge k : kb())
                                    {
                                        if (e.related(k) && (k.replaces(e) || k.subsumes(e)))
                                            {
                                                //if (!e.aware.get(node().index())) k.receivedfrom.set(src.index());
                                                k.aware.or(e.aware);
                                                if(!Param.SILENT){ if(Param.LogAwareness) log(2, "node awareness set updated: "  + k.aware.toString()); }
                                                k.aware_engine().addAll(e.aware_engine());
                                                if(!Param.SILENT){ if(Param.UseRealNetwork && Param.LogAwareness) log(2, "engine awareness set updated: "  + k.aware_engine().toString()); }
                                                k.processedfor = new BitSet();
                                            }
                                    }
                            }

                        return false;
                    }
            }
        return false;
    }

    public void handleStatusDelivery(SimNode src, KnowledgePayload s)
    {
        statusDeliveries++;
        for (ApplicationKnowledge e : s.knowledge()) handleKnowledgeUpdate(src,e);
    }

    public void requestUpdateNode()
    {
        if (Param.MaintainNodeStatusKnowledge && Param.RefreshNodeStatusKnowledge)
            {
                if (SimEngine.time() >= lastNodeStatusRefresh + Param.NodeStatusRefreshInterval)
                    {
                        lastNodeStatusRefresh = SimEngine.time();
                        requestUpdate(new NodeStatusKnowledge(node(),node()));
                    }
            }
        if (Param.MaintainNodeStatisticsKnowledge && Param.RefreshNodeStatisticsKnowledge)
            {
                if (SimEngine.time() >= lastNodeStatisticsRefresh + Param.NodeStatisticsRefreshInterval)
                    {
                        lastNodeStatisticsRefresh = SimEngine.time();
                        requestUpdate(new NodeStatisticsKnowledge(node(),node()));
                    }
            }
    }

    public BundleReceivedKnowledge bundleReceived(SimNode n, int seq)
    {
        ListIterator<ApplicationKnowledge> li = kb().listIterator();
        while (li.hasNext())
            {
                ApplicationKnowledge e = li.next();
                if (e instanceof BundleReceivedKnowledge)
                    {
                        BundleReceivedKnowledge bs = (BundleReceivedKnowledge)e;
                        if (bs.bundleReceived(n,seq)) return bs;
                    }
            }
        return null;
    }

    public BundleDeliveredKnowledge bundleDelivered(SimNode n, int seq)
    {
        ListIterator<ApplicationKnowledge> li = kb().listIterator();
        while (li.hasNext())
            {
                ApplicationKnowledge e = li.next();
                if (e instanceof BundleDeliveredKnowledge)
                    {
                        BundleDeliveredKnowledge bs = (BundleDeliveredKnowledge)e;
                        if (bs.bundleDelivered(n,seq)) return bs;
                    }
            }
        return null;
    }

    public void requestUpdateLinks()
    {
        SimLinkList il = node().incomingLinks();
        if (Param.MaintainLinkStatusKnowledge && Param.RefreshLinkStatusKnowledge)
            {
                if (SimEngine.time() >= lastLinkStatusRefresh + Param.LinkStatusRefreshInterval)
                    {
                        for (SimLink l : il)
                            {
                                requestUpdate(new LinkStatusKnowledge(node(),l));
                            }
                        lastLinkStatusRefresh = SimEngine.time();
                    }
            }
        if (Param.MaintainLinkStatisticsKnowledge && Param.RefreshLinkStatisticsKnowledge)
            {
                if (SimEngine.time() >= lastLinkStatisticsRefresh + Param.LinkStatisticsRefreshInterval)
                    {
                        for (SimLink l : il)
                            {
                                requestUpdate(new LinkStatisticsKnowledge(node(),l));
                            }
                        lastLinkStatisticsRefresh = SimEngine.time();
                    }
            }
    }

    public AddressList knownDestinations()
    {
        return knownDestinations;
    }

    public void addKnownDestination(Address dest)
    {
        if (!knownDestinations.contains(dest)) knownDestinations.add(dest);
    }

    // the following functions are triggered when new knowledge is added

    public int nextHop(int curnode, int dest)
    {
        int next = dest;
        while (dest != curnode)
            {
                next = dest;
                //if (!preddefined[curnode][dest]) return -1;
                //dest = pred[curnode][dest];
            }
        return next;
    }

    public SimNode nextHop(SimNode n, SimNode dest)
    {
        int nextHop = nextHop(n.index(),dest.index());
        //if (nextHop >= 0) return network.node[nextHop];
        if (nextHop >= 0) return null;
        else return null;
    }

    public SimNode nextHop(SimNode dest)
    {
        return nextHop(node(), dest);
    }

    public void addNodeKnowledge(ApplicationKnowledge e)
    {
        NodeParamKnowledge np = (NodeParamKnowledge)e;
        SimNode n = np.node();
        //nodearray[n.index()] = n;
    }

    public void addDeliveredKnowledge(ApplicationKnowledge e)
    {
        int transmissions = 0;

        BundleDeliveredKnowledge bs = (BundleDeliveredKnowledge)e;

        //if (Param.LearnFromBackgroundOnly) if (!bs.bundle().isBackground()) return;
    }

    public NodeList outgoingNeighbors()
    {
        NodeList r = new NodeList();
        //LinkList ol = node().outgoingLinks();
        SimLinkList ol = node().availableLinks();
        for (SimLink l : ol)
            {
                SimNode receiver = l.receiver();
                if (!r.contains(receiver)) r.add(receiver);
            }
        return r;
    }

    public void transmitStatus()
    {
        handleStatusTransmissionEvent(false);
    }
  
    /* Called if status needs to be transmitted.
       if periodic true that is event drivened calling
    */
    private void handleStatusTransmissionEvent(boolean periodic)
    {
        log(2, "status transmission event");
	//	long start = System.nanoTime(); 

        requestUpdateNode();
        requestUpdateLinks();

        boolean disseminateBundleStatus = false;

        if (SimEngine.time() >= nextBundleStatusDissemination)
            {
                nextBundleStatusDissemination = SimEngine.time() + dist.sampleUniform(Param.BundleStatusDisseminationIntervalMin,Param.BundleStatusDisseminationIntervalMax);
                disseminateBundleStatus = true;
            }

        //    if (!Param.KnowledgeDeterministicFlooding || kb().additions())
        //    {
        boolean fullydisseminated = true;

        log(2, "preparing status bundle");

        NodeList outgoingNeighbors = outgoingNeighbors();
        int outgoingNeighborsSize = outgoingNeighbors.size();

        // generate status for all outgoing neighbors even if link not available
        // this is important to get a foot into the bundle queue

        // currently status bundles are individualized for each neighbor
        // to exploit broadcasting it may be better to send the same info to all even if redundant for some

        for (Node n : outgoingNeighbors)
            {
                //KnowledgeSet totransmit = new KnowledgeSet();
                KnowledgeBase totransmit = new KnowledgeBase();

                ListIterator<ApplicationKnowledge> li = kb().listIterator();
                while (li.hasNext())
                    {
                        ApplicationKnowledge k = li.next();

                        boolean disseminate = false;

                        if (k instanceof BundleDeliveredKnowledge || k instanceof BundleReceivedKnowledge)
                            {
                                if (disseminateBundleStatus) disseminate = true;
                            }
                        else
                            {
                                disseminate = true; //????
                            }

                        if (disseminate)
                            {
                                boolean aware = k.aware.get(n.index());
                                boolean sent = k.sentto.get(n.index());
                                boolean processed = k.processedfor.get(n.index());

                                if (!sent && !aware)
                                    {
                                        if(!Param.SILENT){  log(2, "adding knowledge " + k.logRepShort() + " for " + n.logRepShort()); }
                                        if (!totransmit.contains(k)) totransmit.add(k);
                                        if (!k.sentto.get(n.index())) k.sentto.set(n.index());
                                    }
                                //else if(Param.ProbabilisticReflection)
                                else if( Param.DisseminationPolicy == Param.DisseminationPolicyEnum.PROB_REF ||
                                  		Param.DisseminationPolicy == Param.DisseminationPolicyEnum.PROB_REF_FINE)
                                    {
	                                	// JS --------
	                                    if(!aware){
	                                    	if(!Param.SILENT){  log(1, "adding knowledge " + k.logRepShort() + " for " + n.logRepShort() + " - probabilistic reflection"); }
	                                    	k.sentto.set(n.index());
	                                        totransmit.add(k);
	                                    }
	                                    // ------- JS
	                                    else if(!sent && aware)
                                            {
	                                        	// JS: Support Maneesh's Probabilistic Reflection -------
	                                            double r = Param.ReflectionParameter;
	                                            if(Param.DisseminationPolicyEnum.PROB_REF_FINE == Param.DisseminationPolicy)
	                                            	r = k.transmissionProbability;
	                                            // ------- JS
                                            
                                                double _n = outgoingNeighborsSize;
                                                if( dist.sampleUniform(0, _n) < r )
                                                    {
                                                        if(!Param.SILENT){  log(1, "adding knowledge " + k.logRepShort() + " for " + n.logRepShort() + " - probabilistic reflection ["+r/_n+"]"); }
                                                        if (!totransmit.contains(k)) totransmit.add(k);
                                                        if (!k.sentto.get(n.index())) k.sentto.set(n.index());
                                                    }
                                            }
                                    }

                                if (!k.processedfor.get(n.index())) k.processedfor.set(n.index());
                            }

                        if (!li.hasNext() && totransmit.size() > 0 || totransmit.size() >= Param.MaxKnowledgeItemsPerBundle)
                            {
                                if(!Param.SILENT){  log(1, "transmitting status to " + n.logRepShort()); }
                                node().requestStatusTransmission(((SimNode)n).addr(),new KnowledgePayload(totransmit),Param.StatusBundleExpiration);
                                statusTransmissions++;
                                knowledgeItemsTransmitted += totransmit.size();
                                totransmit = new KnowledgeBase();
                            }
                    }
            }
        //      if (Param.KnowledgeDeterministicFlooding) kb().done();

        //    }

        if(periodic)
            node().postEvent(new StatusTransmissionEvent(dist.sampleUniform(Param.KnowledgeDisseminationIntervalMin,Param.KnowledgeDisseminationIntervalMax)));
        //long end = System.nanoTime(); 
        //System.out.println("------ status trans diff " + (end-start)/1000);
        //System.out.flush();
    }

    /* Methods for individual node statistics. */

    public void print()
    {
        if(Param.SILENT){  return; }
        SimEngine.rep.println("KnowledgeManager " + logRepShort());
        SimEngine.rep.begin();
        SimEngine.rep.println("KnowledgeItems: " + kb().size());
        SimEngine.rep.println("DirectUpdates: " + directUpdates);
        SimEngine.rep.println("IndirectUpdates: " + indirectUpdates);
        SimEngine.rep.println("StatusTransmissions: " + statusTransmissions);
        SimEngine.rep.println("StatusDeliveries: " + statusDeliveries + " (" + Log.percentage(statusDeliveries,statusTransmissions) + ")");
        SimEngine.rep.println("KnowledgeItemsTransmitted: " + knowledgeItemsTransmitted);
        SimEngine.rep.println("KnownDestinations: ");
        SimEngine.rep.begin();
        SimEngine.rep.println(knownDestinations().toString());
        SimEngine.rep.end();
        if(Param.SharedKnowledgeBase)
            SimEngine.rep.println("SharedKnowledgeBase: " );
        else
            SimEngine.rep.println("LocalKnowledgeBase: ");
        SimEngine.rep.begin();
        SimEngine.rep.println("size: " + kb().size());
        kb().print();
        SimEngine.rep.end();
        SimEngine.rep.end();
    }

    /* Methods for global node statistics. */

    public static int totStatusTransmissions(SimWorld n)
    {
        int r = 0;
        for (KnowledgeManager km : n.allKM()) r += ((SimKnowledgeManager)km).statusTransmissions;
        return r;
    }

    public static int totStatusDeliveries(SimWorld n)
    {
        int r = 0;
        for (KnowledgeManager km : n.allKM()) r += ((SimKnowledgeManager)km).statusDeliveries;
        return r;
    }

    /** INNER CLASSES **/

    /* Event to trigger status transmission. */
    private class StatusTransmissionEvent extends Event
    {
        int seq;
        StatusTransmissionEvent(double delay) {
            super(delay);
        }
        public void handle() {
            handleStatusTransmissionEvent(true);
        }
    }
}
