/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim.dtn;

import cf.common.Event;
import cf.common.Param;
import cf.sim.SimEngine;
import cf.sim.SimNode;
import cf.sim.SimWorld;
import cf.sim.stage.SimulationInterface;
import cf.sim.stage.WifiConfig;
import cf.sim.stage.WifiInterface;
import cf.sim.stage.WifiMessageCF;
import cf.util.Log;

/** A link denotes a potential, directed connection between to end points.
    It is a potential link because it can be open or closed, where
    closed also subsumes the case where the links is nonexistent (e.g. out of range)
    Bidirectional links are modeled using two directed links with
    the same parameters (in case of symmetric links) or different parameters
    (in case of asymmetric links).

    We use a conservative link model so that actual results should be at least as
    good as those obtained by this model.
*/

public class SimLink
{
    SimWorld world;

    /** Link identifier. */

    private int index;
    public int index() {
        return index;
    }

    /** Link end points. */
    private SimNode transmitter, receiver;

    /** Link parameters. */
    private LinkType type;
    private double cost;
    private double latency;
    private double bandwidth;
    private double errorRate;
    private boolean sensing;

    public double avgUpTime, avgDownTime;

    /** Associated reverse link if any. */
    private SimLink reverse;

    /** Link status can be open or closed. */
    private boolean open;

    /** Perceived information about link (we use the simplifying assumption
        that receiver and transmitter perceive same info). */

    /** Link status can be sensed open or closed. */
    private boolean sensedUp, sensedDown;

    /* Total statistics. */
    private int totSensedChanges;
    private double totSensedUpDuration;
    private double totSensedDownDuration;

    /* Historical statistics. */
    public double maxShortDownDuration = 100.0*60*3;
    public double minLongDownDuration = 100.0*60*3;

    private double lastSensedChangeTime;

    private double lastSensedUpEvent;

    private int lastSensedLongDownEventCount;
    private int lastSensedShortDownEventCount;
    private int lastSensedRelativeUpEventCount;
    private int minSensedRelativeUpEventCount = Integer.MAX_VALUE;
    private int maxSensedRelativeUpEventCount = 0;

    private double lastSensedDownEvent;

    private double lastSensedUpDuration;

    private double lastSensedDownDuration;

    /** Transmitter active on this link. */
    boolean transmitting;

    /** Transmissions on the link (to be received). */
    private TransmissionList intransit;

    private TransmissionList waiting;

    /* Physical link statistics (read only for public). */

    public int changes;
    public double lastChangeTime;

    public double sumUpDuration, sumDownDuration, sumUtilDuration;
    public double sumStatusUtilDuration, sumDataUtilDuration, sumBackgroundUtilDuration, sumAppUtilDuration;

    public double lastTransmissionTime;

    public double transSize;
    public int transInitiated, transRejectedSensedDown, transRejectedBusy, transLostDown;
    public int transTerminated, transReceived;
    public int transDisrupted, transDisruptedCutOff, transDisruptedError;
    public int transLost, transLostError, transLostCutOff;

    /* Observers. */
    public LinkType type() {
        return type;
    }
  
    public SimNode receiver() {
        return receiver;
    }
  
    public SimNode transmitter() {
        return transmitter;
    }
  
    public SimLink reverse() {
        return reverse;
    }

    public int queued() {
        return waiting.size();
    }

    public double cost() {
        return cost;
    }
  
    public double bandwidth() {
        return bandwidth;
    }
  
    public double latency() {
        return latency;
    }
  
    public double errorRate() {
        return errorRate;
    }
  
    public boolean sensing() {
        return sensing;
    }

    public boolean open() {
        return open;
    }
  
    public boolean transmitting() {
        return transmitting;
    }
  
    public boolean available() {
        return !transmitting;
    }
  
    public boolean sensedUp() {
        return sensedUp;
    }
  
    public boolean sensedDown() {
        return sensedDown;
    }
  
    public double lastChangeTime() {
        return lastChangeTime;
    }

    public int totSensedChanges() {
        return totSensedChanges;
    }
  
    public double lastSensedUpEvent() {
        return lastSensedUpEvent;
    }
  
    public double lastSensedDownEvent() {
        return lastSensedDownEvent;
    }
  
    public double lastSensedChangeTime() {
        return lastSensedChangeTime;
    }
  
    public int lastSensedLongDownEventCount() {
        return lastSensedLongDownEventCount;
    }
  
    public int lastSensedShortDownEventCount() {
        return lastSensedShortDownEventCount;
    }
  
    public int lastSensedRelativeUpEventCount() {
        return lastSensedRelativeUpEventCount;
    }
  
    public int minSensedRelativeUpEventCount() {
        return minSensedRelativeUpEventCount;
    }
  
    public int maxSensedRelativeUpEventCount() {
        return maxSensedRelativeUpEventCount;
    }

    /** Informative abbreviation. */

    public String name()
    {
        return "l" + index;
    }

    public final String logRepShort()
    {
        if(Param.SILENT){ return ""; }
        return Log.getClassName(this) + "(" + name() + ")";
    }

    public String logRepLong()
    {
        if(Param.SILENT){ return ""; }
        return logRepShort() + " : " + transmitter.logRepShort() + " -> " + receiver.logRepShort();
    }

    public void log(int level, String msg)
    {
        if(Param.SILENT){ return; }
        SimEngine.log.println(level, logRepLong() + " > " + msg);
    }

    /** Link constructor. */

    public SimLink(SimWorld world, int index, SimNode transmitter, SimNode receiver,
                   LinkType type, double cost, double latency, double bandwidth, double errorRate,
                   double avgUpTime, double avgDownTime)
    {
        this.world = world;
        this.index = index;
        this.transmitter = transmitter;
        this.receiver = receiver;
        this.type = type;
        this.cost = cost;
        this.latency = latency;
        this.bandwidth = bandwidth;
        this.errorRate = errorRate;
        this.sensing = Param.LinkSensing;
        this.waiting = new TransmissionList();
        this.intransit = new TransmissionList();
        this.open = false;
        this.sensedDown = true;
        this.sensedUp = false;
        this.lastSensedDownEvent = Double.NaN;
        this.lastSensedDownDuration = Double.NaN;
        this.lastSensedUpEvent = Double.NaN;
        this.lastSensedUpDuration = Double.NaN;
        this.avgUpTime = avgUpTime;
        this.avgDownTime = avgDownTime;
        transmitter.addOutgoingLink(this);
        receiver.addIncomingLink(this);
    }

    public SimLink(SimWorld world, int index, SimNode transmitter, SimNode receiver,
                   LinkType type, double cost, double latency, double bandwidth, double errorRate)
    {
        this(world, index, transmitter, receiver,
             type, cost, latency, bandwidth, errorRate,
             SimEngine.dist.sampleUniform(Param.AvgLinkUpTimeMin,Param.AvgLinkUpTimeMax),
             SimEngine.dist.sampleUniform(Param.AvgLinkDownTimeMin,Param.AvgLinkDownTimeMax));
    }

    public void initLinkState(boolean open)
    {
        this.open = open;
        this.sensedDown = !open;
        this.sensedUp = open;
        this.transmitter.handleLinkUp(this);
        this.transmitter.handleLinkSensedUp(this);
        this.transmitter.handleLinkAvailable(this);

        SimLink r = reverse();
        if(r != null) 
            {
                r.open = open;
                r.sensedDown = !open;
                r.sensedUp = open;
                r.transmitter.handleLinkUp(r);
                r.transmitter.handleLinkSensedUp(r);
                r.transmitter.handleLinkAvailable(r);
            }
    }
  
    public void setReverse(SimLink reverse)
    {
        this.reverse = reverse;
    }

    /** Accumulated time of link being open. */

    public double openDuration()
    {
        if (open()) return sumUpDuration + (SimEngine.time() - lastChangeTime);
        else return sumUpDuration;
    }

    /** Accumulated time of link being closed. */

    public double closedDuration()
    {
        if (!open()) return sumDownDuration + (SimEngine.time() - lastChangeTime);
        else return sumDownDuration;
    }

    /** Called by network upon change of link status (in view of
        potentially correlated link changes, e.g. in case of
        bidirectional links, we require the higher entity,
        i.e. network, to trigger this event). */

    public void handleLinkChange(boolean open)
    {
        if (!this.open && open)
            {
                log(3,"link open");

                sumDownDuration += SimEngine.time() - lastChangeTime;
                this.open = true;
                changes++;
                lastChangeTime = SimEngine.time();
                transmitter.handleLinkUp(this);
                if (sensing())
                    {
                        if (Param.LinkUpSenseDelay == 0) handleLinkSensedUp();
                        else SimEngine.add(new LinkSensedUpEvent(Param.LinkUpSenseDelay,this));
                    }
            }
        else if (this.open && !open)
            {
                log(3,"link closed");

                sumUpDuration += SimEngine.time() - lastChangeTime;
                this.open = false;
                prepareReception();
                handleCutOff();
                changes++;
                lastChangeTime = SimEngine.time();
                transmitter.handleLinkDown(this);
                if (sensing())
                    {
                        if (Param.LinkUpSenseDelay == 0) handleLinkSensedDown();
                        else SimEngine.add(new LinkSensedDownEvent(Param.LinkDownSenseDelay,this));
                    }
            }
    }

    /** Link has been sensed as open. */

    private class LinkSensedUpEvent extends Event
    {
        SimLink l;
        LinkSensedUpEvent(double delay, SimLink l)
        {
            super(delay);
            this.l = l;
        }
        public void handle() {
            l.handleLinkSensedUp();
        }
    }

    private void updateEventCount()
    {
        if (lastSensedDownDuration >= minLongDownDuration)
            {
                lastSensedLongDownEventCount++;
                if (lastSensedLongDownEventCount >= 2 && lastSensedRelativeUpEventCount < minSensedRelativeUpEventCount)
                    minSensedRelativeUpEventCount = lastSensedRelativeUpEventCount;
                lastSensedRelativeUpEventCount = 0;
            }
        else if (lastSensedDownDuration < maxShortDownDuration)
            {
                lastSensedShortDownEventCount++;
                lastSensedRelativeUpEventCount++;
                if (lastSensedRelativeUpEventCount > maxSensedRelativeUpEventCount)
                    maxSensedRelativeUpEventCount = lastSensedRelativeUpEventCount;
            }
    }

    private void handleLinkSensedUp()
    {
        log(2,"sensing link open");
        lastSensedDownDuration = SimEngine.time() - lastSensedChangeTime;
        totSensedDownDuration += lastSensedDownDuration;
        sensedDown = false;
        sensedUp = true;
        totSensedChanges++;
        lastSensedChangeTime = SimEngine.time();
        lastSensedUpEvent = SimEngine.time();
        updateEventCount();
        transmitter.handleLinkSensedUp(this);
        checkForWaitingTransmissions();
        if (available())
            {
                log(2,"link available");
                transmitter.handleLinkAvailable(this);
            }
    }

    /** Link has been sensed as closed. */

    private class LinkSensedDownEvent extends Event
    {
        SimLink l;
        LinkSensedDownEvent(double delay, SimLink l)
        {
            super(delay);
            this.l = l;
        }
        public void handle() {
            l.handleLinkSensedDown();
        }
    }

    private void handleLinkSensedDown()
    {
        log(2,"sensing link closed");
        lastSensedUpDuration = SimEngine.time() - lastSensedChangeTime;
        totSensedUpDuration += lastSensedUpDuration;
        sensedUp = false;
        sensedDown = true;
        totSensedChanges++;
        lastSensedChangeTime = SimEngine.time();
        lastSensedDownEvent = SimEngine.time();
        transmitter.handleLinkSensedDown(this);
        if (available()) transmitter.handleLinkUnavailable(this);
    }

    public double totSensedUpDuration()
    {
        if (sensedUp()) return totSensedUpDuration + (SimEngine.time() - lastSensedChangeTime);
        else return totSensedUpDuration;
    }

    public double totSensedDownDuration()
    {
        if (sensedDown()) return totSensedDownDuration + (SimEngine.time() - lastSensedChangeTime);
        else return totSensedDownDuration;
    }

    /** Event signaling expected end of transmission. */

    private class TransmittedEvent extends Event
    {
        SimLink l;
        Transmission m;
        TransmittedEvent(double delay, SimLink l, Transmission m)
        {
            super(delay);
            this.l = l;
            this.m = m;
        }
        public void handle() {
            l.handleTransmitted(m);
        }
    }


    /** called if transmission occurred through stage */
    public void handleTransmitted_stage(Transmission m){
        handleTransmitted(m);
    }

    /** Called if transmission ends (possibly earlier than expected if link sensed as closed). */
    private void handleTransmitted(Transmission m)
    {
        if (transmitting)
            {
                prepareReception();
                transmitting = false;
                if (!sensedDown)
                    {
                        if (!checkForWaitingTransmissions())
                            {
                                log(2,"link available");
                                transmitter.handleLinkAvailable(this);
                            }
                    }
            }
    }

    /** Called if transmission is terminated (either completed or disrupted). */

    private void prepareReception(Transmission m)
    {
        if (m.initiated())
            {
                m.prepareReception();
                SimEngine.add(new ReceptionEvent(latency,this,m));
                transTerminated += 1;
                sumUtilDuration += m.utilDuration();
                if (m.payload().isKnowledge()) sumStatusUtilDuration += m.utilDuration();
                if (m.payload().isData()) sumDataUtilDuration += m.utilDuration();
                if (m.payload().isBackground()) sumBackgroundUtilDuration += m.utilDuration();
                if (m.payload().isApp()) sumAppUtilDuration += m.utilDuration();
            }
    }

    public double UtilDuration()
    {
        for (Transmission m : intransit)
            {
                if (m.initiated()) return sumUtilDuration + (SimEngine.time() - lastTransmissionTime);
            }
        return sumUtilDuration;
    }

    public double StatusUtilDuration()
    {
        for (Transmission m : intransit)
            {
                if (m.initiated() && m.payload().isKnowledge()) return sumStatusUtilDuration + (SimEngine.time() - lastTransmissionTime);
            }
        return sumStatusUtilDuration;
    }

    public double BackgroundUtilDuration()
    {
        for (Transmission m : intransit)
            {
                if (m.initiated() && m.payload().isBackground()) return sumBackgroundUtilDuration + (SimEngine.time() - lastTransmissionTime);
            }
        return sumBackgroundUtilDuration;
    }

    public double DataUtilDuration()
    {
        for (Transmission m : intransit)
            {
                if (m.initiated() && m.payload().isData()) return sumDataUtilDuration + (SimEngine.time() - lastTransmissionTime);
            }
        return sumDataUtilDuration;
    }

    public double AppUtilDuration()
    {
        for (Transmission m : intransit)
            {
                if (m.initiated() && m.payload().isApp()) return sumAppUtilDuration + (SimEngine.time() - lastTransmissionTime);
            }
        return sumAppUtilDuration;
    }

    private void prepareReception()
    {
        for (Transmission m : intransit) prepareReception(m);
    }

    /** Called to simulate disruption of transmissions in transit. */
    /** We conservatively assume that all messages in transit are affected. */

    private void handleCutOff()
    {
        while (!intransit.isEmpty())
            {
                Transmission m = intransit.peek();
                handleCutOff(m);
            }
    }

    private void handleCutOff(Transmission m)
    {
        assert m.terminated();
        m.handleCutOff();

        if (m.lost())
            {
                intransit.remove(m);
                transLost++;
                transLostCutOff++;
                log(2,"transmission " + m.logRepLong() + " lost due to disruption");
            }
        else
            {
                transDisrupted++;
                transDisruptedCutOff++;
                log(2,"transmission " + m.logRepLong() + " affected by disruption");
                handleReception(m);
            }
    }

    /** Called after transmission has been received to simulate transmission errors */
    /* We conservatively assume that errors will be detected not before the receipt of the transmission. */

    private void handleError(Transmission m)
    {
        assert m.terminated();
        if (m.handleError())
            {
                if (m.lost())
                    {
                        intransit.remove(m);
                        transLost++;
                        transLostError++;
                        log(2,"transmission " + m.logRepLong() + " lost due to transmission error");
                    }
                else
                    {
                        transDisrupted++;
                        transDisruptedError++;
                        log(2,"transmission " + m.logRepLong() + " affected by transmission error");
                    }
            }
    }

    /** Event signaling completion of reception. */

    private class ReceptionEvent extends Event
    {
        SimLink l;
        Transmission m;
        ReceptionEvent(double delay, SimLink l, Transmission m)
        {
            super(delay);
            this.l = l;
            this.m = m;
        }
        public void handle() {
            //          long start = System.nanoTime(); 
            l.handleReception(m);

            /*          long end = System.nanoTime(); 
                        System.out.println("--------- recepti diff " + (end-start)/1000);
                        System.out.flush();
            */
        }
    }

    /** Called if transmission has been received (possibly disrupted). */

    private void handleReception(Transmission m)
    {
        if (!m.received() && !m.lost()) // can happen earlier than scheduled but only once
            {
                assert m.terminated();
                handleError(m);
                if (!m.lost())
                    {
                        intransit.remove(m);
                        transReceived += 1;
                        log(2,"transmission received:" + m.logRepLong());
                        m.handleReception();
                        receiver.handleReception(this,m);
                    }
            }
    }

    /** Request queuing transmission. */

    public void requestTransmission(TransmissionPayload payload)
    {
        Transmission m = new Transmission(this,payload);
        if (waiting.size() <= Param.LinkQueueLength)
            {
                log(2,"queueing transmission " + m.logRepLong());
                waiting.add(m);
                checkForWaitingTransmissions();
            }
        else
            {
                log(2,"dropping transmission " + m.logRepLong() + " - link queue full");
            }
    }

    public boolean checkForWaitingTransmissions()
    {
        if (open() && available())
            {
                Transmission m = waiting.peek();
                if (m != null)
                    {
                        waiting.remove(m);
                        log(2,"transmitting " + m.logRepLong());
                        intransit.add(m);
                        transmitting = true;
                        transmitter.handleLinkUnavailable(this);
                        m.handleInitiation();
                        transInitiated++;
                        double s = m.realSize();
                        transSize += s;
                        lastTransmissionTime = SimEngine.time();
                        double transDelay = m.realSize()/bandwidth;
                        //                long start = System.nanoTime();
                        //System.out.println("calling time: " + System.nanoTime());
                        if(Param.EnableStage && Param.UseStageTransfer && m.link().type() == LinkType.Wireless && Param.UseSimNetwork) 
                            {
                                /* make new wifi message, send to their neighbor */
                                WifiMessageCF wmb = new WifiMessageCF();

                                //wmb.message = new String("TM");
                                if(m.payload() instanceof Bundle) 
                                    {
                                        if(m.payload().isKnowledge())
                                            {
                                                Bundle b = (Bundle)m.payload();
                                                KnowledgePayload st= (KnowledgePayload)b.payload();
                                                wmb.message = new String("S,"+st.knowledge().size());
                                            }
                                    }
                                else
                                    {
                                        wmb.message = new String("Unknown");
                                    }
                                //                  wmb.l = this;
                                wmb.tm = m;
                                wmb.tm_size = s;
                                wmb.tm_cancled = false;
                                int target_node_index = m.link().receiver().ctrl().GetNodeIndex();//target's wifi node index
                                int src_node_index = m.link().transmitter().ctrl().GetNodeIndex();
                                int target_robot_p =SimulationInterface.GetRobotPVector()[ target_node_index ];
                                int src_robot_p = SimulationInterface.GetRobotPVector()[ src_node_index ];

                                wmb.sender_node_index = src_node_index;
                                wmb.recipient_node_index = target_node_index;

                                WifiConfig target_wc = WifiInterface._GetConfig(target_robot_p);
                                wmb.recipient_ip_in = target_wc.ip_in;

                                log(3,"\nTrying.. "+ logRepLong() + " sending trasnmsision to "+target_wc.ip_in+" , Bundle: "+ m.payload().logRepLong()+ " Link: " + m.link().logRepLong()+" time: " + SimEngine.time());
                                WifiInterface._SendUnicastMessage(wmb, src_robot_p);
                                //                  long end = System.nanoTime();
                                //                  Param.sum += (end-start);
                                //                  System.out.println("option 2: reserve Transmission accumulated time: " + Param.sum+", diff: " + (end-start));
                            }
                        else {
                            SimEngine.add(new TransmittedEvent(transDelay,this,m));
                            // log(1, logRepLong() + " - size: " + m.realSize() );
                            // log(1, logRepLong() + " - bandwidth: " + bandwidth );
                            // log(1, logRepLong() + " - transmission delay: " + Log.precision(4,transDelay) );
                            //                  long end = System.nanoTime();
                            //                  Param.sum += (end-start);
                            //                  System.out.println("option 1: reserve Transmission accumulated time: " + Param.sum+", diff: " + (end-start));
                        }

                        /*
                          if(!open())
                          {
                          intransit.remove(m);
                          transLost++;
                          transLostDown++;
                          log(1, "transmission " + m.logRepLong() + " lost (link closed)");
                          return false;
                          }
                        */
                        return true;
                    }
            }
        return false;
    }

    /* Methods for individual link statistics. */

    public int transInTransit()
    {
        return intransit.size();
    }

    public double openCapacity() {
        return openDuration() * bandwidth;
    }
  
    public double capacity() {
        return (openDuration() + closedDuration()) * bandwidth;
    }
  
    public double UtilCapacity() {
        return UtilDuration() * bandwidth;
    }
  
    public double StatusUtilCapacity() {
        return StatusUtilDuration() * bandwidth;
    }
  
    public double DataUtilCapacity() {
        return DataUtilDuration() * bandwidth;
    }
  
    public double BackgroundUtilCapacity() {
        return BackgroundUtilDuration() * bandwidth;
    }
  
    public double AppUtilCapacity() {
        return AppUtilDuration() * bandwidth;
    }

    public void print()
    {
        if(Param.SILENT){ return; }
        SimEngine.rep.println("Link " + logRepLong());
        SimEngine.rep.begin();
        SimEngine.rep.println("Bandwidth: " + Log.precision(1,bandwidth));
        SimEngine.rep.println("Latency: " + Log.precision(4,latency));
        SimEngine.rep.println("ErrorRate: " + errorRate);
        SimEngine.rep.println("Available: " + available());
        SimEngine.rep.println("Up: " + open());
        SimEngine.rep.println("SensedUp: " + sensedUp());
        SimEngine.rep.println("SensedDown: " + sensedDown());
        SimEngine.rep.println("InTransit: ");
        SimEngine.rep.begin();
        intransit.print();
        SimEngine.rep.end();
        SimEngine.rep.println("LastTransmissionTime: " + Log.tf.format(lastTransmissionTime));
        SimEngine.rep.println("Changes: " + changes);
        SimEngine.rep.println("LastChangeTime: " + Log.tf.format(lastChangeTime));
        SimEngine.rep.println("TransInTransit: " + transInTransit());
        SimEngine.rep.println("TransInitiated: " + transInitiated);
        SimEngine.rep.println("TransRejectedBusy: " + transRejectedBusy);
        SimEngine.rep.println("TransRejectedSensedDown: " + transRejectedSensedDown);
        SimEngine.rep.println("TransTerminated: " + transTerminated  +
                              " (" + Log.percentage(transTerminated,transInitiated) + ")");
        SimEngine.rep.println("TransReceived: " + transReceived  +
                              " (" + Log.percentage(transReceived,transInitiated) + ")");
        SimEngine.rep.println("TransDisrupted: " + transDisrupted +
                              " (" + Log.percentage(transDisrupted,transInitiated) + ")");
        SimEngine.rep.println("TransDisruptedCutOff: " + transDisruptedCutOff +
                              " (" + Log.percentage(transDisruptedCutOff,transInitiated) + ")");
        SimEngine.rep.println("TransDisruptedError: " + transDisruptedError +
                              " (" + Log.percentage(transDisruptedError,transInitiated) + ")");
        SimEngine.rep.println("TransLost: " + transLost +
                              " (" + Log.percentage(transLost,transInitiated) + ")");
        SimEngine.rep.println("TransLostDown: " + transLostDown +
                              " (" + Log.percentage(transLostDown,transInitiated) + ")");
        SimEngine.rep.println("TransLostCutOff: " + transLostCutOff +
                              " (" + Log.percentage(transLostCutOff,transInitiated) + ")");
        SimEngine.rep.println("TransLostError: " + transLostError +
                              " (" + Log.percentage(transLostError,transInitiated) + ")");
        SimEngine.rep.println("AvgTransSize: " + Log.fraction(0,transSize,transInitiated));
        SimEngine.rep.println("OpenDuration: " + Log.tf.format(openDuration()) +
                              " (" + Log.percentage(openDuration(),openDuration()+closedDuration()) + ")");
        SimEngine.rep.println("ClosedDuration: " + Log.tf.format(closedDuration()) +
                              " (" + Log.percentage(closedDuration(),openDuration()+closedDuration()) + ")");
        SimEngine.rep.println("UtilDuration: " + Log.tf.format(UtilDuration()));
        SimEngine.rep.println("StatusUtilDuration: " + Log.tf.format(StatusUtilDuration()));
        SimEngine.rep.println("DataUtilDuration: " + Log.tf.format(DataUtilDuration()));
        SimEngine.rep.println("BackgroundUtilDuration: " + Log.tf.format(BackgroundUtilDuration()));
        SimEngine.rep.println("AppUtilDuration: " + Log.tf.format(AppUtilDuration()));
        SimEngine.rep.println("OpenCapacity: " + Log.precision(0,openCapacity()));
        SimEngine.rep.println("UtilCapacity: " + Log.precision(0,UtilCapacity()) +
                              " (" + Log.percentage(UtilCapacity(),openCapacity()) + ")");
        SimEngine.rep.println("StatusUtilCapacity: " + Log.precision(0,StatusUtilCapacity()) +
                              " (" + Log.percentage(StatusUtilCapacity(),openCapacity()) + ")");
        SimEngine.rep.println("DataUtilCapacity: " + Log.precision(0,DataUtilCapacity()) +
                              " (" + Log.percentage(DataUtilCapacity(),openCapacity()) + ")");
        SimEngine.rep.println("BackgroundUtilCapacity: " + Log.precision(0,BackgroundUtilCapacity()) +
                              " (" + Log.percentage(BackgroundUtilCapacity(),openCapacity()) + ")");
        SimEngine.rep.println("AppUtilCapacity: " + Log.precision(0,AppUtilCapacity()) +
                              " (" + Log.percentage(AppUtilCapacity(),openCapacity()) + ")");
        SimEngine.rep.println("TotSensedChanges: " + totSensedChanges);
        SimEngine.rep.println("LastSensedChangeTime: " + Log.tf.format(lastSensedChangeTime));
        SimEngine.rep.println("LastSensedUpEvent: " + Log.tf.format(lastSensedUpEvent));
        SimEngine.rep.println("TotSensedUpDuration: " + Log.tf.format(totSensedUpDuration()));
        SimEngine.rep.println("LastSensedDownEvent: " + Log.tf.format(lastSensedDownEvent));
        SimEngine.rep.println("TotSensedDownDuration: " + Log.tf.format(totSensedDownDuration()));
        SimEngine.rep.end();
    }

    public static void printAll(SimWorld w)
    {
        for (SimLink l : w.allLinks()) {
            SimEngine.rep.println("");
            l.print();
        }
    }

    /* methods for global link statistics */

    public static int totLinksUp(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.open()?1:0;
        return r;
    }

    public static int totLinksAvailable(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.available()?1:0;
        return r;
    }

    public static int totTransInTransit(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transInTransit();
        return r;
    }

    public static int totTransInitiated(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transInitiated;
        return r;
    }

    public static int totTransRejectedBusy(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transRejectedBusy;
        return r;
    }

    public static int totTransRejectedSensedDown(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transRejectedSensedDown;
        return r;
    }

    public static int totTransLostDown(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transLostDown;
        return r;
    }

    public static int totTransTerminated(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transTerminated;
        return r;
    }

    public static int totTransReceived(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transReceived;
        return r;
    }

    public static int totTransDisrupted(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transDisrupted;
        return r;
    }

    public static int totTransDisruptedCutOff(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transDisruptedCutOff;
        return r;
    }

    public static int totTransDisruptedError(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transDisruptedError;
        return r;
    }

    public static int totTransLost(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transLost;
        return r;
    }

    public static int totTransLostCutOff(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transLostCutOff;
        return r;
    }

    public static int totTransLostError(SimWorld w)
    {
        int r = 0;
        for (SimLink l : w.allLinks()) r += l.transLostError;
        return r;
    }

    public static double totTransSize(SimWorld w)
    {
        double r = 0;
        for (SimLink l : w.allLinks()) r += l.transSize;
        return r;
    }

    public static double totOpenCapacity(SimWorld w)
    {
        double r = 0;
        for (SimLink l : w.allLinks()) r += l.openCapacity();
        return r;
    }

    public static double totCapacity(SimWorld w)
    {
        double r = 0;
        for (SimLink l : w.allLinks()) r += l.capacity();
        return r;
    }

    public static double totUtilCapacity(SimWorld w)
    {
        double r = 0;
        for (SimLink l : w.allLinks()) r += l.UtilCapacity();
        return r;
    }

    public static double totStatusUtilCapacity(SimWorld w)
    {
        double r = 0;
        for (SimLink l : w.allLinks()) r += l.StatusUtilCapacity();
        return r;
    }

    public static double totDataUtilCapacity(SimWorld w)
    {
        double r = 0;
        for (SimLink l : w.allLinks()) r += l.DataUtilCapacity();
        return r;
    }

    public static double totBackgroundUtilCapacity(SimWorld w)
    {
        double r = 0;
        for (SimLink l : w.allLinks()) r += l.BackgroundUtilCapacity();
        return r;
    }

    public static double totAppUtilCapacity(SimWorld w)
    {
        double r = 0;
        for (SimLink l : w.allLinks()) r += l.AppUtilCapacity();
        return r;
    }

    public static void printTotals(SimWorld w)
    {
        if(Param.SILENT){ return; }
        SimEngine.rep.begin();
        SimEngine.rep.println("TotalLinks: " + w.totLinks());
        SimEngine.rep.println("TotalLinksUp: " + totLinksUp(w));
        SimEngine.rep.println("TotalLinksAvailable: " + totLinksAvailable(w));
        SimEngine.rep.println("TotalTransInTransit: " + totTransInTransit(w));
        SimEngine.rep.println("TotalTransInitiated: " + totTransInitiated(w));
        SimEngine.rep.println("TotalTransRejectedBusy: " + totTransRejectedBusy(w));
        SimEngine.rep.println("TotalTransRejectedSensedDown: " + totTransRejectedSensedDown(w));
        SimEngine.rep.println("TotalTransTerminated: " + totTransTerminated(w) +
                              " (" + Log.percentage(totTransTerminated(w),totTransInitiated(w)) + ")");
        SimEngine.rep.println("TotalTransReceived: " + totTransReceived(w) +
                              " (" + Log.percentage(totTransReceived(w),totTransInitiated(w)) + ")");
        SimEngine.rep.println("TotalTransDisrupted: " + totTransDisrupted(w) +
                              " (" + Log.percentage(totTransDisrupted(w),totTransInitiated(w)) + ")");
        SimEngine.rep.println("TotalTransDisruptedCutOff: " + totTransDisruptedCutOff(w) +
                              " (" + Log.percentage(totTransDisruptedCutOff(w),totTransInitiated(w)) + ")");
        SimEngine.rep.println("TotalTransDisruptedError: " + totTransDisruptedError(w) +
                              " (" + Log.percentage(totTransDisruptedError(w),totTransInitiated(w)) + ")");
        SimEngine.rep.println("TotalTransLost: " + totTransLost(w) +
                              " (" + Log.percentage(totTransLost(w),totTransInitiated(w)) + ")");
        SimEngine.rep.println("TotalTransLostDown: " + totTransLostDown(w) +
                              " (" + Log.percentage(totTransLostDown(w),totTransInitiated(w)) + ")");
        SimEngine.rep.println("TotalTransLostCutOff: " + totTransLostCutOff(w) +
                              " (" + Log.percentage(totTransLostCutOff(w),totTransInitiated(w)) + ")");
        SimEngine.rep.println("TotalTransLostError: " + totTransLostError(w) +
                              " (" + Log.percentage(totTransLostError(w),totTransInitiated(w)) + ")");
        SimEngine.rep.println("TotalAvgTransSize: " + Log.fraction(0,totTransSize(w),totTransInitiated(w)));
        SimEngine.rep.println("TotalOpenCapacity: " + Log.precision(0,totOpenCapacity(w)) +
                              " (" + Log.percentage(totOpenCapacity(w),totCapacity(w)) + ")");
        SimEngine.rep.println("TotalUtilCapacity: " + Log.precision(0,totUtilCapacity(w)) +
                              " (" + Log.percentage(totUtilCapacity(w),totOpenCapacity(w)) + ")");
        SimEngine.rep.println("TotalStatusUtilCapacity: " + Log.precision(0,totStatusUtilCapacity(w)) +
                              " (" + Log.percentage(totStatusUtilCapacity(w),totOpenCapacity(w)) + ")");
        SimEngine.rep.println("TotalDataUtilCapacity: " + Log.precision(0,totDataUtilCapacity(w)) +
                              " (" + Log.percentage(totDataUtilCapacity(w),totOpenCapacity(w)) + ")");
        SimEngine.rep.println("TotalBackgroundUtilCapacity: " + Log.precision(0,totBackgroundUtilCapacity(w)) +
                              " (" + Log.percentage(totBackgroundUtilCapacity(w),totOpenCapacity(w)) + ")");
        SimEngine.rep.println("TotalAppUtilCapacity: " + Log.precision(0,totAppUtilCapacity(w)) +
                              " (" + Log.percentage(totAppUtilCapacity(w),totOpenCapacity(w)) + ")");
        SimEngine.rep.end();
    }
}
