/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim.dtn;

import cf.common.Param;
import cf.sim.SimEngine;
import cf.util.Log;

enum TransStatus { Created, Initiated, Terminated, Lost, Received };

public class Transmission
{
  /** Link used for transmission. */

  private SimLink link;

  /** Transmission payload: currently either a bundle or a status update. */

  public TransmissionPayload payload;

  /* Transmission status. */

  private TransStatus status;

  /** (Expected) time of events. */

  private double timeCreated, timeBeginTransmission, timeEndTransmission, timeBeginReception, timeEndReception;

  /* Observers. */
  private  boolean error_report_from_stage;
  public void SetError() { error_report_from_stage = true; }

  public SimLink link() {
    return link;
  }
  public TransmissionPayload payload() {
    return payload;
  }

  public boolean initiated() {
    return status == TransStatus.Initiated;
  }
  public boolean terminated() {
    return status == TransStatus.Terminated;
  }
  public boolean received() {
    return status == TransStatus.Received;
  }

  public boolean lost() {
    return payload.lost();
  }

  /* Constructor. */

  public Transmission(SimLink link, TransmissionPayload payload)
  {
    this.link = link;
    this.payload = payload;
    this.timeCreated = SimEngine.time();
    this.status = TransStatus.Created;
    this.error_report_from_stage = false;
  }

  /** Informative abbreviation for transmission. */

  public String logRepShort() {
      if(Param.SILENT){ return ""; }
      return Log.getClassName(this) + "(" + payload.logRepShort() + ") : " + link().logRepShort();
  }

  public String logRepLong() {
      if(Param.SILENT){ return ""; }
      return Log.getClassName(this) + "(" + payload.logRepLong() + ") : " + link().logRepLong();
  }

  /** Size of correponding real transmission in bytes (model of messages here is an abstraction of reality). */

  public double realSize() {
    return payload.realSize();
  }

  /** Called at begining of transmission. */

  public void handleInitiation()
  {
    status = TransStatus.Initiated;
    timeBeginTransmission = SimEngine.time();
  }

  /** Called at end of transmission. */

  public void prepareReception()
  {
    status = TransStatus.Terminated;
    timeEndTransmission = SimEngine.time();
    timeBeginReception = timeBeginTransmission + link.latency();
    timeEndReception = timeEndTransmission + link.latency();
  }

  /** Called if transmission is cut off. */

  public boolean handleCutOff()
  {
    if (timeEndTransmission > timeBeginReception)
    {
      double cutOffSize = (timeEndTransmission - timeBeginReception) * link.bandwidth();
      payload.handleCutOff(cutOffSize);
      return true;
    }
    return false;
  }

  /** Called if transmission error is encountered. */

  public boolean handleError()
  {
    // using stage's SNR thresh hold model
    if(Param.EnableStage && Param.UseStageTransfer && link().type() == LinkType.Wireless) 
    {
      if(error_report_from_stage) 
      {
        payload.handleError();
        return true;
      }
      return false;
    }
    else 
    {
      if (!SimEngine.dist.sampleBerWithP(Math.pow(1 - link.errorRate(), realSize())))
      {
        payload.handleError();
        return true;
      }
      return false;
    }
  }

  /** Called if transmission has been received. */

  public void handleReception()
  {
    status = TransStatus.Received;
    timeEndReception = SimEngine.time();
  }

  /* Statistics */

  /** Duration of medium utilization. */

  public double utilDuration() {
    return timeEndTransmission - timeBeginTransmission;
  }

  /** Print transmission. */

  public void print() {
      if(Param.SILENT){ return; }
      payload.print();
  }
}