/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim.pn;

import java.util.ListIterator;

import cf.common.Event;
import cf.sim.SimEngine;
import cf.sim.SimNode;
import cf.util.Log;

enum PlaceDist { Det, Exp }

/* Places annotated with distribution of delays - possibly carrying tokens. */

public class Place
{
  private int index;
  private String name;

  private PlaceDist dist;
  private double delay;

  private TokenList committedMarking;
  private TokenList uncommittedMarking;

  private SimMobility sm;

  public TokenList marking() {
    return committedMarking;
  }

  public Place(SimMobility sm, int index, String name, PlaceDist dist, double delay)
  {
    this.sm = sm;
    this.index = index;
    this.name = name;
    this.committedMarking = new TokenList();
    this.dist = dist;
    this.delay = delay;
  }

  /* Check committed state of place. */

  public boolean isEmpty()
  {
    return committedMarking.isEmpty();
  }

  public boolean contains(SimNode n)
  {
    for (Token t : committedMarking)
    {
      if (t.node() == n) return true;
    }
    return false;
  }

  /* Begin transaction. */

  public void begin()
  {
    uncommittedMarking = new TokenList();
    for (Token t : committedMarking) uncommittedMarking.add(t);
  }

  /* Commit transaction. */

  public void commit()
  {
    committedMarking = uncommittedMarking;
  }

  /* Check uncomitted state of place. */

  public boolean available(SimNode n)
  {
    for (Token t : uncommittedMarking)
    {
      if (t.available() && t.node() == n) return true;
    }
    return false;
  }

  /* Modify uncommited state of place - add a token. */

  public void add(SimNode n)
  {
    double actualDelay = 0.0;
    switch (dist)
    {
      case Det:
        actualDelay = delay;
        break;
      case Exp:
        actualDelay = SimEngine.dist.sampleExpWithMean(delay);
        break;
    }
    uncommittedMarking.add(new Token(actualDelay,n));
    SimEngine.add(new TokenAvailableEvent(actualDelay,this,n));
  }

  /* Modify uncommited state of place - remove a token. */

  public boolean remove(SimNode n)
  {
    ListIterator<Token> li = uncommittedMarking.listIterator();
    while (li.hasNext())
    {
      Token t = li.next();
      if (t.available() && t.node() == n)
      {
        li.remove();
        return true;
      }
    }
    return false;
  }

  /* Event signalling availability of delayed tokens. */

  private class TokenAvailableEvent extends Event
  {
    Place p;
    SimNode n;
    TokenAvailableEvent(double delay, Place p, SimNode n)
    {
      super(delay);
      this.p = p;
      this.n = n;
    }
    public void handle() {
      handleTokenAvailable(p,n);
    }
  }

  /* Need to be handled at higher level. */

  private void handleTokenAvailable(Place p, SimNode n)
  {
    // SimEngine.log.println(0, "TokenAvailableEvent - p = " + p.name() + " n = " + n.name());
    sm.handleTokenAvailable(p,n);
  }

  public String name()
  {
    if (name == null) return "" + index;
    else return name;
  }

  public String logRepShort()
  {
    return "P(" + name() + ")";
  }

  public String logRepLong()
  {
    return logRepShort();
  }

  public String markingToString()
  {
    return name() + "[" + committedMarking.toString() + "]";
  }

  public void print()
  {
    SimEngine.rep.println("Place " + logRepShort());
    SimEngine.rep.begin();
    SimEngine.rep.println("Delay: " + Log.precision(2,delay));
    SimEngine.rep.println("Marking: " + (isEmpty() ? "" : markingToString()));
    SimEngine.rep.end();
  }

}

