/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim.pn;

import cf.common.Node;
import cf.common.NodeList;
import cf.sim.SimEngine;
import cf.sim.SimNode;
import cf.sim.SimWorld;
import cf.sim.dtn.LinkType;
import cf.sim.dtn.SimLink;

/*

   Physical resource schedules are part of the global network simulation
   model.  They are an abstract model of a logical or physical topology
   defined by a special class of Petri nets with individual tokens
   (representing network nodes), delays associated to places, and links
   associated to pairs of places. The Petri net can then be regarded as a
   schedule for its associated links. A link is open iff its transmitter
   and receiver are located at the places corresponding to its end points.

   A lot of room exists for improving the efficiency, but for the small
   sample networks the following code should be sufficient.

 */

/** Abstract mobility model as annotated Petri net. */

public class SimMobility
{
  SimWorld world;

  private PlaceList places;
  private TransitionList transitions;

  private SyncGroupList syncgroups;
  private ContactList contacts;

  public PlaceList places() { return places; }
  public TransitionList transitions() { return transitions; }
  public ContactList contacts() { return contacts; }

  private int freshPlaceIndex;
  private int freshPlaceIndex() {
    return freshPlaceIndex++;
  }

  private int freshTransitionIndex;
  private int freshTransitionIndex() {
    return freshTransitionIndex++;
  }

  public SimMobility(SimWorld world)
  {
    this.world = world;
    places = new PlaceList();
    transitions = new TransitionList();
    contacts = new ContactList();
    syncgroups = new SyncGroupList();
    world.setSimMobility(this);
  }

  /* Add place with neg. exponential delay distribution. */

  public Place addDelayedPlace(String name, double delay)
  {
    // if(Param.Deterministic) return addDetPlace(name,delay); else return addExpPlace(name,delay);
    // return addDetPlace(name,delay);
    return addExpPlace(name,delay);
  }

  public Place addExpPlace(String name, double delay)
  {
    Place p = new Place(this,freshPlaceIndex(),name,PlaceDist.Exp,delay);
    places.add(p);
    return p;
  }

  /* Add place with deterministic delay. */

  public Place addDetPlace(String name, double delay)
  {
    Place p = new Place(this,freshPlaceIndex(),name,PlaceDist.Det,delay);
    places.add(p);
    return p;
  }

  /* Add place without delay. */

  public Place addPlace(String name)
  {
    Place p = new Place(this,freshPlaceIndex(),name,PlaceDist.Det,0.0);
    places.add(p);
    return p;
  }

  /* Add token, i.e. node, to given place. */

  public void addToken(Place p, SimNode n)
  {
    p.begin();
    p.add(n);
    p.commit();

    NodeList c = getChildNodes(n);
    for(Node child : c)
    {
      //System.out.println("n- "+ n.logRepShort() + "'s child : "+child.logRepShort());
      addToken(p , (SimNode)child);
    }
  }

  /* add node in child or parent */
  private NodeList getChildNodes(SimNode n)
  {
    NodeList child_node_list = new NodeList();
    for(Node node : world.allNodes() )
    {
      SimNode parent = (SimNode)(node.parent());
      if(parent != null && parent != node)
      {
        if( parent == n) // found child
        {
          //System.out.println("N - " + node.logRepShort() + " , NP - " + parent.logRepShort());
          child_node_list.add( node);
        }
      }
    }
    return child_node_list;
  }

  public Place getPlace(SimNode node)
  {
    for(Place p : places)
    {
      if( p.contains(node) )
        return p;
    }
    return null;
  }

  public Place getPlace(String name)
  {
    for(Place p : places)
    {
      if( p.name().equals( name ) )
      {
        return  p;
      }
    }
    return null;
  }

  public boolean fire(SimNode node, Place before, Place after)
  {
    Transition t = available(before, after);
    if(t == null) 
      return false;
    else
    {
      SimEngine.log.println(0, "SimMobility - firing " + t.logRepShort());

      before.begin();
      after.begin();

      fire( t, node );

      before.commit();
      after.commit();

      handleChange();

      return true;
    }
  }

  void fire(Transition t, SimNode n)
  {
    t.fire(n);
    NodeList c = getChildNodes(n);
    for(Node child : c)
    {
      fire(t, (SimNode)child);
    }
  }

  public Transition available(Place before, Place after)
  {
    for(Transition t : transitions)
    {
      PlaceList preset = t.preset();
      PlaceList postset = t.postset();

      for(Place p : preset)
      {
        if( p == before)
        {
          for(Place q : postset)
          {
            if ( q == after )
              return t;
          }
        }
      }
    }
    return null;
  }
  public Transition addTransition(PlaceList preset, String name, PlaceList postset)
  {
    Transition t = new Transition(this,freshTransitionIndex(),preset,name,postset);
    transitions.add(t);
    return t;
  }

  public Transition addTransition(Place p0, String name, Place q0)
  {
    PlaceList preset = new PlaceList();
    preset.add(p0);
    PlaceList postset = new PlaceList();
    postset.add(q0);
    return addTransition(preset,name,postset);
  }


  /* Add transitions with various input/output cardinalities. */

  public Transition addSyncTransition(PlaceList preset, String name, PlaceList postset)
  {
    Transition t = new Transition(this,freshTransitionIndex(),preset,name,postset);
    transitions.add(t);
    return t;
  }

  public Transition addSyncTransition(Place p0, String name, Place q0)
  {
    PlaceList preset = new PlaceList();
    preset.add(p0);
    PlaceList postset = new PlaceList();
    postset.add(q0);
    return addSyncTransition(preset,name,postset);
  }

  public Transition addSyncTransition(Place p0, Place p1, String name, Place q0)
  {
    PlaceList preset = new PlaceList();
    preset.add(p0);
    preset.add(p1);
    PlaceList postset = new PlaceList();
    postset.add(q0);
    return addSyncTransition(preset,name,postset);
  }

  public Transition addSyncTransition(Place p0, String name, Place q0, Place q1)
  {
    PlaceList preset = new PlaceList();
    preset.add(p0);
    PlaceList postset = new PlaceList();
    postset.add(q0);
    postset.add(q1);
    return addSyncTransition(preset,name,postset);
  }

  public Transition addSyncTransition(Place p0, Place p1, String name, Place q0, Place q1)
  {
    PlaceList preset = new PlaceList();
    preset.add(p0);
    preset.add(p1);
    PlaceList postset = new PlaceList();
    postset.add(q0);
    postset.add(q1);
    return addSyncTransition(preset,name,postset);
  }

  public void addSyncGroup(TransitionList group)
  {
    syncgroups.add(new SyncGroup(this,group));
  }

  public void addSyncGroup(Transition t)
  {
    TransitionList group = new TransitionList();
    group.add(t);
    addSyncGroup(group);
  }

  public void addSyncGroup(Transition t1, Transition t2)
  {
    TransitionList group = new TransitionList();
    group.add(t1);
    group.add(t2);
    addSyncGroup(group);
  }

  public Transition addLocalTransition(PlaceList preset, String name, PlaceList postset)
  {
    Transition t = addSyncTransition(preset,name,postset);
    addSyncGroup(t);
    return t;
  }

  public Transition addLocalTransition(Place p0, String name, Place q0)
  {
    PlaceList preset = new PlaceList();
    preset.add(p0);
    PlaceList postset = new PlaceList();
    postset.add(q0);
    return addLocalTransition(preset,name,postset);
  }

  public Transition addLocalTransition(Place p0, Place p1, String name, Place q0)
  {
    PlaceList preset = new PlaceList();
    preset.add(p0);
    preset.add(p1);
    PlaceList postset = new PlaceList();
    postset.add(q0);
    return addLocalTransition(preset,name,postset);
  }

  public Transition addLocalTransition(Place p0, String name, Place q0, Place q1)
  {
    PlaceList preset = new PlaceList();
    preset.add(p0);
    PlaceList postset = new PlaceList();
    postset.add(q0);
    postset.add(q1);
    return addLocalTransition(preset,name,postset);
  }

  public Transition addLocalTransition(Place p0, Place p1, String name, Place q0, Place q1)
  {
    PlaceList preset = new PlaceList();
    preset.add(p0);
    preset.add(p1);
    PlaceList postset = new PlaceList();
    postset.add(q0);
    postset.add(q1);
    return addLocalTransition(preset,name,postset);
  }

  /* Add a scheduled contact. */

  public void addContact(SimLink link, Place transmitter, Place receiver)
  {
    contacts.add(new Contact(link,transmitter,receiver));
  }

  /* Add a symmetric scheduled contact. */

  public void addSymmetricContact(SimLink link, Place transmitter, Place receiver)
  {
    contacts.add(new Contact(link,transmitter,receiver));
    assert link.reverse() != null;
    contacts.add(new Contact(link.reverse(),receiver,transmitter));
  }

  /* Used only internally to handle change of marking. */

  public void handleChange()
  {
    SimEngine.log.println(0, "SimMobility - new marking: " + places.markingToString());
    updateLinks();
  }

  /* Used only internally to handle new token availability. */

  public void handleTokenAvailable(Place p, SimNode n)
  {
    fireNextSyncGroup();
  }

  /* Check if link should be opened according to any of the contacts. */

  private boolean open(SimLink l)
  {
    for (Contact c : contacts)
    {
      if (c.open(l)) return true;
    }
    return false;
  }

  /* Update link state according to this function. */

  private void updateLinks()
  {
    for (SimLink l : world.allLinks())
    {
      if(l.type() == LinkType.ParentChild)
      {
        return;
      }
      if (open(l))
      {
        //System.out.println("open");
        if (!l.open()) world.handleLinkChange(true,l);
      }
      else
      {
        //System.out.println("close");
        //SimEngine.log.println("LINK: " + l.logRepShort());
        if (l.open()) world.handleLinkChange(false,l);
      }
    }
  }

  /* Count number of enabled groups. */

  public int enabledSyncGroups()
  {
    int r = 0;
    for (SyncGroup s : syncgroups)
    {
      if (s.enabled() != null) r++;
    }
    return r;
  }

  /* Select i-th enabled group. */

  public SyncGroup ithEnabledSyncGroup(int i)
  {
    int r = 0;
    for (SyncGroup s : syncgroups)
    {
      if (s.enabled() != null)
      {
        if (r == i) return s;
        r++;
      }
    }
    return null;
  }

  /* Fire next group using a two level probabilistic selection to ensure fairness. */

  private boolean fireNextSyncGroup()
  {
    int enabledGroups = enabledSyncGroups();
    if (enabledGroups > 0)
    {
      SyncGroup s = ithEnabledSyncGroup(SimEngine.dist.sampleUniform(0,enabledGroups));
      int enabledBindings = s.enabledBindings();
      NodeList b = s.ithEnabledBinding(SimEngine.dist.sampleUniform(0,enabledBindings));
      if (b != null) {
        s.fire(b);
        return true;
      }
    }
    return false;
  }

  /* Invoke resource schedule. */

  public void start()
  {
    SimEngine.log.println(0, "SimMobility - initial marking: " + places.markingToString());
    updateLinks();
    fireNextSyncGroup();
  }

  public void setup()
  {
    SimEngine.log.println(0, "SimMobility - initial marking: " + places.markingToString());
    updateLinks();
  }

  public void print()
  {
    SimEngine.rep.println("Places:");
    SimEngine.rep.begin();
    places.print(true);
    SimEngine.rep.end();
    SimEngine.rep.println("Transitions:");
    SimEngine.rep.begin();
    transitions.print(true);
    SimEngine.rep.end();
    SimEngine.rep.println("Synchronized Transition Groups:");
    SimEngine.rep.begin();
    syncgroups.print();
    SimEngine.rep.end();
    SimEngine.rep.println("Contacts:");
    SimEngine.rep.begin();
    contacts.print();
    SimEngine.rep.end();
  }
}
