/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim.pn;

import cf.common.NodeList;
import cf.sim.SimEngine;
import cf.sim.SimNode;

class SyncGroup
{
  /* Transitions to be synchronized. */

  private TransitionList group;

  public TransitionList group() {
    return group;
  }

  private SimMobility sm;

  /* Combined preset and postset. */

  private PlaceList preset, postset;

  SyncGroup(SimMobility sm, TransitionList group)
  {
    this.sm = sm;
    this.group = group;
    preset = new PlaceList();
    postset = new PlaceList();
    for (Transition t : group)
    {
      preset.addAll(t.preset()); // should remove duplicates
      postset.addAll(t.postset());
    }
  }

  /* Bindings, i.e. lists of bindings for underlying transitions, can be enumerated. */

  public NodeList firstBinding()
  {
    int len = group.size();
    NodeList binding = new NodeList();
    for (int i = 0; i < len; i++) binding.add(sm.world.firstNode());
    return binding;
  }

  public NodeList nextBinding(NodeList binding)
  {
    int len = binding.size();
    for (int i = 0; i < len; i++)
    {
      SimNode n = (SimNode)binding.get(i);
      SimNode nn = sm.world.nextNode(n);
      if (nn == null)
      {
        binding.set(i,sm.world.firstNode());
      }
      else
      {
        binding.set(i,nn);
        return binding;
      }
    }
    return null;
  }

  /* Is group enabled for given binding ? */

  public boolean enabled(NodeList binding)
  {
    for (Place p : preset) p.begin();

    int i = 0;
    for (Transition t : group)
    {
      SimNode n = (SimNode)binding.get(i);
      if (!t.activate(n)) return false;
      i++;
    }
    return true;
  }

  /* Is group enabled for some binding ? */

  public NodeList enabled()
  {
    NodeList binding = firstBinding();
    while (binding != null)
    {
      if (enabled(binding)) return binding;
      binding = nextBinding(binding);
    }
    return null;
  }

  /* Count number of enabled bindings. */

  public int enabledBindings()
  {
    int r = 0;
    NodeList binding = firstBinding();
    while (binding != null)
    {
      if (enabled(binding)) r++;
      binding = nextBinding(binding);
    }
    return r;
  }

  /* Get i-th enabled binding in above ordering. */

  public NodeList ithEnabledBinding(int i)
  {
    int r = 0;
    NodeList binding = firstBinding();
    while (binding != null)
    {
      if (enabled(binding))
      {
        if (r == i) return binding;
        r++;
      }
      binding = nextBinding(binding);
    }
    return null;
  }

  /* Fire group under the given binding. */

  public void fire(NodeList binding)
  {
    SimEngine.log.println(0, "SimMobility - firing " + group.toString());

    for (Place p : preset) p.begin();
    for (Place p : postset) p.begin();

    int i = 0;
    for (Transition t : group)
    {
      SimNode n = (SimNode)binding.get(i);
      t.fire(n);
      i++;
    }

    for (Place p : preset) p.commit();
    for (Place p : postset) p.commit();

    sm.handleChange();
  }

  public void print()
  {
    SimEngine.rep.println(group.toString());
  }
}