/*

  Copyright 2010-2013 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim.pn.devices;

import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.devices.PositionApplication;
import cf.api.devices.PositionMobility;
import cf.common.Event;
import cf.common.Model;
import cf.common.Node;
import cf.common.Param;
import cf.sim.SimNode;
import cf.sim.SimWorld;
import cf.sim.pn.Place;
import cf.sim.pn.SimMobility;
import cf.util.Log;

public class SimPositionMobility extends Model implements PositionMobility
{
  public SimPositionMobility(SimWorld world, Model parent, String name)
  {
    super(world, parent, name);
    current_place = null;
    this.world = world;
  }

  public SimPositionMobility(SimWorld world, Model parent)
  {
    this(world, parent, null);
    this.world = world;
  }

  public SimPositionMobility(SimWorld world)
  {
    this(world, null, null);
    this.world = world;
  }

  SimWorld world;

  public String logRepLong()
  {
    if(current_place!=null)
      return super.logRepLong() + "PositionMobility: " + "("+ current_place.logRepShort() + ")";
    else
      return super.logRepLong() + "PositionMobility: " + "("+ "?" + ")";
  }

  public void addApplication(PositionApplication app)
  {
    addApplication( (CyberApplication) app );
  }

  private Place current_place;

  public Place getPlace()
  {
    if(current_place != null)
      return current_place;

    SimNode node = getTopParentNode();
    SimMobility sm = world.sm();
    Place p = sm.getPlace(node);

    if(p != null)
    {
      current_place = p;
      return p;
    }
    else
    {
      return null;
    }
  }

  private SimNode getTopParentNode()
  {
    SimNode node;
    for(CyberApplication app : allApplications())
    {
      if( app instanceof CyberApp)
      {
        CyberApp ca = (CyberApp) app;
        node = (SimNode)ca.node;
        return getTopParent(node);
      }
    }
    return null;
  }

  private SimNode getTopParent(SimNode node)
  {
    if ( node.parent() != null &&  node.parent() instanceof SimNode)
      return getTopParent( (SimNode)node.parent() ) ;
    else
      return node;
  }


  private MoveEvent lastMoveEvent;

  class MoveEvent extends Event
  {
    SimPositionMobility position_mobility;
    Place p_togo;
    SimNode n;

    MoveEvent(SimPositionMobility position_mobility, double delay, SimNode n, Place p_togo)
    {
      super(delay);
      this.position_mobility = position_mobility;
      this.p_togo = p_togo;
      this.n = n;
    }

    public void handle()
    {	    
      position_mobility.handleMoveEvent(this);
    }
  }

  private void handleMoveEvent(MoveEvent e)
  {
    log("completing move");

    Place current_place = getPlace();
    Place togo_place = e.p_togo;

    if( !world.sm().fire(e.n,  current_place , togo_place ) )
    {
      Log.stderr.println( "there's no transition from " + current_place.logRepShort() + " to " + togo_place.logRepShort() );
      log( "there's no transition from " + current_place.logRepShort() + " to " + togo_place.logRepShort() );
    }

    this.current_place = togo_place;

    for(CyberApplication app : allApplications())
    {
      if(app instanceof PositionApplication) {
        ((PositionApplication)app).handlePositionUpdate();
      }
    }
    lastMoveEvent = null;
  }

  public Place move(String placeToGo)
  {
    log("initiating move");

    SimNode node = getTopParentNode();
    Place current_place = getPlace();
    Place togo_place = world.sm().getPlace(placeToGo);

    if( current_place == null )
    {
      Log.stderr.println( node.logRepShort() + " has no place ");
      log( node.logRepShort() + " has no place ");
      return null;
    }

    if( togo_place == null )
    {
      Log.stderr.println( "there's no place whose name is " + placeToGo );
      log( "there's no place whose name is " + placeToGo );
      return null;
    }

    if( current_place == togo_place )
    {
      Log.stderr.println( "try to move to same place " + placeToGo );
      log( "try to move to same place " + placeToGo );
      return null;
    }

    if( world.sm().available(current_place , togo_place ) == null )
    {
      Log.stderr.println( "there's no transition from " + current_place.logRepShort() + " to " + togo_place.logRepShort() );
      log( "there's no transition from " + current_place.logRepShort() + " to " + togo_place.logRepShort() );
      return null;
    }

    if(lastMoveEvent != null && lastMoveEvent.pending()) 
    {
      //lastMoveEvent.cancel();
      //log("cancelling previous move");
      log("there's un-finished move");
      return null;
    }

    lastMoveEvent = new MoveEvent(this, Param.MobilityTransitionDelay , node, togo_place);
    Node my_node = getNode();
    if(my_node != null)
      my_node.postEvent(lastMoveEvent);
    else 
      return null;

    return togo_place;
  }
  
  public void stop()
  {
    if(lastMoveEvent != null && lastMoveEvent.pending()) 
    {
      lastMoveEvent.cancel();
      log("stopping previous move");
    }
  }
}
